/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rg3DRom.h"
#include "rg3DRom_csg.h"

int sk3DRomCSGCube::save_data( std::ostream &des) const
{
	write_dword( des, sk3DRomCSGCube::version());
	write_dword( des, m_flags);
	write_dword( des, m_bone_id);
	rgCSGCube::save_data( des);
	return 0;
}

int sk3DRomCSGCube::load_data( std::istream &src)
{
	// XXX: Read in the version number and load appropriate data.
        switch( read_dword( src)) {
	case 0: return load_version_0_data( src);
	default: return -1;
	}
}

int sk3DRomCSGCube::load_version_0_data( std::istream &src)
{
	m_flags = read_dword( src);
	m_bone_id = read_dword( src);
	rgCSGCube::load_data( src);
	return 0;
}

void sk3DRomCSGCube::dbg_bone_loss_check()
{
	// FIXME: This check is needed if a bone is deleted/removed
	//	without properly notifying the brush. This is needed
	//	because the bone's display-list/etc is out-of-sync.
	//	However, the ideal solution is to make sure the bone notifies
	//	all brushes before it is deleted/removed.
	if( m_bone_id >= 0) {
		m_pos.p = rgVector();
		m_bone_id = -1;
		refresh();
	}
}


void sk3DRomCSGCube::draw()
{
	if( group_id_valid()) {
		if( (m_model->group[ m_group_id]).hidden()) return;
		(m_model->group[ m_group_id]).set_color3();
	} else glColor3f( 0.4, 0.4, 0.4);

	if( !bone_id_valid()) {
		dbg_bone_loss_check();
		rgCSGCube::draw();
		return;
	}
	//const rgCoordSystem tmp( m_pos);
	//m_pos = coord_system();
	// FIXME: Should check if refresh is neccasary first.
	refresh();
	rgCSGCube::draw();
	//m_pos = tmp;
}

void sk3DRomCSGCube::gen_verts( std::vector <rgVector> &v) const
{
	map_verts( v, coord_system());
}

void sk3DRomCSGCube::real_draw_cursor( const rgVector vc, const rgVector ec)
{
	if( !bone_id_valid()) {
		dbg_bone_loss_check();
		rgCSGCube::real_draw_cursor( vc, ec);
		return;
	}
	//const rgCoordSystem tmp( m_pos);
	//m_pos = coord_system();
	// FIXME: Should check if refresh is neccasary first.
	refresh();
	rgCSGCube::real_draw_cursor( vc, ec);
	//m_pos = tmp;
}

void sk3DRomModel::bind_brush_to_bone( rgCSGBrush *b, const int bid)
{
	if( !b) return;
	if( (bid < 0) || (bid >= int( skel.b.size()))) return;

	const rgCoordSystem cs( b->coord_system());
	switch( b->type()) {
	case RG3DROM_CSG_CUBE: ((sk3DRomCSGCube *)b)->m_bone_id = bid; break;
	default: return;
	}

	b->set_coord_system( skel.b[bid].fcs.change_csys_csys( cs));
	b->refresh();
}

void sk3DRomModel::unbind_brush( rgCSGBrush *b) const
{
	if( !b) return;
	int *bid = 0;
	switch( b->type()) {
	case RG3DROM_CSG_CUBE: bid = &(((sk3DRomCSGCube *)b)->m_bone_id);
		break;
	default: return;
	}
	if( ((*bid) >= 0) && ((*bid) < int( skel.b.size()))) {
		b->set_coord_system( skel.b[ *bid].fcs.map_csys(
			b->coord_system()));
	}
	*(bid) = -1;
	b->refresh();
}

