/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <rgPort.h>
#include <iostream>
#include <string.h>
#include <math.h>
#include "rg3DRom.h"

void sk3DRomSkel::inbetween_skeleton( const float w,
	const sk3DRomSkelFrame &ska, const sk3DRomSkelFrame &skb,
	sk3DRomSkelFrame &skc) const
{
	// IMPORTANT NOTE: The skeleton interpolation can be done
	//	bone-by-bone in any arbitrary order. This is because every
	//	bone is controlled by a coordinate system. By simply
	//	interpolating the coordinate systems, the entire skeleton
	//	itself will be interpolated. Transformation and rendering
	//	functions gaurentee that bones will remain linked properly
	//	throughout the interpolation (so there is no need to check
	//	for that in this function).
	//
	//	Also, the ability for bone-by-bone transformation means that
	//	the output skeleton "skc" can be the same as "ska" or "skb"
	//	(that is "skc" has the same address as "ska" or "skb").

	int count = int( ska.b.size());
	std::vector <sk3DRomBoneFrame> :: const_iterator ia = ska.b.begin();
	std::vector <sk3DRomBoneFrame> :: const_iterator ib = skb.b.begin();
	std::vector <sk3DRomBoneFrame> :: iterator ic = skc.b.begin();
	while( count--) {
		rgCoordSystem::interpolate_csys(
			(*ia).rcs, (*ib).rcs, (*ic).rcs, w);
		++ia;
		++ib;
		++ic;
	}
}

void sk3DRomSkel::copy_mount_data( sk3DRomSkelFrame &des,
	const sk3DRomSkelFrame &src) const
{
	if( src.b.size() != des.b.size()) return;
	std::vector <sk3DRomBoneFrame> :: const_iterator isrc = src.b.begin();
	std::vector <sk3DRomBoneFrame> :: iterator ides = des.b.begin();
	int count = int( src.b.size());
	while( count--) {
		{
			const sk3DRomBoneFrame &sf = (*isrc);
			sk3DRomBoneFrame &df = (*ides);
			df.msb = sf.msb;
			df.mmb = sf.mmb;
			df.cms = sf.cms;
		}
		++isrc;
		++ides;
	}
}
