/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <rgPort.h>
#include <iostream>
#include <string.h>
#include "rg3DRom.h"

void sk3DRomDrawContext::init_fcs( const sk3DRomSkel &skel)
{
	m_fcs.create( skel.b.size());
}

int sk3DRomDrawContext::gen_frame_fcs( const sk3DRomSkel &s,
	const sk3DRomSkelFrame &f)
{
	if( m_fcs.size() != int( s.b.size())) init_fcs( s);
	if( m_fcs.size() != int( f.b.size())) return -1;
	// FIXME: Use a more efficient way of generating all "rcs"'s.
	for( int i = 0; i < m_fcs.size(); i++)
		m_fcs.data[i] = s.get_full_bone_rcs( i, f);
	return 0;
}

sk3DRomAniFrame::sk3DRomAniFrame()
{
	m_fid = -1;
	m_time = 250;	// By default, a frame will last a quarter of a second
	m_code = SK3DROM_ANI_CODE_TRANSITION;

	memset( data, 0, sizeof( int) * 6);	// FIXME: Write this better.
}

sk3DRomSequence::sk3DRomSequence()
{
	memset( m_name, 0, SK3DROM_SEQUENCE_NAME_LEN);
	edata = 0;
}

void sk3DRomSkel::delete_sequence( const int sid)
{
	if( (sid < 0) || (sid >= sequence_count())) return;
	// FIXME: Some sequences may reference the deleted sequence. Those
	//	sequences must be notified of the delete.
	seq.erase( seq.begin() + sid);
}

int sk3DRomSkel::get_total_seq_time( const int sid)
{
	if( sid < 0) return 0;
	if( sid >= int( seq.size())) return 0;

	int t = 0;
	const int istop = int( seq[sid].frame.size());
	for( int i = 0; i < istop; i++) {
		t += seq[sid].frame[i].time();
	}
	return t;
}

void sk3DRomSkel::get_seq_frame_and_scalar( const int sid,
	const int ttime, const float p, int &fid, float &s)
{
	const float k = float( ttime) * p;
	const int t = int( k);
	s = k - float( int( k));

	int n = 0;
	fid = 0;
	while( (n + seq[sid].frame[fid].time()) < t) {
		n += seq[sid].frame[fid].time();
		fid++;
	}

	// Determine amount of animation this frame spans.
	s = float( t - n) / float( seq[sid].frame[fid].time());
}

sk3DRomSkelFrame &sk3DRomSkel::animate( const int sid, const int fid,
	const int p_time, const int p_elapsed)
{
	return animate( sid, fid, p_time, p_elapsed, 0);
}

sk3DRomSkelFrame &sk3DRomSkel::animate( const int sid, const int fid,
	const int p_time, const int p_elapsed, sk3DRomDrawContext *p_dc)
{
	sk3DRomDrawContext *dc = &m_draw_context;
	if( p_dc) dc = p_dc;
	if( (sid < 0) || (sid >= sequence_count())) dc->m_cur_seq = -1;
	else dc->m_cur_seq = sid;

	if(	(dc->m_cur_seq < 0) ||
		(fid < 0) ||
		(fid >= seq[dc->m_cur_seq].frame_count()))
		dc->m_cur_frame = -1;
	else dc->m_cur_frame = fid;

	sk3DRomSkelFrame *dc_skout = &(dc->m_skout);
	if( dc == (&m_draw_context)) dc_skout = &skout;

	dc->init_current_frame( *this, *dc_skout);
	dc->m_frame_cur_time = 0;
	dc->m_frame_total_time = RG_MAX( 1, p_time);
	{ // Prepare to continue off the current frame (prevents animation
	  // "snapping").
		dc->m_skold = *dc_skout;
	}
	return animate( p_elapsed, dc);
}

sk3DRomSkelFrame &sk3DRomSkel::animate( const int p_elapsed)
{
	return animate( p_elapsed, 0);
}

sk3DRomSkelFrame &sk3DRomSkel::animate( const int p_elapsed,
	sk3DRomDrawContext *p_dc)
{
	sk3DRomDrawContext *dc = &m_draw_context;
	if( p_dc) dc = p_dc;
	sk3DRomSkelFrame *dc_skout = &(dc->m_skout);
	if( dc == (&m_draw_context)) dc_skout = &skout;

	if( dc_skout->b.size() != b.size()) init_skel_frame( *dc_skout);
	if( sequence_count() <= 0) return (*dc_skout);

	dc->m_frame_cur_time += p_elapsed;
	if( dc->m_frame_cur_time < dc->m_frame_total_time) {
		dc->gen_animation_frame( *this, *dc_skout);
		return (*dc_skout);
	}

	do {
		dc->m_frame_cur_time -= dc->m_frame_total_time;
		dc->choose_next_frame( *this, *dc_skout);
	} while( dc->m_frame_cur_time >= dc->m_frame_total_time);
	// FIXME: UPDATE: "dc_skout" is now being used. How does this effect
	//	the below problem?
	// FIXME: Is there a better way to set "m_skold" to continue off of
	//	the last frame? In general, "skout" does not have to contain
	//	the continuation of "m_skold". However, the continuation
	//	(currently transformed) skeleton frame has to be stored
	//	somewhere. One solution is to store it in an animation
	//	internal skeleton frame, or to use "skout", or finally, to
	//	have the user always supply this continuation frame.
	dc->m_skold = (*dc_skout);
	dc->gen_animation_frame( *this, *dc_skout);
	return (*dc_skout);
}

sk3DRomDrawContext::sk3DRomDrawContext()
{
	m_use_vertex_array = false;
	m_transform_direct = false;
	m_lock_fcs = false;

	m_cur_seq = 0;
	m_cur_frame = 0;
	m_frame_start = 0;
	m_frame_stop = 0;
	m_frame_cur_time = 0;
	m_frame_total_time = 0;
	m_root_master = 0;
	m_skel = 0;
}

void sk3DRomDrawContext::gen_animation_frame( sk3DRomSkel &skel,
	sk3DRomSkelFrame &dc_skout)
{
	gen_animation_frame( float( m_frame_cur_time) /
		float( m_frame_total_time), skel, dc_skout);
}

void sk3DRomDrawContext::gen_animation_frame( const float w, sk3DRomSkel &skel,
	sk3DRomSkelFrame &dc_skout)
{
	if( m_cur_frame < 0) return;
	sk3DRomAniFrame *f = &(skel.seq[m_cur_seq].frame[ m_cur_frame]);
	if( (f->fid() < 0) || (f->fid() >= skel.frame_count())) return;
	if( m_skold.b.size() != skel.b.size()) {
		m_skold = skel.sf[f->fid()];
		return;
	}
	skel.inbetween_skeleton( w, m_skold, skel.sf[f->fid()], dc_skout);
	apply_skeletal_lock( &m_old_frame, dc_skout, skel);
}

void sk3DRomDrawContext::set_skeletal_lock( const sk3DRomAniFrame *f,
	sk3DRomSkel &skel, sk3DRomSkelFrame &dc_skout)
{
	m_lock_type_set = SK3DROM_GENERIC_LOCK_SET;
	if( SK3DROM_ANI_CODE_JUMP == f->code()) {
		sk3DRomAniFrame_jump j;
		if( SK3DROM_ANI_LOCK_TYPE_JOINT == j.get_lock_type( *f)) {
			const int bid = j.get_target_bone( *f);
			if( (bid < 0) || (bid >= int( skel.b.size()))) return;
			const rgCoordSystem rcs(
				skel.get_full_bone_rcs( bid, dc_skout));
			m_lock_cs.p =
				rcs.map_point( skel.b[bid].get_jcs_rotate());
			return;

		} else if( SK3DROM_ANI_LOCK_TYPE_BONE == j.get_lock_type( *f)){
			const int bid = j.get_target_bone( *f);
			if( (bid < 0) || (bid >= int( skel.b.size()))) return;
			const rgCoordSystem rcs(
				skel.get_full_bone_rcs( bid, dc_skout));
			m_lock_cs = rcs;
//			m_lock_cs.p =
//				rcs.map_point( skel.b[bid].get_jcs_rotate());
			m_prev_orientation = m_orientation;
			return;

		} else if( SK3DROM_ANI_LOCK_TYPE_YAXIS == j.get_lock_type(*f)){
			const int bid = j.get_target_bone( *f);
			if( (bid < 0) || (bid >= int( skel.b.size()))) return;
			const rgCoordSystem rcs(
				skel.get_full_bone_rcs( bid, dc_skout));
			m_lock_cs = rcs;
			m_lock_cs.p =
				rcs.map_point( skel.b[bid].get_jcs_rotate());
		}
	}
}

void sk3DRomDrawContext::apply_skeletal_lock( const sk3DRomAniFrame *f,
	sk3DRomSkelFrame &s, sk3DRomSkel &skel)
{
	if( SK3DROM_NO_LOCK_SET == m_lock_type_set) {
		apply_frame_transformation();	
		set_skeletal_lock( f, skel, s);
	}
//	m_lock_type_set = SK3DROM_NO_LOCK_SET;
	// XXX: Apply the lock to the target bone's "root master". That is,
	//	apply the lock to the bone which has control over the target
	//	bone but has no master/controller itself. "root master" is
	//	the Master Mount Bone (mmb) which has no master itself.
	if( SK3DROM_ANI_CODE_JUMP == f->code()) {
		sk3DRomAniFrame_jump j;
		if( SK3DROM_ANI_LOCK_TYPE_JOINT == j.get_lock_type( *f)) {
			const int bid = j.get_target_bone( *f);

			if( this == &(skel.m_draw_context)) {
				const rgVector v( m_lock_cs.p -
				(skel.get_full_bone_rcs( bid, s)).map_point(
					skel.b[bid].get_jcs_rotate()));

				const int mid = skel.get_root_master( bid, s);
				skel.translate_rcs( mid, s, v);
				return;
			}

			m_frame_transform.loadIdentity();
			m_frame_transform.translate(
				(mat()) * (m_lock_cs.p -
				(skel.get_full_bone_rcs( bid, s)).map_point(
					skel.b[bid].get_jcs_rotate())));
			return;

		} else if( SK3DROM_ANI_LOCK_TYPE_BONE == j.get_lock_type( *f)){
			const int bid = j.get_target_bone( *f);
//			const int mid = skel.get_root_master( bid, s);
			const rgCoordSystem rcs( skel.get_full_bone_rcs( bid,
				s));
			m_orientation = m_lock_cs.map_csys(
				(rcs.change_csys_csys( m_prev_orientation)));
//			const rgVector pivot =
//				rcs.map_point( skel.b[bid].get_jcs_rotate());
//			skel.translate_rcs( mid, s, m_lock_cs.p -
//				rcs.map_point( skel.b[bid].get_jcs_rotate()));
			return;

		} else if( SK3DROM_ANI_LOCK_TYPE_YAXIS == j.get_lock_type(*f)){
			const int bid = j.get_target_bone( *f);
/*
			if( this == &(skel.m_draw_context)) {
				const rgVector v( m_lock_cs.p -
				(skel.get_full_bone_rcs( bid, s)).map_point(
					skel.b[bid].get_jcs_rotate()));

				const int mid = skel.get_root_master( bid, s);
				skel.translate_rcs( mid, s, v);
//				skel.rotate_rcs( mid, s, ang, rgVector(0,1,0));
			m_orientation = m_prev_orientation;
			m_orientation.rotate( ang, rgVector( 0, 1, 0));
				return;
			}
*/

///*


//			m_frame_translate.loadIdentity();
//			m_frame_translate.translate(
//			m_frame_transform.translate(
//			const rgVector tmp_move(
//				m * (m_lock_cs.p -
//				(skel.get_full_bone_rcs( bid, s)).map_point(
//					skel.b[bid].get_jcs_rotate())));
//*/

			//const int mid = skel.get_root_master( bid, s);
			const rgCoordSystem rcs( skel.get_full_bone_rcs( bid,
				s));

			rgVector prev( m_lock_cs.map_vector(
				rgVector( 1, 0, 0)));
			rgVector next( rcs.map_vector( rgVector( 1, 0, 0)));

			prev = (rgMakePlanar( prev, rgVector( 0, 1, 0))).
				normalize();
			next = (rgMakePlanar( next, rgVector( 0, 1, 0))).
				normalize();
			const float ang = -rgDotAngle( prev, next);

			const rgMatrix m( mat());
			m_frame_transform.loadIdentity();
			const rgVector tmp_pivot( m.pointMul( m_lock_cs.p));
			m_frame_transform.translate( tmp_pivot);
			m_frame_transform.rotate( ang, rgVector( 0, 1, 0));
			m_frame_transform.translate( -tmp_pivot);
			m_frame_transform.translate(
				tmp_pivot -
				m.pointMul( rcs.map_point(
					skel.b[bid].get_jcs_rotate())));
			return;
		}
	}
}

rgCoordSystem sk3DRomDrawContext::get_full_bone_rcs(
	const sk3DRomSkel &s,
	const int bid,
	const sk3DRomSkelFrame &sf,
	rgCoordSystem &mcs) const
{

//	if( this == &(s.m_draw_context)) {
		rgCoordSystem cs( s.get_full_bone_rcs( bid, sf, mcs));
		mcs = m_orientation.map_csys( mcs);
		return m_orientation.map_csys( cs);
//	}

//	rgCoordSystem cs( s.get_full_bone_rcs( bid, sf, mcs));
	
}

void sk3DRomDrawContext::init_current_frame( sk3DRomSkel &skel,
	sk3DRomSkelFrame &dc_skout)
{
	m_lock_type_set = SK3DROM_NO_LOCK_SET;
	if( m_cur_frame < 0) return;
	sk3DRomAniFrame *f = &(skel.seq[m_cur_seq].frame[ m_cur_frame]);
	m_old_frame = *f;
	switch( f->code()) {
	case SK3DROM_ANI_CODE_STOP:
	case SK3DROM_ANI_CODE_TRANSITION:
		break;

	case SK3DROM_ANI_CODE_JUMP:
		//{ sk3DRomAniFrame_jump j; j.reset_skip_count( *f);}
		break;
	}

	// XXX: Bone mountings are not interpolated between frames. Instead,
	//	the newly set frame will have its mounting data be in effect.
	//	Thus mounting data only changes when a new frame is set.
	if( (f->fid() < 0) || (f->fid() >= skel.frame_count())) return;
	skel.copy_mount_data( dc_skout, skel.sf[f->fid()]);
}

void sk3DRomDrawContext::choose_next_frame( sk3DRomSkel &skel,
	sk3DRomSkelFrame &dc_skout)
{
	if( m_cur_frame < 0) {
		m_frame_cur_time = 0;
		m_frame_total_time = 1;
		return;
	}

	sk3DRomAniFrame *f = &(skel.seq[m_cur_seq].frame[m_cur_frame]);
	m_frame_total_time = RG_MAX( 1, f->time());
	set_skeletal_lock( f, skel, dc_skout);
	m_old_frame = *f;
	switch( f->code()) {
	case SK3DROM_ANI_CODE_STOP: break;

	case SK3DROM_ANI_CODE_TRANSITION:
		if( (++(m_cur_frame)) >= skel.seq[m_cur_seq].frame_count())
			m_cur_frame = 0;
		break;

	case SK3DROM_ANI_CODE_JUMP:
		choose_next_jump_frame( f, skel);
		break;
	}

	apply_frame_transformation();

	// FIXME: "f" MUST always point to a valid frame. Perhaps an
	//	"ASSERT" should be placed here.
	f = &(skel.seq[m_cur_seq].frame[ m_cur_frame]);

	// XXX: Bone mountings are not interpolated between frames. Instead,
	//	the newly set frame will have its mounting data be in effect.
	//	Thus mounting data only changes when a new frame is set.
	if( (f->fid() >= 0) && (f->fid() < skel.frame_count()))
		skel.copy_mount_data( dc_skout, skel.sf[f->fid()]);
	//std::cerr << "f = " << (void *)f << "\n";
	//std::cerr << "m_cur_frame = " << m_cur_frame << "\n";
	//std::cerr << "m_frame_total_time = " << m_frame_total_time << "\n";
	return;
}

void sk3DRomDrawContext::choose_next_jump_frame( sk3DRomAniFrame *f,
	sk3DRomSkel &skel)
{
	sk3DRomAniFrame_jump j;
	// Check jump skip
	if( j.get_times_already_skipped( *f) < j.get_times_to_skip( *f)) {
		if( (++m_cur_frame) >= skel.seq[m_cur_seq].frame_count())
			m_cur_frame = 0;
		j.increment_skip_count( *f);
		return;
	}

	// Check jump probability
	const int n = int( j.get_probability( *f) * RAND_MAX);
	// FIXME: Why is int( float( 1.0) * RAND_MAX) == -2147483648 ?
	if( (n >= 0) && (rand() > n)) {
		if( (++m_cur_frame) >= skel.seq[m_cur_seq].frame_count())
			m_cur_frame = 0;
		return;
	}

	// Jump delay is not implemented since it is not neccasary

	// XXX: The jump must only be taken if the destination sequence and
	//	frame exist.
	int sid = j.get_des_sequence( *f);
	int fid = j.get_des_frame( *f);
	if( (sid < 0) || (sid >= skel.sequence_count())) {
		sid = -1;
		fid = -1;

	} else {
		if( (fid < 0) || (fid >= skel.seq[m_cur_seq].frame_count()))
			fid = -1;
	}
	if( (sid >= 0) && (fid >= 0)) {
		m_cur_seq = sid;
		m_cur_frame = fid;
	}
}


