/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rg3DRom_vertex_array.h"
#include <GL/gl.h>
#include <GL/glext.h>

int sk3DRomVertexArray::non_alpha_face_count( const sk3DRomModel &mdl)
{
        int count = 0;
        for( int i = 0; i < mdl.face_count(); i++)
                if( !((mdl.f[i]).is_alpha())) count++;
        return count;
}
                                                                                
void sk3DRomVertexArray::non_alpha_faces( const sk3DRomModel &mdl,
        std::vector <int> &naf)
{
        naf.clear();
        for( int i = 0; i < mdl.face_count(); i++)
                if( !((mdl.f[i]).is_alpha())) naf.push_back( i);
}

rgWORD *sk3DRomVertexArray::create_face_index_array(
        const sk3DRomModel &mdl, int &count,
        std::vector <sk3DRomVertexArray_RenderGroup> &rgrp)
{
        rgrp.clear();
        std::vector <int> naf;
	if( m_force_use_all_faces)
		for( int i = 0; i < mdl.face_count(); i++) naf.push_back( i);
	else non_alpha_faces( mdl, naf);
        if( naf.size() <= 0) return 0;

        count = int( naf.size());
	rgTempArray <rgWORD> tmp_idx( count * 3);
	rgTempArray <int> fid( count);
        if( tmp_idx.empty() || fid.empty()) return 0;
        rgWORD *idx = tmp_idx.detach();

	int *fptr = fid.data;
        int element = 0;
        for( int t = -1; t < mdl.texture_count(); t++) {
                sk3DRomVertexArray_RenderGroup rg;
                rg.m_start = element;
                rg.m_mdl_texid = t;
                if( t >= 0) rg.m_texid = (mdl.texture[t]).texture_id();
                for( int i = 0; i < int( naf.size()); i++) {
                        const sk3DRomFace &f = mdl.f[ naf[i]];
                        if( f.tid == t) {
                                idx[ element++] = f.vid[0];
                                idx[ element++] = f.vid[1];
                                idx[ element++] = f.vid[2];
				*(fptr++) = naf[i];
                                rg.m_count += 3;
//				if( t < 0) {
//			std::cout << " i = " << i << "\n";
//			std::cout << " naf[i] = " << naf[i] << "\n";
//			std::cout << " f.vid[0] = " << f.vid[0] << "\n";
//			std::cout << " f.vid[1] = " << f.vid[1] << "\n";
//			std::cout << " f.vid[2] = " << f.vid[2] << "\n";
//				}
                        }
                }
                if( rg.m_count > 0) rgrp.push_back( rg);
        }

	const int auxc = create_auxiliary_vertices( mdl, count * 3, idx,
		fid.data);
	if( auxc < 0) {
		delete [] idx;
		return 0;
	}
	std::cout << "auxc = " << auxc << "\n";

	create_accelerated_transformation_array( mdl);

	prepare_render_range( idx, rgrp);

        return idx;
}

void sk3DRomVertexArray::prepare_render_range(
	const rgWORD *idx,
	std::vector <sk3DRomVertexArray_RenderGroup> &rgrp) const
{
	if( !idx) return;
	for( int i = 0; i < int( rgrp.size()); i++) {
		rgrp[i].m_min_vertex_index = 0xFFFFFFFF;
		rgrp[i].m_max_vertex_index = 0;
		const rgWORD *p = idx + (rgrp[i]).m_start;
		for( int j = 0; j < (rgrp[i]).m_count; j++) {
			(rgrp[i]).m_min_vertex_index =
				RG_MIN( (rgrp[i]).m_min_vertex_index, p[j]);
			(rgrp[i]).m_max_vertex_index =
				RG_MAX( (rgrp[i]).m_max_vertex_index, p[j]);
		}
		std::cout << "\tvertex range = ["
			<< (rgrp[i]).m_min_vertex_index << ", "
			<< (rgrp[i]).m_max_vertex_index << "]\n";
	}
}


class tmp_unique_vertex
{
public:
	int vid;	// Real (model) vertex data
	int fid;	// Which contains vertex data
	int face_vid;
	float u;	// Texture "u" coord (which makes it unique)
	float v;	// Texture "v" coord (which makes it unique)
	float w;	// Texture "w" coord (which makes it unique)
	rgVector n;	// Normal (which makes it unique)
	bool used;

	tmp_unique_vertex() { used = false;}

	bool operator ==( const tmp_unique_vertex &vert) const {
	// FIXME: All faces MUST share vertex normals since currently the
	//	transformation system does not transform these extra
	//	normals. If different normals are required, then the modeler
	//	must manual duplicate the vertices. This solution has the
	//	draw back that those extra vertices add redundant vertex
	//	transformations. However, flat shaded models (or models
	//	with different vertex normals) are not optimal anyway.
	//	const float epsilon = 0.001f;
	//	if( fabs( n.x - vert.n.x) > epsilon) return false;
	//	if( fabs( n.y - vert.n.y) > epsilon) return false;
	//	if( fabs( n.z - vert.n.z) > epsilon) return false;

		const float texcoord_epsilon = 0.0001f;
		if( fabs( u - vert.u) > texcoord_epsilon) return false;
		if( fabs( v - vert.v) > texcoord_epsilon) return false;
		if( fabs( w - vert.w) > texcoord_epsilon) return false;
		return true;
	}

	friend std::ostream &operator <<( std::ostream &o,
		const tmp_unique_vertex &u) {
		o	<< "{\n"
			<< "\tused = " << u.used << "\n"
			<< "\tu = " << u.u << "\n"
			<< "\tv = " << u.v << "\n"
			<< "\tw = " << u.w << "\n"
			<< "\tvid = " << u.vid << "\n"
			<< "\tfid = " << u.fid << "\n"
			<< "\tface_vid = " << u.face_vid << "\n}\n";
		return o;
	}
};

bool claim_vertex(
	rgWORD &vid,
	const tmp_unique_vertex &v,
	const int core_count,
	std::vector <tmp_unique_vertex> &vert)
{
	if( !(vert[vid].used) || (v == vert[vid])) {
		vert[vid] = v;
		return true;
	}

	for( int i = core_count; i < int( vert.size()); i++) {
		if( !(vert[i].used) || (v == vert[i])) {
			vid = i;
			return true;
		}
	}

	vid = vert.size();
	vert.push_back( v);
	return false;
}

int sk3DRomVertexArray::create_auxiliary_vertices(
	const sk3DRomModel &mdl,
	const int idx_size,
	rgWORD *idx,
	const int *fid)
{

	if( m_aux_vertex) {
		m_aux_vertex_count = 0;
		delete [] m_aux_vertex;
		m_aux_vertex = 0;
	}

	if( m_vertex_fid) {
		m_vertex_fid_count = 0;
		delete [] m_vertex_fid;
		m_vertex_fid = 0;
	}

	// Create core vertices
	std::vector <tmp_unique_vertex> vert;
	{
		for( int i = 0; i < mdl.vertex_count(); i++)
			vert.push_back( tmp_unique_vertex());
	}

	// Every face will try to claim three vertices (creating new ones if it
	// cannot claim any existing vertices).
	const int core_count = int( vert.size());
	std::cout << "core_count = " << core_count << "\n";
	{
		for( int i = 0; i < idx_size; i++) {
			tmp_unique_vertex v;
			v.used = true;
			v.vid = idx[i];
			v.fid = fid[ i / 3];

			{
				const int j = v.face_vid = i % 3;
				v.n = mdl.f[ v.fid].n[j];
				v.u = mdl.f[ v.fid].map[j].x;
				v.v = mdl.f[ v.fid].map[j].y;
				v.w = mdl.f[ v.fid].map[j].z;
			}

			claim_vertex(
				idx[i],
				v,
				core_count,
				vert);
		}
	}

	// Every face now has a compatible vertex index
	m_aux_vertex_count = RG_MAX( 0, int( vert.size()) - core_count);
	if( m_aux_vertex_count > 0) {
		m_aux_vertex = new int[ m_aux_vertex_count];
		if( !m_aux_vertex) {
			m_aux_vertex_count = 0;
			return -2;
		}
		for( int i = core_count; i < int( vert.size()); i++)
			m_aux_vertex[ i - core_count] = vert[i].vid;
	}
	
	{
		m_vertex_fid_count = int( vert.size());
		m_vertex_fid = new int[ m_vertex_fid_count];
		for( int i = 0; i < m_vertex_fid_count; i++) {
			if( vert[i].used) {
			if( (vert[i].face_vid < 0) ||
				(vert[i].face_vid >= 3)) {
				std::cerr << "sk3DRomVertexArray::"
					"create_auxiliary_vertices: "
					<< "vid fail = "
					<< vert[i].face_vid << "\n";
				std::cerr << "i = " << i << "\n";
				std::cerr << vert[i] << "\n";
				exit(0);
			}
				set_vertex_face_fid( i, vert[i].fid);
				set_vertex_face_vid( i, vert[i].face_vid);
//std::cout << i << ", " << vert[i].fid << ", " << vert[i].face_vid << "\n";
			} else {
				set_vertex_face_fid( i, 0);
				set_vertex_face_vid( i, 0);
			}
		}
	}

	return m_aux_vertex_count;
}

void hlp_gen_bone_weight_lists(
	std::vector <int> &wlist,
	std::vector <int> &ulist,
	const sk3DRomModel &mdl,
	const sk3DRomBone &b)
{
	for( int j = 0; j < int( b.vid.size()); j++) {
		if( 1.0f == (mdl.v[ b.vid[j]]).weight)
			ulist.push_back( b.vid[j]);
		else wlist.push_back( b.vid[j]);
	}
}

// Returns the size 
int sk3DRomVertexArray::calc_accel_array_size( const sk3DRomModel &mdl) const
{
	if( mdl.skel.bone_count() <= 0) return -1;

	// XXX: Safety checks are performed because this is a high performance
	//	hardware dependent function. Even so, it uses "standard"
	//	descriptions of data sizes to make sure it will work on
	//	different platforms. If a platform has a strange hardware
	//	setup, then acceleration will be turned off, rather than
	//	crashing.
	//
	//	The accelerated array has a granularity of "sizeof( rgDWORD)"
	//	which should be 32 bits.
	// 
	//	Some elements of the array will be interpreted as a 32 bit word
	//	instead of a 32 bit float.
	if( sizeof( rgDWORD) != sizeof( rgFLOAT)) return -2;

	const int jcs_size = sizeof( mdl.skel.b[0].jcs);
	if( (jcs_size % sizeof( rgDWORD))) return -3;


	int count = 0;
	for( int i = 0; i < mdl.skel.bone_count(); i++) {
		count += jcs_size;		// Bone jcs
		count += sizeof( rgDWORD);	// Bone mid

		std::vector <int> wlist;
		std::vector <int> ulist;
		hlp_gen_bone_weight_lists( wlist, ulist, mdl, mdl.skel.b[i]);

		// Weighted vertex elements have 8 values (all 32 bits).
		//	(off, x, y, z, w, nx, ny, nz).
		// Unweighted vertex elements have 7 values (all 32 bits).
		//	(off, x, y, z, nx, ny, nz).
		count += sizeof( rgDWORD);	// Weighted vertex count
		count += sizeof( rgDWORD) * 8 * int( wlist.size());

		count += sizeof( rgDWORD);	// Unweighted vertex count
		count += sizeof( rgDWORD) * 7 * int( ulist.size());
	}
	return count;
}

bool sk3DRomVertexArray::create_accelerated_transformation_array(
	const sk3DRomModel &mdl)
{
	if( m_vn_accel) { delete [] m_vn_accel; m_vn_accel = 0;}
	if( !m_vertex_fid) return false;	// Must support index array
	if( normal_stride() <= 0) return false;	// Must support normals

	const int array_size = calc_accel_array_size( mdl) / sizeof( rgDWORD);
	if( array_size <= 0) return false;
	if( !(m_vn_accel = new rgDWORD[ array_size])) return false;

	std::cout << "array_size = " << (array_size * sizeof( rgDWORD)) << ", "
		<< array_size << " DWORDs\n";
	int wtotal = 0;
	int utotal = 0;

	rgDWORD *p = m_vn_accel;
	for( int i = 0; i < mdl.skel.bone_count(); i++) {
		// jcs (Joint Coordinate System)
		*((rgCoordSystem *)p) = mdl.skel.b[i].jcs;
		p += sizeof( rgCoordSystem) / sizeof( rgDWORD);

		// mid
		// XXX: It is NOT really "mid" (master bone index), instead
		//	it is "mid" with an offset applied. This is an
		//	optimization which allows the accelerated array
		//	reader to increment through bones inorder AND to
		//	"randomly" index them at the same time. This means that
		//	"mid" may index "behind" the pointer. This may throw
		//	off the cpu cache prediction, however the fact that
		//	bones are "randomly" indexed would have thrown off the
		//	cache prediction any ways.
		// XXX: Profiling has confirmed this optimization to work.
		*p = mdl.skel.b[i].mid - i;
		p++;

		std::vector <int> wlist;
		std::vector <int> ulist;
		hlp_gen_bone_weight_lists( wlist, ulist, mdl, mdl.skel.b[i]);
		wtotal += wlist.size();
		utotal += ulist.size();

		// Weighted vertex count
		*(p++) = int( wlist.size());
		{ // Weighted vertices and normals
			rgFloat *f = (rgFloat *)p;
			for( int j = 0; j < int( wlist.size()); j++) {
				const int vid = wlist[j];

				// Offset
				((rgDWORD *)f)[0] = vid * vertex_stride();

				// Vertex
				f[1] = (mdl.v[ vid]).x;
				f[2] = (mdl.v[ vid]).y;
				f[3] = (mdl.v[ vid]).z;

				// Weight
				f[4] = (mdl.v[ vid]).weight;

				// Normal
				//const rgCoordSystem cs( mdl.skel.b[i].jcs);
				const rgCoordSystem cs(
					mdl.skel.get_full_bone_jcs( i));
				const rgVector n( cs.change_vector_csys(
					mdl.f[ vertex_face_fid( vid)].n[
						vertex_face_vid( vid)]));
				f[5] = n.x;
				f[6] = n.y;
				f[7] = n.z;

				f += 8;
			}
			p = (rgDWORD *)f;
		}

		// Unweighted vertex count
		*(p++) = int( ulist.size());
		{ // Unweighted vertices and normals
			rgFloat *f = (rgFloat *)p;
			for( int j = 0; j < int( ulist.size()); j++) {
				const int vid = ulist[j];

				// Offset
				((rgDWORD *)f)[0] = vid * vertex_stride();

				// Vertex
				f[1] = (mdl.v[ vid]).x;
				f[2] = (mdl.v[ vid]).y;
				f[3] = (mdl.v[ vid]).z;

				// Normal
				//const rgCoordSystem cs( mdl.skel.b[i].jcs);
				const rgCoordSystem cs(
					mdl.skel.get_full_bone_jcs( i));
				const rgVector n( cs.change_vector_csys(
					mdl.f[ vertex_face_fid( vid)].n[
						vertex_face_vid( vid)]));
				f[4] = n.x;
				f[5] = n.y;
				f[6] = n.z;

				f += 7;
			}
			p = (rgDWORD *)f;
		}
	}

	std::cout << "Total weighted = " << wtotal << ", unweighted = " <<
		utotal << "\n";
	return true;
}

sk3DRomVNTArray::sk3DRomVNTArray()
{
	m_vertex_count = 0;
	m_data = 0;
}

sk3DRomVNTArray::~sk3DRomVNTArray()
{
	delete_arrays();
}

bool sk3DRomVNTArray::export_to_OBJ( std::ostream &o, const sk3DRomModel &mdl)
{
	if( !m_force_share_vertex) {
		o <<
			"# Error: Non-sharing vertices selected.\n"
			"#	Currently only supporting sharing vertices.\n";
		return false;
	}

	{
		o << "# Vertices: " << m_vertex_count << "\n";
		const float *p = vertex_array_start();
		for( int i = 0; i < m_vertex_count; i++) {
			o << "v " << p[0] << " " << p[1] << " " << p[2]
				<< "\n";
			p += vertex_stride();
		}
		o << "\n";
	}

	{
		o << "# Texture Coordniates: " << m_vertex_count << "\n";
		// XXX: Note that vertex and texcoord arrays are interleaved.
		//	So vertex_stride() == texcoord_stride().
		const float *p = texcoord_offset( vertex_array_start());
		for( int i = 0; i < m_vertex_count; i++) {
			o << "vt " << p[0] << " " << p[1] << "\n";
			p += vertex_stride();
		}
		o << "\n";
	}

	{
		o << "# Normals: " << m_vertex_count << "\n";
		// XXX: Note that vertex and normal arrays are interleaved.
		//	So vertex_stride() == normal_stride().
		const float *p = normal_offset( vertex_array_start());
		for( int i = 0; i < m_vertex_count; i++) {
			o << "vn " << p[0] << " " << p[1] << " " << p[2]
				<< "\n";
			p += vertex_stride();
		}
		o << "\n";
	}

	if( index_array()) {
		const std::vector <sk3DRomVertexArray_RenderGroup> &rg =
	                face_render_group();
		o << "# Groups: " << rg.size() << "\n";
		for( int i = 0; i < int( rg.size()); i++) {
			{
				o << "g " << "group_" << i << "\n";
				o << "#usemtl ";
				const int tid = rg[i].m_mdl_texid;
				if( (tid < 0) || (tid >= mdl.texture_count()))
					o << "none\n";
				else {
					// FIXME: Find a better way to get
					//	file name.
					const int fname_len = 300;
					char fname[ fname_len];
					mdl.texture[tid].filename( fname,
						fname_len);
					o << fname << " # <-- This is a "
						"file name\n";
				}
			}
			
			o << "# Faces (strictly triangles) in this group: "
				<< (rg[i].m_count / 3) << "\n";
			const rgWORD *p = index_array() + rg[i].m_start;
			for( int j = 0; j < rg[i].m_count; j += 3){
				o << "f "
					// XXX: OBJ is backwards compared to
					//	3DRomED.
					<< (p[2] + 1) << "/"
					<< (p[2] + 1) << "/"
					<< (p[2] + 1) << " "

					<< (p[1] + 1) << "/"
					<< (p[1] + 1) << "/"
					<< (p[1] + 1) << " "

					<< (p[0] + 1) << "/"
					<< (p[0] + 1) << "/"
					<< (p[0] + 1) << "\n";
				p += 3;
			}
			o << "\n";
		}

	} else {
		o << "# Error: Cannot get face indices, cannot write faces\n";
		return false;
	}

	return true;
}

int sk3DRomVNTArray::create_arrays( const sk3DRomModel &mdl)
{
	delete_arrays();
	if( mdl.vertex_count() <= 0) return -1;
	if( mdl.face_count() <= 0) return -2;
        
	// 64 bytes per vertex
        // 4=<x,y,z,w> 3=<nx, ny, nz> 1=<fid> 3=<u, v, w> 5=<pad>

        if( m_force_share_vertex) return create_shared_vertex_array( mdl);
        else return create_non_shared_vertex_array( mdl);
}

void sk3DRomVNTArray::dump_vertex( std::ostream &o, const int i)
{
	if( (i < 0) || (i >= m_vertex_count))
		o << "[bad vertex id = " << i << "]";
	float *p = vertex_array_start() + (i * vertex_stride());
	o	<< "V( " << p[0] << ", " << p[1] << ", " << p[2] << "), "
		<< "N( " << p[4] << ", " << p[5] << ", " << p[6] << "), "
		<< "U=" << (*(texcoord_offset( p))) << ", "
		<< "V=" << (*(texcoord_offset( p) + 1)) << ", "
		<< "W=" << (*(texcoord_offset( p) + 2));
}

int sk3DRomVNTArray::create_shared_vertex_array( const sk3DRomModel &mdl)
{
	create_face_index_array( mdl);

	m_vertex_count = aux_vertex_count() + mdl.vertex_count();

	m_data = new float[ m_vertex_count * vertex_stride()];
	if( !m_data) return -3;
	if( !vertex_face_id_array()) return -4;

	float *p = vertex_array_start();
	for( int i = 0; i < m_vertex_count; i++) {
		const sk3DRomFace &f = mdl.f[ vertex_face_fid( i)];
		const int vid = vertex_face_vid( i);
		*(normal_offset( p)) = f.n[vid].x;
		*(normal_offset( p) + 1) = f.n[vid].y;
		*(normal_offset( p) + 2) = f.n[vid].z;

		*(texcoord_offset( p)) = f.map[vid].x;
		*(texcoord_offset( p) + 1) = f.map[vid].y;
		*(texcoord_offset( p) + 2) = f.map[vid].z;
		p += vertex_stride();
	}
        return 0;
}

int sk3DRomVNTArray::create_non_shared_vertex_array( const sk3DRomModel &mdl)
{
                                                                                
        std::vector <int> naf;
        non_alpha_faces( mdl, naf);
        if( naf.size() <= 0) return -3;

        m_vertex_count = int( naf.size()) * 3;
        if( !(m_data = new float[ m_vertex_count * vertex_stride()]))
		return -4;

	int count = 0;
	float *p = m_data;
	for( int t = -1; t < mdl.texture_count(); t++) {
                                                                                
                sk3DRomVertexArray_RenderGroup rg;
                rg.m_start = count;
                rg.m_mdl_texid = t;
                if( t >= 0) rg.m_texid = mdl.texture[t].texture_id();
                for( int i = 0; i < int( naf.size()); i++) {
                        const sk3DRomFace &f = mdl.f[ naf[i]];
                        if( t == f.tid) {
                                fid_value( p) = naf[i];
                                *(texcoord_offset( p)) = f.map[0].x;
                                *(texcoord_offset( p) + 1) = f.map[0].y;
                                *(texcoord_offset( p) + 2) = f.map[0].z;
                                p += vertex_stride();

                                fid_value( p) = naf[i];
                                *(texcoord_offset( p)) = f.map[1].x;
                                *(texcoord_offset( p) + 1) = f.map[1].y;
                                *(texcoord_offset( p) + 2) = f.map[1].z;
                                p += vertex_stride();

                                fid_value( p) = naf[i];
                                *(texcoord_offset( p)) = f.map[2].x;
                                *(texcoord_offset( p) + 1) = f.map[2].y;
                                *(texcoord_offset( p) + 2) = f.map[2].z;
                                p += vertex_stride();

                                rg.m_count++;
                                count++;
                        }
                }
                if( rg.m_count > 0) m_vertex_render_group.push_back( rg);
        }
                                                                                
        return 0;
}


void sk3DRomVNTArray::set_use_all_faces( const bool b)
{
        m_force_use_all_faces = b;
        delete_arrays();
}

void sk3DRomVNTArray::set_share_vertex( const bool b)
{
        m_force_share_vertex = b;
        delete_arrays();
}
                                                                                
void sk3DRomVNTArray::delete_arrays()
{
        m_vertex_render_group.clear();
        m_vertex_count = 0;
        if( m_data) { delete [] m_data; m_data = 0;}
}

void sk3DRomVNTArray::update_auxiliary_vertices()
{
	if( !m_aux_vertex) return;
	const int *av = m_aux_vertex;
	float *des = m_data + ((m_vertex_count - m_aux_vertex_count) *
		vertex_stride());
	const int vstride = vertex_stride();
	for( int i = 0; i < m_aux_vertex_count; i++) {
		// Copy vertex data
		const float *src = m_data + ((*(av++)) * vstride);
		des[0] = src[0];
		des[1] = src[1];
		des[2] = src[2];

		// Copy normal data
		// XXX: The normal offset is harded coded with index numbers
		//	instead of using "normal_offset()" for speed.
		des[4] = src[4];
		des[5] = src[5];
		des[6] = src[6];

		// Vertices, normals and texture coords are packed together.
		// So incrementing by "vertex_stride" increments normals and
		// texture coords as well.
		des += vstride;
	}
}

void sk3DRomVNTArray::update_arrays( const sk3DRomModel &mdl)
{
        if( !m_data) return;
        update_texture_ids( mdl);
        if( m_force_share_vertex) {
		update_auxiliary_vertices();
		return;
	}
        float *p = m_data;
        const float *stop = m_data + (m_vertex_count * vertex_stride());
        while( p < stop) {
                const sk3DRomFace &f = mdl.f[ fid_value( p)];
                update_vertex( p, mdl, f, 0);
                update_vertex( p + 16, mdl, f, 1);
                update_vertex( p + 32, mdl, f, 2);
                p += 48;
        }
}

void sk3DRomVNTArray::update_texture_ids( const sk3DRomModel &mdl)
{
        if( m_force_share_vertex) {
                std::vector <sk3DRomVertexArray_RenderGroup> &rg =
                        face_render_group();
                for( int i = 0; i < int( rg.size()); i++) {
                        if( rg[i].m_mdl_texid < 0) rg[i].m_texid = 0;
                        else rg[i].m_texid = (mdl.texture[ rg[i].m_mdl_texid]).
				texture_id();
                }
		return;
	}

	std::vector <sk3DRomVertexArray_RenderGroup> &rg =
		m_vertex_render_group;
	for( int i = 0; i < int( rg.size()); i++) {
		if( rg[i].m_mdl_texid < 0) rg[i].m_texid = 0;
		else rg[i].m_texid = (mdl.texture[ rg[i].m_mdl_texid]).
			texture_id();
	}
}


void sk3DRomVNTArray::render_arrays()
{
	if( !m_data) return;

	glVertexPointer( 3, GL_FLOAT, 64, m_data);
	glNormalPointer( GL_FLOAT, 64, normal_offset( m_data));
	switch( m_texcoord_size) {
	case 1:
		glTexCoordPointer( 1, GL_FLOAT, 64, texcoord_offset( m_data));
		break;
	case 3:
		glTexCoordPointer( 3, GL_FLOAT, 64, texcoord_offset( m_data));
		break;
	default:
		glTexCoordPointer( 2, GL_FLOAT, 64, texcoord_offset( m_data));
		break;
	}
		

	glEnableClientState( GL_TEXTURE_COORD_ARRAY);
	glEnableClientState( GL_NORMAL_ARRAY);
	glEnableClientState( GL_VERTEX_ARRAY);

	if( m_force_share_vertex) {
		if( !index_array()) return;
		const std::vector <sk3DRomVertexArray_RenderGroup> &rg =
	                face_render_group();
	        for( int i = 0; i < int( rg.size()); i++) {
			glBindTexture( GL_TEXTURE_2D, rg[i].m_texid);
			glDrawRangeElements(
				GL_TRIANGLES,
				rg[i].m_min_vertex_index,
				rg[i].m_max_vertex_index, 
				rg[i].m_count,
				GL_UNSIGNED_SHORT,
				index_array() + rg[i].m_start);
		}

	} else {

	        const std::vector <sk3DRomVertexArray_RenderGroup> &rg =
	                m_vertex_render_group;
		for( int i = 0; i < int( rg.size()); i++) {
			glBindTexture( GL_TEXTURE_2D, rg[i].m_texid);
			glDrawArrays(   GL_TRIANGLES,
				rg[i].m_start * 3,
				rg[i].m_count * 3);
		}
	}

	glDisableClientState( GL_VERTEX_ARRAY);
	glDisableClientState( GL_NORMAL_ARRAY);
	glDisableClientState( GL_TEXTURE_COORD_ARRAY);
}
