/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgNURBS.h"

rgNURBS::rgNURBS()
{
	set_degree( 4);
}

rgNURBS::rgNURBS( const int p_degree)
{
	set_degree( p_degree);
}

void rgNURBS::set_degree( const int d)
{
	if( d < 0) return;
	m_degree = d;

	m_cpoint.clear();
	m_knot.clear();
	for( int i = 0; i < (m_degree + cpoint_count() + 1); i++)
		m_knot.push_back( 1.0f);
	m_start_knot = 0;
}

#define Check_base_case \
	if( p_degree <= 0) { \
		if( (Ki <= k) && (k <= Kj)) return 1.0f; \
		return 0.0f; \
	}

rgFloat rgNURBS::N( const int i, const int p_degree, const rgFloat k,
	std::vector <rgFloat> :: const_iterator p_Knot) const
{
	const rgFloat Ki = *(p_Knot + i);
	const rgFloat Kj = *(p_Knot + i + 1);

	Check_base_case;

	const rgFloat Kd = *(p_Knot + i + p_degree);
	const rgFloat Ke = *(p_Knot + i + p_degree + 1);
	const float a = Kd - Ki;
	const float b = Ke - Kj;

	#define TMP_LEFT ((k - Ki) / a) * N( i, p_degree - 1, k, p_Knot)
	#define TMP_RIGHT ((Ke - k) / b) * N( i + 1, p_degree - 1, k, p_Knot)
	if( 0.0f == a) {
		if( 0.0f == b) return 0;
		else return TMP_RIGHT;
	} else {
		if( 0.0f == b) return TMP_LEFT;
		else return TMP_LEFT + TMP_RIGHT;
	}
	#undef TMP_LEFT
	#undef TMP_RIGHT
}

rgVector rgNURBS::P( const rgFloat u)
{
	/*
	set_start_knot( u);
	std::cout << "sk = " << m_start_knot << ",\t";

	const int start_point = m_start_knot - (m_degree - 1);
	std::vector <rgNURBS_Vec> :: const_iterator cp =
		m_cpoint.begin() + start_point;
	rgVector p;
	float k = 0.0f;
	for( int i = 0; i < m_degree; i++) {
		const float n = N( start_point + i, u);
		const rgNURBS_Vec &v = *cp;
		p += (v.m_w * v) * n;
		k += v.m_w * n;
		++cp;
	}
	return (k * p);
	*/

	std::vector <rgNURBS_Vec> :: const_iterator cp = m_cpoint.begin();
	rgVector p;
	float k = 0.0f;
	for( int i = 0; i < int( m_cpoint.size()); i++) {
		const float n = N( i, u);
		const rgNURBS_Vec &v = *cp;
		const float t = v.m_w * n;
		p += t * v;
		k += t;
		++cp;
	}
	return (k * p);
}

void rgNURBS::gen_standard_knot_sequence()
{
	if( m_knot.empty()) return;

	const int kcount = int( m_knot.size());

	{ for( int i = 0; i <= m_degree; i++) m_knot[i] = 0.0f;}

	{
		const rgFloat ustep = 1.0f / (kcount - (m_degree * 2));
		rgFloat u = ustep;
		for( int i = m_degree + 1; i < (kcount - m_degree); i++) {
			m_knot[i] = u;
			u += ustep;
		}
	}
	

	{ for( int i = kcount - m_degree - 1; i < kcount; i++)
		m_knot[i] = 1.0f;
	}
}

void rgNURBS::gen_uniform_knots()
{
	if( m_knot.empty()) return;
	rgFloat u = 0.0f;
	const rgFloat ustep = 1.0f / rgFloat( m_knot.size());
	std::vector <rgFloat> :: iterator K = m_knot.begin();
	int count = int( m_knot.size());
	while( count--) {
		*K = u;
		u += ustep;
		++K;
	}
}
