/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <rgStandard.h>
#include <vector>

class rgNURBS_Vec : public rgVector
{
public:
	rgFloat m_w;

	rgNURBS_Vec() { m_w = 1.0f;}
	rgNURBS_Vec( const rgVector &v) : rgVector( v) { m_w = 1.0f;}
	rgNURBS_Vec &operator =( const rgVector &v) {
		x = v.x; y = v.y; z = v.z; return (*this);}

	friend std::ostream &operator <<( std::ostream &o,
		const rgNURBS_Vec &v) {
		return (o << "( " << v.x << ", " << v.y << ", " << v.z <<
			", " << v.m_w << ")");	
	}
};

class rgNURBS
{
public:

	rgNURBS();
	rgNURBS( const int p_degree);


	void prepend_cpoint( const rgVector &p) {
		m_cpoint.insert( m_cpoint.begin(), p);
		m_knot.insert( m_knot.begin() + m_degree, 1.0f);
	}

	void append_cpoint( const rgVector &p) {
		m_cpoint.push_back( p);
		m_knot.push_back( 1.0f);
	}

	void insert_cpoint( const rgVector &p, int i) {
		if( i < 0) prepend_cpoint( p);
		else if( i > int( m_cpoint.size())) append_cpoint( p);
		else {
			m_cpoint.insert( m_cpoint.begin() + i, p);
			m_knot.insert( m_knot.begin() + i + m_degree, 1.0f);
		}
	}

	rgVector get_point( const rgFloat u) {
		if( cpoint_count() <= m_degree) return rgVector();
		if( u <= 0.0f) return P( 0.0f);
		if( u >= 1.0f) return P( 1.0f);
		return P( u);
	}

	rgFloat N( int i, const rgFloat k) const {
		std::vector <rgFloat> :: const_iterator K = m_knot.begin();
		return N( i, m_degree, k, K);
	}

	void set_degree( const int d);
	void gen_uniform_knots();
	void gen_standard_knot_sequence();

	int cpoint_count() const { return int( m_cpoint.size());}

	void list_knots( std::ostream &o) {
		for( int i = 0; i < int( m_knot.size()); i++)
			o << "K[" << i << "] = " << m_knot[i] << "\n";
	}

private:
	int m_degree;
	int m_start_knot;

	std::vector <rgNURBS_Vec> m_cpoint;
	std::vector <rgFloat> m_knot;

	rgVector P( const rgFloat u);

	rgFloat N( const int i, const int d, const rgFloat k,
		std::vector <rgFloat> :: const_iterator K) const;

	void set_start_knot( const rgFloat u) {
		std::vector <rgFloat> :: const_iterator K = m_knot.begin();
		if( u < (*(K + m_start_knot)))
			m_start_knot = find_start_knot( u, K, 0) + m_degree;
		else if( u >= (*(K + m_start_knot + 1))) {
			m_start_knot = find_start_knot( u, K + m_start_knot,
				m_start_knot);
		}
	}

	int find_start_knot( const rgFloat u,
		std::vector <rgFloat> :: const_iterator K,
		int i) const
	{
		//std::cout << "i = " << i << "\n";
		int count = int( m_knot.size()) - i;
		while( count--) {
			if( u < (*(++K))) return i;
			++i;
		}
		return i - 1;
	}
};

