/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGOBJ_H
#define RGOBJ_H

#include <fstream>
#include <vector>
#include <rgStandard.h>
#include <string>

#define RGOBJ_MAX_COMMAND_LENGTH 63
#define RGOBJ_MAX_ERR_MSG_LENGTH 255

class rgOBJVertex : public rgVector
{
public:
	
};

class rgOBJNormal : public rgVector
{
public:
};

class rgOBJTexCoord : public rgVector
{
public:
};


class rgOBJFaceIndices
{
public:
	int m_vid;
	int m_nid;
	int m_tcid;
	int m_pad;

	rgOBJFaceIndices() { 
		m_vid = -1;
		m_nid = -1;
		m_tcid = -1;
	}
};

class rgOBJFace
{
public:
	std::vector <rgOBJFaceIndices> m_indices;
};

class rgOBJGroup
{
public:
	std::string m_name;
	std::vector <rgOBJFace> m_face;

	int face_count() const { return int( m_face.size());}
};

class rgOBJ
{
public:

	bool fail_on_errors;
	std::string m_emsg;


        std::vector <rgOBJVertex> m_vert;
	std::vector <rgOBJNormal> m_normal;
	std::vector <rgOBJTexCoord> m_texcoord;
	std::vector <rgOBJGroup> m_group;

	rgOBJ();
	~rgOBJ();

	void clear();
	int load( const char *fname);
	int load( std::istream &inf);

	bool face_is_good( rgOBJFace &f) const;

	void remove_duplicate_vertices();

	int vertex_count() const { return int( m_vert.size());}
	int normal_count() const { return int( m_normal.size());}
	int texcoord_count() const { return int( m_texcoord.size());}
	int group_count() const { return int( m_group.size());}
	int total_face_count() const {
		int count = 0;
		for( int i = 0; i < group_count(); i++)
			count += m_group[i].face_count();
		return count;
	}

	rgOBJFace *get_face( const int id);

	rgVector face_normal( rgOBJFace &f, const int vid) {
		if( vid < 0) return rgVector();
		if( vid >= int( f.m_indices.size())) return rgVector();
		if( f.m_indices[ vid].m_nid < 0) return rgVector();
		if( f.m_indices[ vid].m_nid >= normal_count())
			return rgVector();
		return m_normal[ f.m_indices[ vid].m_nid];
	}
	rgVector face_texcoord( rgOBJFace &f, const int vid) {
		if( vid < 0) return rgVector();
		if( vid >= int( f.m_indices.size())) return rgVector();
		if( f.m_indices[ vid].m_tcid < 0) return rgVector();
		if( f.m_indices[ vid].m_tcid >= texcoord_count())
			return rgVector();
		return m_texcoord[ f.m_indices[ vid].m_tcid];
	}

private:

};

int get_command( std::istream &inf, char *cmd, int &i);
int match_command( char *cmd, const int len);
void end_line( std::istream &inf);

#endif
