/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <iostream>
#include <stdio.h>
#include "rgOBJ_parser.h"

void rgobj_parser_generic_error_check( Obj_parse_data *p)
{
	if( p->inf->fail()) {
		if( !(p->inf->eof())) {
			p->obj->m_emsg = "Line " + rg_to_string( p->line) +
				": File read error.";
			p->err = CM_FILE_ERROR;
		}
		p->done = true;
	}
}

void hf_comment( Obj_parse_data *p)
{
        end_line( *(p->inf));
        p->line++;
	rgobj_parser_generic_error_check( p);
}
                                                                                
void hf_unknown( Obj_parse_data *p)
{
        end_line( *(p->inf));
	p->line++;
	rgobj_parser_generic_error_check( p);
}
                                                                                
void hf_vertex( Obj_parse_data *p)
{
        std::istream &f = *(p->inf);
	const int e = p->get_line_end( p->inf);

	rgOBJVertex v;
	f >> v.x;
	if( f.tellg() >= e) {
		p->obj->m_emsg = "Line " + rg_to_string( p->line) +
			": Unexpected '\\n' (new line) after reading X "
			"coordinate.";
		p->err = CM_PARSE_ERROR;
		return;
	}
	
	f >> v.y;
	if( f.tellg() >= e) {
		p->obj->m_emsg = "Line " + rg_to_string( p->line) +
			": Unexpected '\\n' (new line) after reading Y "
			"coordinate.";
		p->err = CM_PARSE_ERROR;
		return;
	}

	f >> v.z;
	if( f.tellg() < e) end_line( f);
	p->line++;

	p->obj->m_vert.push_back( v);
	rgobj_parser_generic_error_check( p);
}

void hf_normal( Obj_parse_data *p)
{
	std::istream &f = *(p->inf);
	const int e = p->get_line_end( p->inf);

	rgOBJNormal n;
	f >> n.x;
	if( f.tellg() >= e) {
		p->obj->m_emsg = "Line " + rg_to_string( p->line) +
			": Unexpected '\\n' (new line) after reading X "
			"coordinate.";
		p->err = CM_PARSE_ERROR;
		return;
	}
	
	f >> n.y;
	if( f.tellg() >= e) {
		p->obj->m_emsg = "Line " + rg_to_string( p->line) +
			": Unexpected '\\n' (new line) after reading Y "
			"coordinate.";
		p->err = CM_PARSE_ERROR;
		return;
	}

	f >> n.z;
	if( f.tellg() < e) end_line( f);
	p->line++;

	p->obj->m_normal.push_back( n);

	rgobj_parser_generic_error_check( p);
}

void hf_texcoord( Obj_parse_data *p)
{
	std::istream &f = *(p->inf);
	const int e = p->get_line_end( p->inf);

	rgOBJTexCoord tc;
	f >> tc.x;
	if( f.tellg() >= e) {
		p->obj->m_emsg = "Line " + rg_to_string( p->line) +
			": Unexpected '\\n' (new line) after reading U "
			"coordinate.";
		p->err = CM_PARSE_ERROR;
		return;
	}
	
	f >> tc.y;
	if( f.tellg() < e) {
		f >> tc.z;	// Read 'W' coord if present
		if( f.fail()) {
			tc.z = 1.0f;
			f.clear();
		}
	}

	if( f.tellg() < e) end_line( f);
	p->line++;

	p->obj->m_texcoord.push_back( tc);

	rgobj_parser_generic_error_check( p);
}

void hf_group( Obj_parse_data *p)
{
	std::istream &f = *(p->inf);
	const int e = p->get_line_end( p->inf);

	p->obj->m_group.push_back( rgOBJGroup());
	rgOBJGroup &g = p->obj->m_group[ p->obj->m_group.size() - 1];

	f >> g.m_name;
	if( f.tellg() < e) end_line( f);

	rgobj_parser_generic_error_check( p);
}

int hlp_read_face_sub_index( std::istream &f, const int end_of_line, char &c)
{
	std::string tmp;
	bool leading_white_space = true;
	while( f.tellg() < end_of_line) {
		c = f.get();
		if( (' ' == c) || ('\t' == c)) {
			if( !leading_white_space) break;

		} else if( ('/' == c) || ('\\' == c)) {
			break;

		} else {
			tmp += c;
			leading_white_space = false;
		}
	}
	if( tmp.length() <= 0) return -1;
	return atoi( tmp.c_str());
}

int hlp_read_face_indices(
	Obj_parse_data *p,
	std::istream &f,
	const int end_of_line,
	rgOBJFaceIndices &fid)
{
	fid = rgOBJFaceIndices();

	// XXX: Must subtract one from face indices since OBJ files start from
	//	1 istead of 0.
	char c;
	fid.m_vid = hlp_read_face_sub_index( f, end_of_line, c);
	if( fid.m_vid < 0)
		fid.m_vid = p->obj->vertex_count() + fid.m_vid;
	else fid.m_vid--;
	if( (' ' == c) || ('\t' == c)) return 0;
	else if( ('/' != c) && ('\\' != c)) return -1;

	fid.m_tcid = hlp_read_face_sub_index( f, end_of_line, c);
	if( fid.m_tcid < 0)
		fid.m_tcid = p->obj->texcoord_count() + fid.m_tcid;
	else fid.m_tcid--;
	if( (' ' == c) || ('\t' == c) || ('#' == c)) return 0;

	fid.m_nid = hlp_read_face_sub_index( f, end_of_line, c);
	if( fid.m_nid < 0)
		fid.m_nid = p->obj->normal_count() + fid.m_nid;
	else fid.m_nid--;
	if( (' ' == c) || ('\t' == c)) return 0;
	else if( ('/' != c) && ('\\' != c)) return -2;

	return -3;
}

void hf_face( Obj_parse_data *p)
{
	std::istream &f = *(p->inf);
	const int e = p->get_line_end( p->inf);

	rgOBJFace face;
	while( f.tellg() < e) {

		rgOBJFaceIndices fid;
		hlp_read_face_indices( p, f, e, fid);
		face.m_indices.push_back( fid);
	}

	if( !(p->obj->face_is_good( face))) {
		rgobj_parser_generic_error_check( p);
		return;
	}

	if( p->obj->m_group.empty())
		p->obj->m_group.push_back( rgOBJGroup());
	rgOBJGroup &g = p->obj->m_group[ p->obj->m_group.size() - 1];
	g.m_face.push_back( face);
	rgobj_parser_generic_error_check( p);
}
