/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 *
 * IMPORTANT NOTE: This is an internal header file and should only be included
 *	by files that implement parts of the rgOBJ parser/loader.
 */

#ifndef RGOBJ_PARSER_H
#define RGOBJ_PARSER_H

#include "rgOBJ.h"

#define CM_NONE -1
#define CM_UNKNOWN -2
#define CM_COMMENT -3
#define CM_FILE_ERROR -5
#define CM_TOO_LONG -6
#define CM_EOF -7
#define CM_PARSE_ERROR -8

/*
#define RGOBJ_PARSER_GENERIC_ERROR_CHECK \
	{if( p->inf->fail()) { \
		if( !(p->inf->eof())) { \
			p->obj->m_emsg = "Line " + rg_to_string( p->line) + \
				": File read error."; \
			p->err = CM_FILE_ERROR; \
		} \
                p->done = true; \
        }}
*/

/* Data used while parsing obj files. */
class Obj_parse_data
{
public:

        std::istream *inf;
        int line;
        int err;
        bool done;
        rgOBJ *obj;

	char *cmd;
	int cmd_len;

	int m_negative_vertex_offset;
	int m_negative_normal_offset;
	int m_negative_texcoord_offset;

	Obj_parse_data();
	~Obj_parse_data();

	int get_line_end( std::istream *inf) {
		if( !inf) return -1;
		const int t = inf->tellg();
        	end_line( *inf);
		int e = inf->tellg();
		if( inf->fail()) e = -1;
        	inf->clear( std::ios::goodbit);
       		inf->seekg( t, std::ios::beg);
		return e;
	}


};
                                                                                
/*********** FUNCTION HANDLER PROTOTYPES ************/
/* Append your new handler function at the end of the list. */

void hf_comment( Obj_parse_data *p);
void hf_unknown( Obj_parse_data *p);     /* Use this to ignore any command */
void hf_vertex( Obj_parse_data *p);
void hf_normal( Obj_parse_data *p);
void hf_texcoord( Obj_parse_data *p);
void hf_group( Obj_parse_data *p);
void hf_face( Obj_parse_data *p);

/*********** END OF FUNCTION HANDLER PROTOTYPES ***********/

#endif
