/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgPhysics.h"
#include <rgOpenGL.h>

void rgCollisionQuad::render() const
{
	glNormal( normal());
	glVertex( m_v[0]);
	glVertex( m_v[1]);
	glVertex( m_v[2]);
	glVertex( m_v[3]);
}

void rgCollisionQuad::render_normal( const float k) const
{
	const rgVector p( (m_v[0] + m_v[2]) * 0.5f);
	glVertex( p);
	glVertex( p + (normal() * k));
}

// XXX: Quads overlap if "this" quad contains any "cq" point, or if "this"
//	quad splits any "cq" edge, or if "this" quad is contained by "cq".
//
// FIXME: The below statment is not true due to the "robustness" of
//	  "rgPlanarPointInQuad".
//	Also, quads "MUST" be facing in the exact same direction (normals are
//	parallel and in the same direction).
bool rgCollisionQuad::overlaps( const rgCollisionQuad &cq) const
{
	if( contains_quad_points( cq)) return true;
	if( cq.contains_quad_points( *this)) return true;
	if( splits_quad_edges( cq)) return true;
	return false;
}

void rgCollisionShadow::set_verts( const rgCollisionPrisim &cp,
	short a, short b, short c, short d, const rgVector n)
{
	m_v[0] = rgMakePlanar( cp.m_v[a], n);
	m_v[1] = rgMakePlanar( cp.m_v[b], n);
	m_v[2] = rgMakePlanar( cp.m_v[c], n);
	m_v[3] = rgMakePlanar( cp.m_v[d], n);
	m_vcount = 4;
}

void rgCollisionShadow::set_verts( const rgCollisionPrisim &cp,
	short a, short b, short c, short d, short e, short f,
	const rgVector n)
{
	m_v[0] = rgMakePlanar( cp.m_v[a], n);
	m_v[1] = rgMakePlanar( cp.m_v[b], n);
	m_v[2] = rgMakePlanar( cp.m_v[c], n);
	m_v[3] = rgMakePlanar( cp.m_v[d], n);
	m_v[4] = rgMakePlanar( cp.m_v[e], n);
	m_v[5] = rgMakePlanar( cp.m_v[f], n);
	m_vcount = 6;
}

rgCollisionShadow::rgCollisionShadow( const rgCollisionPrisim &cp,
	const rgVector n)
{
	m_center = cp.center();
	char quad_key = 0;

	if( !add_shadow_quad( cp.normal_A(), n, 001, quad_key))
		add_shadow_quad( cp.normal_C(), n, 004, quad_key);

	if( !add_shadow_quad( cp.normal_B(), n, 002, quad_key))
		add_shadow_quad( cp.normal_D(), n, 010, quad_key);

	if( !add_shadow_quad( cp.normal_E(), n, 020, quad_key))
		add_shadow_quad( cp.normal_F(), n, 040, quad_key);

	switch( quad_key) {
	// F E D|C B A
	// 5 4 3|2 1 0
	case 001: set_verts( cp, 0, 1, 2, 3, n); break; // A
	case 002: set_verts( cp, 1, 7, 6, 2, n); break; // B
	case 004: set_verts( cp, 7, 4, 5, 6, n); break; // C
	case 010: set_verts( cp, 4, 0, 3, 5, n); break; // D
	case 020: set_verts( cp, 4, 7, 1, 0, n); break; // E
	case 040: set_verts( cp, 3, 2, 6, 5, n); break; // F

	case 021: set_verts( cp, 2, 3, 4, 7, n); break; // AE
	case 003: set_verts( cp, 0, 7, 6, 3, n); break; // AB
	case 041: set_verts( cp, 0, 1, 6, 5, n); break; // AF
	case 011: set_verts( cp, 1, 2, 5, 4, n); break; // AD
	case 006: set_verts( cp, 1, 4, 5, 2, n); break; // BC
	case 022: set_verts( cp, 2, 0, 4, 6, n); break; // BE
	case 042: set_verts( cp, 1, 7, 5, 3, n); break; // BF
	case 014: set_verts( cp, 7, 0, 3, 6, n); break; // CD
	case 024: set_verts( cp, 1, 0, 5, 6, n); break; // CE
	case 044: set_verts( cp, 7, 4, 3, 2, n); break; // CF
	case 030: set_verts( cp, 7, 1, 3, 5, n); break; // DE
	case 050: set_verts( cp, 4, 0, 2, 6, n); break; // DF

	case 031: set_verts( cp, 7, 1, 2, 3, 5, 4, n); break; // ADE (tri at 0)
	case 023: set_verts( cp, 0, 4, 7, 6, 2, 3, n); break; // ABE (tri at 1)
	case 043: set_verts( cp, 0, 1, 7, 6, 5, 3, n); break; // ABF (tri at 2)
	case 051: set_verts( cp, 1, 2, 6, 5, 4, 0, n); break; // ADF (tri at 3)
	case 034: set_verts( cp, 3, 5, 6, 7, 1, 0, n); break; // CDE (tri at 4)
	case 054: set_verts( cp, 7, 4, 0, 3, 2, 6, n); break; // CDF (tri at 5)
	case 046: set_verts( cp, 1, 7, 4, 5, 3, 2, n); break; // BCF (tri at 6)
	case 026: set_verts( cp, 5, 6, 2, 1, 0, 4, n); break; // BCE (tri at 7)
	// 4----7
	// | E  |
	// |    |
	// 0----1----7----4----0
	// | A  | B  | C  | D  |
	// |    |    |    |    |
	// 3----2----6----5----3
	// | F  |
	// |    |
	// 5----6

	// XXX: "default" MUST NEVER occur.
	default: *((int *)0) = 0; break;
	}
}

void rgCollisionShadow::draw() const
{
	glBegin( GL_POLYGON);
	for( int i = 0; i < m_vcount; i++) glVertex( m_v[i]);
	glEnd();
}

bool rgCollisionShadow::overlaps( const rgCollisionShadow &cs) const
{
	if( contains_shadow_points( cs)) return true;
	if( cs.contains_shadow_points( *this)) return true;
	if( splits_shadow_edges( cs)) return true;
        return false;
}

//#define HELP_SMP_CPUS_UNDERSTAND
#ifdef HELP_SMP_CPUS_UNDERSTAND
// XXX: Symetric Multiprocessing Cpus should optimize this code. If they do
//	not, then they do not deserve the SMP distinction.
//	Also, if "hyper-threading" does not optimize this, then it is a load
//	of garbage.
bool rgCollisionShadow::splits_shadow_edges( const rgCollisionShadow &cs) const
{
	#define ISECT( a, b, c, d) rgLineSegLineSegIntersect( \
		m_v[a], m_v[b], cs.m_v[c], cs.m_v[d])
	#define IF_ISECT( a, b, c, d) if( ISECT( a, b, c, d))
	if( 4 == cs.m_vcount) {
		for( int i = 0; i < (m_vcount - 1); i++) {
			bool tmp[] = { false, false, false, false};
			tmp[0] = ISECT( i, i + 1, 0, 1);
			tmp[1] = ISECT( i, i + 1, 1, 2);
			tmp[2] = ISECT( i, i + 1, 2, 3);
			tmp[3] = ISECT( i, i + 1, 3, 0);
			if( tmp[0] || tmp[1] || tmp[2] || tmp[3]) return true;
		}
		IF_ISECT( m_vcount - 1, 0, 0, 1) return true;
		IF_ISECT( m_vcount - 1, 0, 1, 2) return true;
		IF_ISECT( m_vcount - 1, 0, 2, 3) return true;
		IF_ISECT( m_vcount - 1, 0, 3, 0) return true;

	} else {
		for( int i = 0; i < (m_vcount - 1); i++) {
			bool tmp[] = { false, false, false, false, false,
				false};
			tmp[0] = ISECT( i, i + 1, 0, 1);
			tmp[1] = ISECT( i, i + 1, 1, 2);
			tmp[2] = ISECT( i, i + 1, 2, 3);
			tmp[3] = ISECT( i, i + 1, 3, 4);
			tmp[4] = ISECT( i, i + 1, 4, 5);
			tmp[5] = ISECT( i, i + 1, 5, 0);
			if( tmp[0] || tmp[1] || tmp[2] || tmp[3] || tmp[4] ||
				tmp[5]) return true;
		}
		IF_ISECT( m_vcount - 1, 0, 0, 1) return true;
		IF_ISECT( m_vcount - 1, 0, 1, 2) return true;
		IF_ISECT( m_vcount - 1, 0, 2, 3) return true;
		IF_ISECT( m_vcount - 1, 0, 3, 4) return true;
		IF_ISECT( m_vcount - 1, 0, 4, 5) return true;
		IF_ISECT( m_vcount - 1, 0, 5, 0) return true;
	}
	return false;
}
#else
bool rgCollisionShadow::splits_shadow_edges( const rgCollisionShadow &cs) const
{
	#define ISECT( a, b, c, d) rgLineSegLineSegIntersect( a, b, c, d)
	#define IF_ISECT( a, b, c, d) if( ISECT( m_v[a], m_v[b], cs.m_v[c], \
		cs.m_v[d]))
	if( 4 == cs.m_vcount) {
		for( int i = 0; i < (m_vcount - 1); i++) {
			IF_ISECT( i, i + 1, 0, 1) return true;
			IF_ISECT( i, i + 1, 1, 2) return true;
			IF_ISECT( i, i + 1, 2, 3) return true;
			IF_ISECT( i, i + 1, 3, 0) return true;
		}
		IF_ISECT( m_vcount - 1, 0, 0, 1) return true;
		IF_ISECT( m_vcount - 1, 0, 1, 2) return true;
		IF_ISECT( m_vcount - 1, 0, 2, 3) return true;
		IF_ISECT( m_vcount - 1, 0, 3, 0) return true;

	} else {
		for( int i = 0; i < (m_vcount - 1); i++) {
			IF_ISECT( i, i + 1, 0, 1) return true;
			IF_ISECT( i, i + 1, 1, 2) return true;
			IF_ISECT( i, i + 1, 2, 3) return true;
			IF_ISECT( i, i + 1, 3, 4) return true;
			IF_ISECT( i, i + 1, 4, 5) return true;
			IF_ISECT( i, i + 1, 5, 0) return true;
		}
		IF_ISECT( m_vcount - 1, 0, 0, 1) return true;
		IF_ISECT( m_vcount - 1, 0, 1, 2) return true;
		IF_ISECT( m_vcount - 1, 0, 2, 3) return true;
		IF_ISECT( m_vcount - 1, 0, 3, 4) return true;
		IF_ISECT( m_vcount - 1, 0, 4, 5) return true;
		IF_ISECT( m_vcount - 1, 0, 5, 0) return true;
	}
	return false;
}
#endif


/*
bool rgCollisionShadow::overlaps( const rgCollisionShadow &cs) const
{
	for( int i = 0; i < m_qcount; i++) {
		for( int j = 0; j < cs.m_qcount; j++)
			if( m_quad[i].overlaps( cs.m_quad[j])) return true;
	}
	return false;
}
*/

#ifdef asdfasdfasdf3244r43f32f32fewferfwrf
bool rgCollisionShadow::span_overlaps( const rgCollisionShadow &ics,
	const rgCollisionShadow &fcs) const
{
	// TODO: Must have a minimum span (there is no span if there is no
	//	minimum span).
	const rgVector icenter( ics.center());
	const rgVector fcenter( fcs.center());
	if( (icenter - fcenter).length_squared() < 1.0f)
		return rotate_span_overlaps( ics, fcs, icenter);
	return linear_span_overlaps( ics, fcs, icenter, fcenter);
}

bool rgCollisionShadow::rotate_span_overlaps(
	const rgCollisionShadow &ics,
	const rgCollisionShadow &fcs,
	const rgVector p_center) const
{
	//if( m_vcount <= 0) return false;
	
	return false;
}

bool rgCollisionShadow::linear_span_overlaps(
	const rgCollisionShadow &ics,
	const rgCollisionShadow &fcs,
	const rgVector icen,
	const rgVector fcen) const
{
	return false;
}
#endif

// Prisim definition:
// XXX: Without the gaurd bars on the right, the drawing would produce
//      a "multi-line comment" warning/error.
//                  |
// 0-----1          |
// |\    |\         |
// | 3-----2        |
// | |   | |        |
// 4-|---7 |        |
//  \|    \|        |
//   5-----6        |
//
// Point 4 is at the origin point (0, 0, 0);
rgCollisionPrisim &rgCollisionPrisim::create( float x_scale, float y_scale,
	float z_scale)
{
	m_v[0] = rgVector( 0, y_scale, 0);
	m_v[1] = rgVector( x_scale, y_scale, 0);
	m_v[2] = rgVector( x_scale, y_scale, z_scale);
	m_v[3] = rgVector( 0, y_scale, z_scale);
	m_v[4] = rgVector( 0, 0, 0);
	m_v[5] = rgVector( 0, 0, z_scale);
	m_v[6] = rgVector( x_scale, 0, z_scale);
	m_v[7] = rgVector( x_scale, 0, 0);
	//refresh_normals();

	inv_xlen = 1.0f / x_scale;
	inv_ylen = 1.0f / y_scale;
	inv_zlen = 1.0f / z_scale;
	return *this;
}

rgCollisionPrisim &rgCollisionPrisim::create( const std::vector <rgVector> &v,
	const rgMatrix &m)
{
	if( v.size() != 8) return *this;
	m_v[0] = m.pointMul( v[0]);
	m_v[1] = m.pointMul( v[1]);
	m_v[2] = m.pointMul( v[2]);
	m_v[3] = m.pointMul( v[3]);
	m_v[4] = m.pointMul( v[4]);
	m_v[5] = m.pointMul( v[5]);
	m_v[6] = m.pointMul( v[6]);
	m_v[7] = m.pointMul( v[7]);
	inv_xlen = 1.0f / (m_v[0] - m_v[1]).length();
	inv_ylen = 1.0f / (m_v[0] - m_v[4]).length();
	inv_zlen = 1.0f / (m_v[0] - m_v[3]).length();
	return *this;
}

void rgCollisionPrisim::render() const
{
	(quad_A()).render();
	(quad_B()).render();
	(quad_C()).render();
	(quad_D()).render();
	(quad_E()).render();
	(quad_F()).render();
}

std::ostream &operator <<( std::ostream &o, const rgCollisionPrisim &p)
{
	o << "rgCollisionPrisim ( " << ((void *)&p) << ") {\n\t" <<
		p.m_v[0] << "\n\t" <<
		p.m_v[1] << "\n\t" <<
		p.m_v[2] << "\n\t" <<
		p.m_v[3] << "\n\t" <<
		p.m_v[4] << "\n\t" <<
		p.m_v[5] << "\n\t" <<
		p.m_v[6] << "\n\t" <<
		p.m_v[7] << "\n}";
	return o;
}

bool rgCollisionPrisim::collision( const rgVector p) const
{
	const rgVector x( (m_v[7] - m_v[4]) * inv_xlen);
	const rgVector y( (m_v[0] - m_v[4]) * inv_ylen);
	const rgVector z( (m_v[5] - m_v[4]) * inv_zlen);
	const rgVector u( p - m_v[4]);
	const rgVector w( u.x * inv_xlen, u.y * inv_ylen, u.z * inv_zlen);
	const float tx = w.dot( x);
	if( (tx < 0.0f) || (tx > 1.0f)) return false;
	const float ty = w.dot( y);
	if( (ty < 0.0f) || (ty > 1.0f)) return false;
	const float tz = w.dot( z);
	if( (tz < 0.0f) || (tz > 1.0f)) return false;
	return true;
}

bool rgCollisionPrisim::collision( const rgVector p, const float radius) const
{
	const rgVector x( (m_v[7] - m_v[4]) * inv_xlen);
	const rgVector y( (m_v[0] - m_v[4]) * inv_ylen);
	const rgVector z( (m_v[5] - m_v[4]) * inv_zlen);
	const rgVector u( p - m_v[4]);
	const rgVector w( u.x * inv_xlen, u.y * inv_ylen, u.z * inv_zlen);
	const float rx = radius * inv_xlen;
	const float ry = radius * inv_ylen;
	const float rz = radius * inv_zlen;
	const float tx = w.dot( x);
	if( (tx < -rx) || (tx > (1.0f + rx))) return false;
	const float ty = w.dot( y);
	if( (ty < -ry) || (ty > (1.0f + ry))) return false;
	const float tz = w.dot( z);
	if( (tz < -rz) || (tz > (1.0f + rz))) return false;
	return true;
}

// Description: Performs collision detection and returns collision information.
//	This collision detection function is for a single point. "a" is the
//	starting point, "b" is then ending point.
//
// FIXME: UPDATE: The function now tolerates point "a" being within the prisim.
// Precondition: Point "a" MUST be outside of the prisim volume, otherwise all
//	the collision check algorithms break down. That is, if point "a" is
//	within the prisim, then a collision has already occured, and this
//	function will have no idea how it happened, and thus cannot produce
//	meaningfull collision data.
bool rgCollisionPrisim::collision(
	const rgVector a,
	const rgVector b,
	const float radius,
	rgVector &surface_normal,
	rgVector &surface_point,
	rgVector &intersect_point,
	rgVector &intrusion_vector) const
{
	// Prepare basis vectors
	const rgVector x( (m_v[7] - m_v[4]) * inv_xlen);
	const rgVector y( (m_v[0] - m_v[4]) * inv_ylen);
	const rgVector z( (m_v[5] - m_v[4]) * inv_zlen);

	// Expand prisim to simulate as if the points have radius (ie. are
	// spheres).
	const float rx = radius * inv_xlen;
	const float ry = radius * inv_ylen;
	const float rz = radius * inv_zlen;

	// Intrusion point relative to m_v[4] (point 4) = (ix, iy, iz).
	// If intrusion point does not intrude at all, then no collision
	// occured. If the intrusion point is planar to a quad edge then it
	// is NOT intruding (the point MUST be inside).
//	{
		const rgVector u( b - m_v[4]);
//		const rgVector w( u.x * inv_xlen, u.y * inv_ylen,
//			u.z * inv_zlen);

		// FIXME: Does "pipe-lining" really work? So far, it seems to
		//	be just a marketing gimick. Also, are compilers smart
		//	enough to pipeline this automatically? A large porition
		//	of the low level functions can benefit greatly from
		//	real pipe-lining.
		const float ix = u.dot( x) * inv_xlen;
		if( (ix <= -rx) || (ix >= (1.0f + rx))) return false;
		const float iy = u.dot( y) * inv_ylen;
		if( (iy <= -ry) || (iy >= (1.0f + ry))) return false;
		const float iz = u.dot( z) * inv_zlen;
		if( (iz <= -rz) || (iz >= (1.0f + rz))) return false;
//	}

//	const rgVector u( a - m_v[4]);
//	const rgVector w( u.x * inv_xlen, u.y * inv_ylen, u.z * inv_zlen);
//	const float ox = w.dot( x);
//	const float oy = w.dot( y);
//	const float oz = w.dot( z);


	int winner = 0;
	{
		const float golf_score[6] = {

			1.0f - (iy + ry),	// Quad A
			1.0f - (ix + rx),	// Quad B
			iy - ry,		// Quad C
			ix - rx,		// Quad D
			iz - rz,		// Quad E
			1.0f - (iz + rz)};	// Quad F

		float best_score = golf_score[0];
		for( int i = 1; i < 6; i++) {
			if( golf_score[i] < best_score) {
				best_score = golf_score[i];
				winner = i;
			}
		}
	}
	switch( winner) {
	case 1:	surface_normal = x;
		surface_point = plane_point_B();
		break;

	case 2:	surface_normal = -y;
		surface_point = plane_point_C();
		break;

	case 3:	surface_normal = -x;
		surface_point = plane_point_D();
		break;

	case 4:	surface_normal = -z;
		surface_point = plane_point_E();
		break;

	case 5:	surface_normal = z;
		surface_point = plane_point_F();
		break;

	default:
		surface_normal = y;
		surface_point = plane_point_A();
		break;
	}
	intrusion_vector = rgVector();
	const rgVector tmp( b - (surface_normal * radius));
	intersect_point = tmp -
		(surface_normal.dot( tmp - surface_point) * surface_normal);
	return true;
}


#ifdef asdfasdfsdf
{
	// Prepare basis vectors
	const rgVector x( (m_v[7] - m_v[4]) * inv_xlen);
	const rgVector y( (m_v[0] - m_v[4]) * inv_ylen);
	const rgVector z( (m_v[5] - m_v[4]) * inv_zlen);

	// Expand prisim to simulate as if the points have radius (ie. are
	// spheres).
	const float rx = (radius + 1.0) * inv_xlen;
	const float ry = (radius + 1.0) * inv_ylen;
	const float rz = (radius + 1.0)* inv_zlen;

	// Intrusion point relative to m_v[4] (point 4) = (ix, iy, iz).
	// If intrusion point does not intrude at all, then no collision
	// occured. If the intrusion point is planar to a quad edge then it
	// is NOT intruding (the point MUST be inside).
	{
		const rgVector u( b - m_v[4]);
		const rgVector w( u.x * inv_xlen, u.y * inv_ylen,
			u.z * inv_zlen);

		// FIXME: Does "pipe-lining" really work? So far, it seems to
		//	be just a marketing gimick. Also, are compilers smart
		//	enough to pipeline this automatically? A large porition
		//	of the low level functions can benefit greatly from
		//	real pipe-lining.
		const float ix = w.dot( x);
		if( (ix <= -rx) || (ix >= (1.0f + rx))) return false;
		const float iy = w.dot( y);
		if( (iy <= -ry) || (iy >= (1.0f + ry))) return false;
		const float iz = w.dot( z);
		if( (iz <= -rz) || (iz >= (1.0f + rz))) return false;
	}

	// Gaurd against null vector AB.
	/*
	if( (fabs( a.x - b.x) + fabs( a.y - b.y) + fabs( a.z - b.z)) <
		0.0001) {
		surface_normal = rgVector();
		intrusion_vector = rgVector();
		surface_point = a;
		intersect_point = a;
		return true;
	}
	*/

	const rgVector u( a - m_v[4]);
	const rgVector w( u.x * inv_xlen, u.y * inv_ylen, u.z * inv_zlen);
	const float ox = w.dot( x);
	const float oy = w.dot( y);
	const float oz = w.dot( z);

/*
	if(	(ox >= -rx) && (ox <= (1.0f + rx)) &&
		(oy >= -ry) && (oy <= (1.0f + ry)) &&
		(oz >= -rz) && (oz <= (1.0f + rz))) {
		surface_normal = rgVector();
		intrusion_vector = rgVector();
		surface_point = a;
		intersect_point = a;
		return true;
	}
*/

	// Determine which quad/plane/surface splits the segment AB. An
	// "optimized" and confusing tecqnique is used (the below diagrams
	// are provided to "aid" in understanding).
	// 0-----1          | Basis vectors:
	// |\    |\         | X = P7 - P4
	// Y 3-----2        | Y = P0 - P4
	// | |   | |        | Z = P5 - P4
	// 4-|-X-7 |        |
	//  Z|    \|        |
	//   5-----6        |
	//
	// 4----7
	// | E  |
	// |    |
	// 0----1----7----4----0
	// | A  | B  | C  | D  |
	// |    |    |    |    |
	// 3----2----6----5----3
	// | F  |
	// |    |
	// 5----6

	// ba: normalized vector point in the direction (b - a).
	const rgVector ba( (b - a).normalize());

	// Intrusion vector = Point"b" minus collision point "p" + point
	// radius (with appropriate scaling).
	// FIXME: Does placing braces around the statments help the compiler
	//	in optimizing the calculations?
/*
	#define SET_INTRUSION_VECTOR( v, n) \
		{ \
			(v).x = b.x - p.x; \
			(v).y = b.y - p.y; \i
			(v).z = b.z - p.z; \
			if( ba.dot( v) < 0) { \
				(v) = radius 
			} \
		}

				const float k = float( \
					sqrt( 3.0f * radius * radius)); \
				(v).x += ba.x * k; \
				(v).y += ba.y * k; \
				(v).z += ba.z * k; \
				if( ba.dot( v) < 0) { \
			std::cerr << "Failure!!!\n"; \
			std::cerr << (v) << "\n"; \
			std::cerr << "rad = " << radius << "\n"; \
			std::cerr << "n = " << (n) << "\n"; \
			std::cerr << "ba = " << ba << "\n"; \
			std::cerr << "a = " << a << "\n"; \
			std::cerr << "b = " << b << "\n"; \
			std::cerr << "p = " << p << "\n"; \
			exit(0); \
				} \
			} \
		}


				const float k = float( \
					sqrt( 3.0f * radius * radius)); \
				(v).x += ba.x * k; \
				(v).y += ba.y * k; \
				(v).z += ba.z * k; \
			} \
		}


				(v).x -= (n).x * radius; \
				(v).y -= (n).y * radius; \
				(v).z -= (n).z * radius; \
			} \
		}

*/
	// Check if quad "A" splits segment AB.
	rgVector &p = intersect_point;
	if( oy > (1.0f + ry)) {
		// NOTE: Vectors "y" and "ba" are NOT orthogonal.
		p = a + (((y.dot( plane_point_A()) - y.dot( a)) / y.dot( ba)) *
			ba - m_v[4]);
		const rgVector q( p - m_v[4]);
		const rgVector w( q.x * inv_xlen, q.y * inv_ylen, q.z *
			inv_zlen);
		const float px = x.dot( w);
		const float pz = z.dot( w);
		if(	(px >= -rx) && (px <= (1.0f + rx)) &&
			(pz >= -rz) && (pz <= (1.0f + rz))) {
			surface_normal = y;
			surface_point = plane_point_A();
//			SET_INTRUSION_VECTOR( intrusion_vector,
//				surface_normal);
/*
			std::cerr << "a = " << a << "\n";
			std::cerr << "b = " << b << "\n";
			std::cerr << "plane_point_A = " << plane_point_A()
				<< "\n";
			std::cerr << "ba = " << ba << "\n";
			std::cerr << "radius = " << radius << "\n";
			std::cerr << "p = " << p << "\n";
			std::cerr << "intrusion = " << intrusion_vector
				<< "\n";
			exit(0);
*/
			return true;
		}
	}

	// Check if quad "B" splits segment AB.
	if( ox > (1.0f + rx)) {
		// NOTE: Vectors "x" and "ba" are NOT orthogonal.
		p = a + (((x.dot( plane_point_B()) - x.dot( a)) / x.dot( ba)) *
			ba - m_v[4]);
		const rgVector q( p - m_v[4]);
		const rgVector w( q.x * inv_xlen, q.y * inv_ylen, q.z *
			inv_zlen);
		const float py = y.dot( w);
		const float pz = z.dot( w);
		if(	(py >= -ry) && (py <= (1.0f + ry)) &&
			(pz >= -rz) && (pz <= (1.0f + rz))) {
			surface_normal = x;
			surface_point = plane_point_B();
//			SET_INTRUSION_VECTOR( intrusion_vector,
//				surface_normal);
			return true;
		}
	}

	// Check if quad "C" splits segment AB.
	if( oy < -ry) {
		// NOTE: Vectors "y" and "ba" are NOT orthogonal.
		//	Also, using normal = "-y".
		p = a + (((y.dot( plane_point_C()) - y.dot( a)) / y.dot( ba)) *
			ba - m_v[4]);
		const rgVector q( p - m_v[4]);
		const rgVector w( q.x * inv_xlen, q.y * inv_ylen, q.z *
			inv_zlen);
		const float px = x.dot( w);
		const float pz = z.dot( w);
		if(	(px >= -rx) && (px <= (1.0f + rx)) &&
			(pz >= -rz) && (pz <= (1.0f + rz))) {
			surface_normal = -y;
			surface_point = plane_point_C();
//			SET_INTRUSION_VECTOR( intrusion_vector,
//				surface_normal);
			return true;
		}
	}

	// Check if quad "D" splits segment AB.
	if( ox < -rx) {
		// NOTE: Vectors "x" and "ba" are NOT orthogonal.
		//	Also, using normal = "-x".
		p = a + (((x.dot( plane_point_D()) - x.dot( a)) / x.dot( ba)) *
			ba);
		const rgVector q( p - m_v[4]);
		const rgVector w( q.x * inv_xlen, q.y * inv_ylen, q.z *
			inv_zlen);
		const float py = y.dot( w);
		const float pz = z.dot( w);
		std::cout << " x.dot( a) = " << x.dot( a) << "\n";
		std::cout << " x.dot( ba) = " << x.dot( ba) << "\n";
		std::cout << " x.dot( D) = " << x.dot( plane_point_D())<<"\n";
		std::cout << "ba = " << ba << "\n";
		std::cout << "a = " << a << "\n";
		std::cout << "b = " << b << "\n";
		std::cout << "D = " << plane_point_D() << "\n";
		std::cout << "p = " << p << "\n";
		std::cout << "x = " << x << "\n";
		std::cout << "m4 = " << m_v[4] << "\n";
		std::cout << "q = " << q << "\n";
		std::cout << "w = " << w << "\n";
		std::cout << "py = " << py << "\n";
		std::cout << "pz = " << pz << "\n";
		
		if(	(py >= -ry) && (py <= (1.0f + ry)) &&
			(pz >= -rz) && (pz <= (1.0f + rz))) {
			surface_normal = -x;
			surface_point = plane_point_D();
//			SET_INTRUSION_VECTOR( intrusion_vector,
//				surface_normal);
			return true;
		}
	}

	// Check if quad "E" splits segment AB.
	if( oz < -rz) {
		// NOTE: Vectors "z" and "ba" are NOT orthogonal.
		//	Also, using normal = "-z".
		p = a + (((z.dot( plane_point_E()) - z.dot( a)) / z.dot( ba)) *
			ba - m_v[4]);
//		std::cout << "E = " << plane_point_E() << "\n";
//		std::cout << "a = " << a << "\n";
//		std::cout << "b = " << b << "\n";
//		std::cout << "p = " << p << "\n";
//		std::cout << "m4 = " << m_v[4] << "\n";
		const rgVector q( p - m_v[4]);
		const rgVector w( q.x * inv_xlen, q.y * inv_ylen, q.z *
			inv_zlen);
//		std::cout << "q = " << q << "\n";
//		std::cout << "w = " << w << "\n";
		const float px = x.dot( w);
		const float py = y.dot( w);
//		std::cout << "px = " << px << "\n";
//		std::cout << "py = " << py << "\n";
//		exit(0);
		if(	(px >= -rx) && (px <= (1.0f + rx)) &&
			(py >= -ry) && (py <= (1.0f + ry))) {
			surface_normal = -z;
			surface_point = plane_point_E();
//			SET_INTRUSION_VECTOR( intrusion_vector,
//				surface_normal);
			return true;
		}
	}

	// Check if quad "F" splits segment AB.
	if( oz > (1.0f + rz)) {
		// NOTE: Vectors "z" and "ba" are NOT orthogonal.
		p = a + (((z.dot( plane_point_F()) - z.dot( a)) / z.dot( ba)) *
			ba - m_v[4]);
		const rgVector q( p - m_v[4]);
		const rgVector w( q.x * inv_xlen, q.y * inv_ylen, q.z *
			inv_zlen);
		const float px = x.dot( w);
		const float py = y.dot( w);
		if(	(px >= -rx) && (px <= (1.0f + rx)) &&
			(py >= -ry) && (py <= (1.0f + ry))) {
			surface_normal = z;
			surface_point = plane_point_F();
//			SET_INTRUSION_VECTOR( intrusion_vector,
//				surface_normal);
			return true;
		}
	}
	#undef SET_INTRUSION_VECTOR

	// Point "a" is within the prisim (which it shouldn't be). Try to
	// give the best collision data results regardless of the corrupt
	// parameters. The collision surface is the quad that is closest to
	// the point "a".
	std::cout << "Corruption\n";
	int winner = 0;
	{
		const float golf_score[6] = {

			1.0f - (oy + ry),	// Quad A
			1.0f - (ox + rx),	// Quad B
			oy - ry,		// Quad C
			ox - rx,		// Quad D
			oz - rz,		// Quad E
			1.0f - (oz + rz)};	// Quad F

/*
	std::cout << "a = " << a << ", b = " << b << "\n";
	std::cout << "g[0] = " << golf_score[0] << "\n";
	std::cout << "g[2] = " << golf_score[2] << "\n";

	std::cout << "g[1] = " << golf_score[1] << "\n";
	std::cout << "g[3] = " << golf_score[3] << "\n";

	std::cout << "g[4] = " << golf_score[4] << "\n";
	std::cout << "g[5] = " << golf_score[5] << "\n";
*/
//			(1.0f + ry) - oy,	// Quad A
//			(1.0f + rx) - ox,	// Quad B
//			oy + ry,		// Quad C
//			ox + rx,		// Quad D
//			oz + rz,		// Quad E
//			(1.0f + rz) - oz};	// Quad F

		float best_score = golf_score[0];
		for( int i = 1; i < 6; i++) {
			if( golf_score[i] < best_score) {
				best_score = golf_score[i];
				winner = i;
			}
		}
	}
	switch( winner) {
	case 1:	surface_normal = x;
		surface_point = plane_point_B();
		break;

	case 2:	surface_normal = -y;
		surface_point = plane_point_C();
		break;

	case 3:	surface_normal = -x;
		surface_point = plane_point_D();
		break;

	case 4:	surface_normal = -z;
		surface_point = plane_point_E();
		break;

	case 5:	surface_normal = z;
		surface_point = plane_point_F();
		break;

	default:
		surface_normal = y;
		surface_point = plane_point_A();
		break;
	}
	intrusion_vector = rgVector();
	intersect_point = a;
	return true;
}
#endif

rgCollisionShadow rgCollisionPrisim::collision_shadow( const rgVector n) const
{
	return rgCollisionShadow( *this, n);
}

bool rgCollisionPrisim::collision( const rgCollisionPrisim &tar) const
{
	if(	collision( *this, tar, rgVector( 0, 1, 0)) &&
		collision( *this, tar, rgVector( 0, 0, 1))) return true;
	return false;
}

bool rgCollisionPrisim::collision( const rgCollisionPrisim &des,
	const rgCollisionPrisim &tar)
{
	if(	collision( des, tar, rgVector( 0, 1, 0)) &&
		collision( des, tar, rgVector( 0, 0, 1))) return true;
	return false;
}

bool rgCollisionPrisim::collision( const rgCollisionPrisim &dyn,
	const rgCollisionShadow &staty,
	const rgCollisionShadow &statz)
{
	{	const rgCollisionShadow fcs( dyn, rgVector( 0, 1, 0));
		if( !(staty.overlaps( fcs))) return false;
	}
	{	const rgCollisionShadow fcs( dyn, rgVector( 0, 0, 1));
		if( !(statz.overlaps( fcs))) return false;
	}
	return true;
}

bool rgCollisionPrisim::collision( const rgCollisionPrisim &des,
	const rgCollisionPrisim &tar, const rgVector n)
{
	//const rgCollisionShadow ics( *this, n);
	const rgCollisionShadow fcs( des, n);
	const rgCollisionShadow tcs( tar, n);
	//if( tcs.overlaps( ics)) return true;
	if( tcs.overlaps( fcs)) return true;
	//if( tcs.span_overlaps( ics, fcs)) return true;
	return false;
}

bool rgCollisionPrisim::collision( const rgCollisionPlane &tar) const
{
	if( tar.m_normal.dot( m_v[0] - tar.m_point) < 0.0f) return true;
	if( tar.m_normal.dot( m_v[1] - tar.m_point) < 0.0f) return true;
	if( tar.m_normal.dot( m_v[2] - tar.m_point) < 0.0f) return true;
	if( tar.m_normal.dot( m_v[3] - tar.m_point) < 0.0f) return true;
	if( tar.m_normal.dot( m_v[4] - tar.m_point) < 0.0f) return true;
	if( tar.m_normal.dot( m_v[5] - tar.m_point) < 0.0f) return true;
	if( tar.m_normal.dot( m_v[6] - tar.m_point) < 0.0f) return true;
	if( tar.m_normal.dot( m_v[7] - tar.m_point) < 0.0f) return true;
	return false;
}

/*
void rgCollisionPrisim::refresh_normals()
{
	m_n[0] = (quad_A()).m_normal;
	m_n[1] = (quad_B()).m_normal;
	m_n[2] = (quad_C()).m_normal;
	m_n[3] = (quad_D()).m_normal;
	m_n[4] = (quad_E()).m_normal;
	m_n[5] = (quad_F()).m_normal;
}

// XXX: If all faces of "this" prisim contain at least on vertex of collision
//	prisim "cp", then the two collision prisims intersect.
// TODO: This function has the potential to be heavily optimized for speed.
bool rgCollisionPrisim::intersect( const rgCollisionPrisim &cp) const
{
	if(	collision( cp.plane_A()) &&
		collision( cp.plane_B()) &&
		collision( cp.plane_C()) &&
		collision( cp.plane_D()) &&
		collision( cp.plane_E()) &&
		collision( cp.plane_F()) &&

		cp.collision( plane_A()) &&
		cp.collision( plane_B()) &&
		cp.collision( plane_C()) &&
		cp.collision( plane_D()) &&
		cp.collision( plane_E()) &&
		cp.collision( plane_F())) return true;
	return false;
}
*/

void rgCollisionPrisim::bounding_box( rgVector &vmin, rgVector &vmax) const
{
	const float big_num = 16777816.0f;
	vmin = rgVector( big_num, big_num, big_num);
	vmax = rgVector( -big_num, -big_num, -big_num);
	for( int i = 0; i < 8; i++) {
		if( m_v[i].x > vmax.x) vmax.x = m_v[i].x;
		if( m_v[i].x < vmin.x) vmin.x = m_v[i].x;
		if( m_v[i].y > vmax.y) vmax.y = m_v[i].y;
		if( m_v[i].y < vmin.y) vmin.y = m_v[i].y;
		if( m_v[i].z > vmax.z) vmax.z = m_v[i].z;
		if( m_v[i].z < vmin.z) vmin.z = m_v[i].z;
	}
}

rgPhyBoundSphere &rgPhyBoundSphere::gen_containing(
	const std::vector <rgCollisionPrisim> &p)
{
	const float big_num = 16777816.0f;
	rgVector vmin( big_num, big_num, big_num);
	rgVector vmax( -big_num, -big_num, -big_num);
	for( int i = 0; i < int( p.size()); i++) {
		rgVector pmin, pmax;
		(p[i]).bounding_box( pmin, pmax);
		if( pmax.x > vmax.x) vmax.x = pmax.x;
		if( pmin.x < vmin.x) vmin.x = pmin.x;

		if( pmax.y > vmax.y) vmax.y = pmax.y;
		if( pmin.y < vmin.y) vmin.y = pmin.y;

		if( pmax.z > vmax.z) vmax.z = pmax.z;
		if( pmin.z < vmin.z) vmin.z = pmin.z;
	}
	center = (vmin + vmax) * 0.5f;
	radius = (vmax - vmin).length() * 0.5f;
	return *this;
}

rgPhyBoundSphere &rgPhyBoundSphere::gen_containing(
	const std::vector <rgPhyBoundSphere> &s)
{
	const float big_num = 16777816.0f;
	rgVector vmin( big_num, big_num, big_num);
	rgVector vmax( -big_num, -big_num, -big_num);
	for( int i = 0; i < int( s.size()); i++) {
		if( (s[i].center.x + s[i].radius) > vmax.x)
			vmax.x = (s[i].center.x + s[i].radius);
		if( (s[i].center.x - s[i].radius) < vmin.x)
			vmin.x = (s[i].center.x - s[i].radius);

		if( (s[i].center.y + s[i].radius) > vmax.y)
			vmax.y = (s[i].center.y + s[i].radius);
		if( (s[i].center.y - s[i].radius) < vmin.y)
			vmin.y = (s[i].center.y - s[i].radius);

		if( (s[i].center.z + s[i].radius) > vmax.z)
			vmax.z = (s[i].center.z + s[i].radius);
		if( (s[i].center.z - s[i].radius) < vmin.z)
			vmin.z = (s[i].center.z - s[i].radius);
	}
	center = (vmin + vmax) * 0.5f;
	radius = (vmax - vmin).length() * 0.5f;
	return *this;
}


