/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgPort.h"
#include <string.h>
#include <rgStandard.h>

#ifdef UNIX_LIKE_OS
#include <unistd.h>
#endif

#ifdef UNIX_LIKE_OS
char *rgport_getcwd( char *buf, const int len)
{
	if( !buf) return buf;
	if( len <= 0) return 0;
	char *tmp = getcwd( buf, len - 1);
	buf[len - 1] = '\0';
	return tmp;
}
#else
char *rgport_getcwd( char *buf, const int len)
{
	if( !buf) return buf;
	if( len <= 0) return 0;
	// FIXME: Implement this for non Unix systems.
	return 0;
}
#endif

rgWORD rgPortLittleToNativeWORD( rgWORD n)
{
	#ifdef RGPORT_LITTLE_ENDIAN
	return n;	// Native format is little endian
	#endif

	#ifdef RGPORT_BIG_ENDIAN
	return rgPortLittleToBigWORD( n);	// Convert from little to big
	#endif
}

rgDWORD rgPortLittleToNativeDWORD( rgDWORD n)
{
	#ifdef RGPORT_LITTLE_ENDIAN
	return n;	// Native format is little endian
	#endif

	#ifdef RGPORT_BIG_ENDIAN
	return rgPortLittleToBigDWORD( n);	// Convert from little to big
	#endif
}

rgFLOAT rgPortLittleToNativeFLOAT( rgFLOAT n)
{
	#ifdef RGPORT_LITTLE_ENDIAN
	return n;	// Native format is little endian
	#endif

	#ifdef RGPORT_BIG_ENDIAN
	return rgPortLittleToBigFLOAT( n);	// Convert from little to big
	#endif
}

rgWORD rgPortBigToNativeWORD( rgWORD n)
{
	#ifdef RGPORT_BIG_ENDIAN
	return n;	// Native format is big endian
	#endif

	#ifdef RGPORT_LITTLE_ENDIAN
	return rgPortBigToLittleWORD( n);	// Convert from big to little
	#endif
}

rgDWORD rgPortBigToNativeDWORD( rgDWORD n)
{
	#ifdef RGPORT_BIG_ENDIAN
	return n;	// Native format is big endian
	#endif

	#ifdef RGPORT_LITTLE_ENDIAN
	return rgPortBigToLittleDWORD( n);	// Convert from big to little
	#endif
}

rgFLOAT rgPortBigToNativeFLOAT( rgFLOAT n)
{
	#ifdef RGPORT_BIG_ENDIAN
	return n;	// Native format is big endian
	#endif

	#ifdef RGPORT_LITTLE_ENDIAN
	return rgPortBigToLittleFLOAT( n);	// Convert from big to little
	#endif
}


rgWORD rgPortLittleEndianWORD( rgWORD n)
{
	#ifdef RGPORT_LITTLE_ENDIAN
	return n;	// Already little endian
	#endif
	
	#ifdef RGPORT_BIG_ENDIAN
	return rgPortBigToLittleWORD( n);	// Convert from big to little
	#endif
}

rgDWORD rgPortLittleEndianDWORD( rgDWORD n)
{
	#ifdef RGPORT_LITTLE_ENDIAN
	return n;	// Already little endian
	#endif

	#ifdef RGPORT_BIG_ENDIAN
	return rgPortBigToLittleDWORD( n);	// Convert from big to little
	#endif
}

rgFLOAT rgPortLittleEndianFLOAT( rgFLOAT n)
{
	#ifdef RGPORT_LITTLE_ENDIAN
	return n;	// Already little endian
	#endif

	#ifdef RGPORT_BIG_ENDIAN
	return rgPortBigToLittleFLOAT( n);	// Convert from big to little
	#endif
}

rgWORD rgPortBigEndianWORD( rgWORD n)
{
	#ifdef RGPORT_BIG_ENDIAN
	return n;	// Already big endian
	#endif

	#ifdef RGPORT_LITTLE_ENDIAN
	return rgPortLittleToBigWORD( n);	// Convert from little to big
	#endif
}

rgDWORD rgPortBigEndianDWORD( rgDWORD n)
{
	#ifdef RGPORT_BIG_ENDIAN
	return n;	// Already big endian
	#endif

	#ifdef RGPORT_LITTLE_ENDIAN
	return rgPortLittleToBigDWORD( n);	// Convert from little to big
	#endif
}

rgFLOAT rgPortBigEndianFLOAT( rgFLOAT n)
{
	#ifdef RGPORT_BIG_ENDIAN
	return n;	// Already big endian
	#endif

	#ifdef RGPORT_LITTLE_ENDIAN
	return rgPortLittleToBigFLOAT( n);	// Convert from little to big
	#endif
}

bool rgPortEndianCheck()
{
	rgBYTE dword[4] = { 1, 2, 3, 4};

	#ifdef RGPORT_LITTLE_ENDIAN
	/* Little endian machines will swap "dword" to "4321". */
	if( (((rgDWORD *)dword)[0] & 0xFF) != 1) return false;
	return true;
	#endif

	#ifdef RGPORT_BIG_ENDIAN
	/* Big endian machines will keep "dword" as "1234". */
	if( (((rgDWORD *)dword)[0] & 0xFF) != 4) return false;
	return true;
	#endif
}

int rgPortWhichEndian_32()
{
	const rgBYTE n[4] = { 0x1A, 0x2B, 0x3C, 0x4D};
	if( 	(n[0] == 0x1A) &&
		(n[1] == 0x2B) &&
		(n[2] == 0x3C) &&
		(n[3] == 0x4D)) {
		return 0;	/* Big endian */

	} else if(
		(n[0] == 0x4D) &&
		(n[1] == 0x3C) &&
		(n[2] == 0x2B) &&
		(n[3] == 0x1A)) {
		return 1;	/* Little endian */

	} else if(
	 	(n[0] == 0x3C) &&
		(n[1] == 0x4D) &&
		(n[2] == 0x1A) &&
		(n[3] == 0x2B)) {
		/* FIXME: Is this big-little? Which machine used this? */
		return 2;	/* Big-little endian */

	}

	/* Unknown endian */
	return -1;
}

/*
rgDWORD rgPort_Little_to_Native( rgDWORD n)
{
}

rgDWORD rgPort_Big_to_Native( rgDWORD n)
{
	
}
*/

// LITTLE --> BIG
rgBYTE rgPortLittleToBig( rgBYTE n)
{
	return n;
}

rgWORD rgPortLittleToBigWORD( rgWORD n)
{
	return rgPortBigLittleSwapWORD( n);
}

rgDWORD rgPortLittleToBigDWORD( rgDWORD n)
{
	return rgPortBigLittleSwapDWORD( n);
}

rgFLOAT rgPortLittleToBigFLOAT( rgFLOAT n)
{
	return rgPortBigLittleSwapFLOAT( n);
}

// BIG --> LITTLE
rgWORD rgPortBigToLittleWORD( rgWORD n)
{
	return rgPortBigLittleSwapWORD( n);
}

rgDWORD rgPortBigToLittleDWORD( rgDWORD n)
{
	return rgPortBigLittleSwapDWORD( n);
}

rgFLOAT rgPortBigToLittleFLOAT( rgFLOAT n)
{
	return rgPortBigLittleSwapFLOAT( n);
}




rgWORD rgPortBigLittleSwapWORD( rgWORD n)
{
	rgBYTE a[2];
	rgBYTE *b = (rgBYTE *)&n;
	a[0] = b[1];
	a[1] = b[0];
	return *((rgWORD *)a);
}

rgDWORD rgPortBigLittleSwapDWORD( rgDWORD n)
{
	rgBYTE a[4];
	rgBYTE *b = (rgBYTE *)&n;
	a[0] = b[3];
	a[1] = b[2];
	a[2] = b[1];
	a[3] = b[0];
	return *((rgDWORD *)a);
}

rgFLOAT rgPortBigLittleSwapFLOAT( rgFLOAT n)
{
	// Swaping a FLOAT is exactly like swaping a DWORD, just make sure
	// that no rounding is done (just do a perfect binary endian swap).
	rgDWORD *d = (rgDWORD *)&n;
	(*d) = rgPortBigLittleSwapDWORD( *d);
	return n;
}




void rg_get_unix_file_name_only( char *des, const char *src, int max)
{
	if( max <= 0) return;
	if( !des) return;
	if( !src) {
		des[0] = '\0';
		return;
	}

	// Start from the end of "src".
	int i = 0;
	while( src[i]) i++;

	// Backpedal until the beginning of the string is reached, or
	// a slash '/' is found.
	while( (i > 0) && (src[i] != '/')) i--;

	// Copy the resulting string to des.
	if( '/' == src[i]) i++;
	while( (--max) && src[i]) *(des++) = src[i++];
	*des = '\0';
}

void rg_get_unix_path_name_only( char *des, const char *src, int max)
{
	if( max <= 0) return;
	if( !des) return;
	if( !src) {
		des[0] = '\0';
		return;
	}

	// Start from the end of "src".
	int i = 0;
	while( src[i]) i++;

	// Backpedal until the beginning of the string is reached, or
	// a slash '/' is found.
	while( (i > 0) && (src[i] != '/')) i--;

	// Copy the resulting string to des.
	if( '/' == src[i]) i++;
	int j = 0;
	while( (--max) && (j < i)) *(des++) = src[j++];
	*des = '\0';
}

bool rg_slash_terminates_unix_path( const char *path)
{
	if( !path) return false;
	if( !(*path)) return false;
	while( *(path + 1)) path++;
	if( '/' == (*path)) return true;
	return false;
}

std::string rg_get_extension( const std::string &fname)
{
	int i = int( fname.length() - 1);
	for( ; i >= 0; i--) {
		if( '.' == fname[i]) {
			i++;
			break;
		}
	}
	std::string tmp;
	if( i < 0) return tmp;
	for( ; i < int( fname.length()); i++) tmp += fname[i];
	return tmp;
}



int rgport_mkstemp( char *s)
{
	#ifdef UNIX_LIKE_OS
	return mkstemp( s);
	#else
	// FIXME: Implement a non-unix and Windows version of "mkstemp()".
	#endif
}

void rgport_close( int fd)
{
	#ifdef UNIX_LIKE_OS
	close( fd);
	#else
	// FIXME: Implement a non-unix and Windows version of "mkstemp()".
	#endif
}

std::string rgport_file_copy( std::istream &src, FILE *des)
{
	if( !des) return "null FILE pointer passed";
	const rgDWORD buf_size = 64000;
	rgTempArray <char> buf( buf_size);
	if( buf.empty()) return "could not allocate copy buffer";

	while( true) {
		src.read( buf.data, buf_size);
		const rgDWORD nread = src.gcount();
		if( 0 == nread) return std::string();
		if( nread != fwrite( buf.data, 1, nread, des))
			return "failed while writing";
		if( src.eof()) return std::string();
		if( src.fail()) return "unspecified failure";
	}
}
