/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef AI_H
#define AI_H

#include "../things/things.h"

#define DECLARE_AI_TASK( t) \
	class t : public phAI_Task { public: void do_task() { throw this;}}
#define wait_for_next_frame return

#define AIOUT( m) (std::cout << (get_host())->m_key << ": ") << m

#define PH_AI_MAX_MESSAGES	100

class phAI_Task
{
public:
	bool m_first_time;
	int m_frame_delay;
	int m_check_point;
	int m_priority;

	phAI_Task() {
		m_priority = 0;
		m_first_time = true;
		m_frame_delay = 0;
		m_check_point = 0;
	}

	//virtual const char *m_name() const { return "phAI_Task";}

	bool operator >=( phAI_Task *t) {
		if( m_priority >= t->m_priority) return true;
		return false;
	}
	virtual void do_task(){ throw this;}

private:

	phAI_Task( const phAI_Task &);
	phAI_Task &operator =( const phAI_Task &);
};

class phAI_Message
{
public:
	ThingControlBlock *m_sender;
	std::vector <std::string> m_msg;

	phAI_Message( ThingControlBlock *src, const std::string &p_msg) {
		init();
		m_sender = src;
		m_msg.push_back( p_msg);
	}

	phAI_Message(	ThingControlBlock *src,
			const std::string &p_msg1,
			const std::string &p_msg2) {
		init();
		m_sender = src;
		m_msg.push_back( p_msg1);
		m_msg.push_back( p_msg2);
	}

	phAI_Message(	ThingControlBlock *src,
			const std::string &p_msg1,
			const std::string &p_msg2,
			const std::string &p_msg3) {
		init();
		m_sender = src;
		m_msg.push_back( p_msg1);
		m_msg.push_back( p_msg2);
		m_msg.push_back( p_msg3);
	}

	const std::string message( const int i) const {
		if( (i < 0) || (i >= int( m_msg.size()))) return std::string();
		return m_msg[i];
	}

private:
	friend class phAI;
	int m_ref_count;
	int ref_count() { return m_ref_count;}
	void inc_reference() { m_ref_count++;}
	void dec_reference() { m_ref_count--;}

	void init() {
		m_ref_count = 0;
	}

	phAI_Message( const phAI_Message &);
	phAI_Message &operator =( const phAI_Message &);
};

class phAI : public gmThing
{
public:
	RPH_THING_MODULE

	phAI();
	virtual ~phAI();

	std::string type() const { return std::string( "phAI");}
	bool is_a( const std::string &s) const {
		if( s != phAI::type()) return gmThing::is_a( s);
		return true;
	}

	gmTE call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	// Note: If a host is free, then it is dettached.
	ThingControlBlock *get_host() {
		if(	(!m_host) ||
			(m_host->is_free()) ||
			(!accept_host( m_host))) m_host = 0;
		return m_host;
	}

	gmThing *get_host_thing() {
		ThingControlBlock *t = get_host();
		if( !t) return 0;
		return t->m_thing;
	}

	bool set_host( ThingControlBlock *t) {
		if(	!t ||
			(t->is_free()) ||
			(!accept_host( t))) {
			m_host = 0;
			return false;
		}
		m_host = t;
		m_host->m_thing->set_ai( tcb());
		return true;
	}

	int message_count() const { return int( m_message.size());}
	phAI_Message *message( const int mid) {
		if( (mid < 0) || (mid >= message_count())) return 0;
		return (m_message[mid]);
	}

	void send_message( ThingControlBlock *des, phAI_Message *msg);
	void send_message( const std::string &des, phAI_Message *msg);
	void send_message( gmThing *des, phAI_Message *msg) {
		if( !des) return;
		send_message( des->tcb(), msg);
	}

	void send_message( gmThing *des, const std::string &m) {
		if( !des) return;
		send_message( des->tcb(), new phAI_Message( get_host(), m));
	}

	void send_message(
		gmThing *des,
		const std::string &m1,
		const std::string &m2) {
		if( !des) return;
		send_message( des->tcb(),
			new phAI_Message( get_host(), m1, m2));
	}

	void send_message(
		gmThing *des,
		const std::string &m1,
		const std::string &m2,
		const std::string &m3) {
		if( !des) return;
		send_message( des->tcb(),
			new phAI_Message( get_host(), m1, m2, m3));
	}

	void clear_messages() {
		while( !(m_message.empty())) delete_message( 0);
	}
	void delete_message( const int mid) {
		if( (mid < 0) || (mid >= message_count())) return;
		if( m_message[mid]) {
			(m_message[mid])->dec_reference();
			if( (m_message[mid])->ref_count() <= 0)
				delete (m_message[mid]);
		}
		m_message.erase( m_message.begin() + mid);
	}

	bool process_higher_level_AI( ThingManager *tm);

	phAI_Task *current_task() {
		if( m_task.empty()) return 0;
		return (m_task.back());
	}

	void set_sub_task( phAI_Task *t);

	void continue_previous_task() {
		if( m_task.size() <= 1) return;
		if( m_task.back()) delete (m_task.back());
		m_task.erase( --(m_task.end()));
	}

	void add_priority_task( phAI_Task *t); 

	void set_task( phAI_Task *t) {
		if( !t) return;
		if( m_task.empty()) {
			m_task.push_back( t);
			return;
		}
		phAI_Task *&tmp = m_task.back();
		if( tmp) delete tmp;
		tmp = t;
	}

	void clear_task_list() {
		while( !(m_task.empty())) {
			if( m_task.front()) delete (m_task.front());
			m_task.erase( m_task.begin());
		}
	}

	void set_single_task( phAI_Task *t) {
		clear_task_list();
		set_task( t);
	}

protected:

	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;

	// Some AI's only make sense when controlling specific hosts.
	virtual bool accept_host( ThingControlBlock *t) { return true;}

	std::string say( const char *s) {
		gmThing *t = get_host_thing();
		if( !t) return (std::string( "[No host]: ") + s);
		return (((t->tcb())->m_key + ": ") + s);
	}


	//std::vector <phAI_Task *> m_task;
	std::list <phAI_Task *> m_task;

	friend gmTE phAI_set_host( const gmCCArg &p);

private:
	std::vector <phAI_Message *> m_message;

	// What "thing" the AI controls.
	ThingControlBlock *m_host;

	phAI( const phAI &);
	phAI &operator =( const phAI &);
};

#endif
