/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "ai_coach.h"
#include "engine/engine_game.h"
#include <rg3DRom.h>
#include "arena.h"
#include "team.h"
#include "referee.h"
#include "coach.h"
#include "player.h"
#include "puck.h"







std::map <const std::string, THING_COMMAND_FN> phAI_Coach::m_ccmd;

extern "C" gmThing *spw_phAI_Coach()
{
	return new phAI_Coach();
}

extern "C" int del_phAI_Coach( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phAI_Coach)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phAI_Coach",
		spw_phAI_Coach);

	p_ge->m_thing_manager->add_to_delete_list( "phAI_Coach",
		del_phAI_Coach);

	return 0;
}

phAI_Coach::phAI_Coach()
{
	if( !(m_ccmd.empty())) return;
}

phAI_Coach::~phAI_Coach()
{
}

gmTE phAI_Coach::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phAI::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phAI_Coach::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phAI_Coach::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += phAI::list_ccmd( p_prefix);
	return tmp;
}


void phAI_Coach::message_players( const std::string &grp, phTeam *p_team,
	phAI_Message *p_msg)
{
	if( !p_team) return;
	std::vector <phPerson *> tmp;
	p_team->get_people( tmp, grp);
	message_players( tmp, p_msg);
}

void phAI_Coach::message_players( const std::vector <phPerson *> &p,
	phAI_Message *p_msg)
{
	if( !p_msg) return;
	for( int i = 0; i < int( p.size()); i++)
		send_message( p[i], p_msg);
}

void phAI_Coach::bench_everyone( phTeam *t)
{
	// Bench everyone.
	const char *bench_command = "BENCH!!!";
	send_message( t->get_position( "C1"), bench_command);
	send_message( t->get_position( "C2"), bench_command);
	send_message( t->get_position( "C3"), bench_command);
	send_message( t->get_position( "C4"), bench_command);
	send_message( t->get_position( "LW1"), bench_command);
	send_message( t->get_position( "LW2"), bench_command);
	send_message( t->get_position( "LW3"), bench_command);
	send_message( t->get_position( "LW4"), bench_command);
	send_message( t->get_position( "RW1"), bench_command);
	send_message( t->get_position( "RW2"), bench_command);
	send_message( t->get_position( "RW3"), bench_command);
	send_message( t->get_position( "RW4"), bench_command);
	send_message( t->get_position( "LD1"), bench_command);
	send_message( t->get_position( "LD2"), bench_command);
	send_message( t->get_position( "LD3"), bench_command);
	send_message( t->get_position( "RD1"), bench_command);
	send_message( t->get_position( "RD2"), bench_command);
	send_message( t->get_position( "RD3"), bench_command);
	send_message( t->get_position( "E1"), bench_command);
	send_message( t->get_position( "E2"), bench_command);
	send_message( t->get_position( "G1"), bench_command);
	send_message( t->get_position( "G2"), bench_command);
}


// XXX: Tasks that the coach AI can perform.
DECLARE_AI_TASK( task___prepare_for_game);
DECLARE_AI_TASK( task___call_faceoff);
DECLARE_AI_TASK( task___ice_players);
DECLARE_AI_TASK( task___watch_game);

void phAI_Coach::process( ThingManager *tm)
{
	//std::cout << "Coach \"" + (tcb())->m_key + "\" is thinking\n";
	// Make sure there is a host coach to think for.
	phCoach *coa;
	{
		ThingControlBlock *h = get_host();
		if( !h) return;
		coa = (phCoach *)(h->m_thing);
	}

	// Get the arena and accessories
        phArena *t_arena = coa->arena();
        if( !t_arena) return;
        phPuck *t_puck = t_arena->game_puck();
        if( !t_puck) return;

	// Get the teams
	phTeam *us = coa->team();
	phTeam *them = t_arena->home_team();
	if( us == them) them = t_arena->away_team();
	if( (!us) || (!them)) return;

	// Get referee
	//phReferee *ref = t_arena->primary_referee();
	//if( !ref) return;

	// XXX: Check messages.
	while( message_count() > 0) {
		//AIOUT( "I got a message\n");
		phAI_Message *m = message( 0);
		if( "primary_ref" == m->m_sender->m_key) {
			const std::string msg( m->message( 0));
			if( "faceoff" == msg) {
				// XXX: Adding a priority task is very
				//	dangerous if the AI (Artificial
				//	Intelligence) is not robust. Setting a
				//	single task is safe, but limits the
				//	"intelligence" of the AI.
				add_priority_task( new task___call_faceoff);
				//set_single_task( new task___call_faceoff);

			} else if( "prepare_for_game" == msg) {
				// Drop everything and prepare for the game.
				set_single_task( new task___prepare_for_game);
			}
		}
		delete_message( 0);
	}

	// XXX:
	while( !(m_task.empty())) {
		try { (current_task())->do_task();}

		
		// Preperation for the game is done in one frame to prevent
		// the task from being interrupted.
		catch( task___prepare_for_game *)
		{
			us->create_person_list( "ice");

			// TODO: Add more creative names for special lines.
			//	However, the single/double letter lines are
			//	standard.

			// Create standard people lists.
			us->create_person_list( "F1");
			us->create_person_list( "F2");
			us->create_person_list( "F3");
			us->create_person_list( "F4");
			us->create_person_list( "D1");
			us->create_person_list( "D2");
			us->create_person_list( "D3");
			us->create_person_list( "E1");
			us->create_person_list( "E2");
			us->create_person_list( "G1");
			us->create_person_list( "G2");
			us->create_person_list( "PP1");
			us->create_person_list( "PP2");
			us->create_person_list( "PK1");
			us->create_person_list( "PK2");

			// Populate the lists.
			us->add_person( "F1", us->get_position_name( "C1"));
			us->add_person( "F1", us->get_position_name( "LW1"));
			us->add_person( "F1", us->get_position_name( "RW1"));

			{
				std::vector <std::string> tmp;
				tmp.push_back( "F1");
				//tmp.push_back( "D1");
				//tmp.push_back( "G1");
				us->schedule_lines( tmp);
			}

			bench_everyone( us);

			AIOUT( "My team is ready to play.\n");
			clear_task_list();
			wait_for_next_frame;
		}



		// Description: Makes sure appropriate players are on the
		//	ice, and tells them all to setup for the faceoff.
		catch( task___call_faceoff *t)
		{
			// Tell sheduled players to get on the ice, and
			// everyone else to get off.
			if( t->m_first_time) {
				set_sub_task( new task___ice_players);
				t->m_first_time = false;
				wait_for_next_frame;
			}

			//message_players( "ice", us,
			//	new phAI_Message( coa->tcb(),
			//		"faceoff"));

			// Tell the scheduled line (which may or may not be
			// on the ice) to setup for the faceoff.
			std::vector <phPerson *> sline;
			us->get_people( sline, us->get_sched_lines());
			message_players( sline,
				new phAI_Message( coa->tcb(), "faceoff"));

			AIOUT( "I told my team to setup for the "
				"faceoff.\n");
			set_single_task( new task___watch_game);
			wait_for_next_frame;
		}



		// Description: Ensures that the maximum number of allowed
		//	players are on the ice.
		//		If too many players are on the ice, then all
		//	players on the ice will be benched except for the
		//	players who's line is currently scheduled to be on the
		//	ice.
		//		If too few players are on the ice, then those
		//	players who's line is scheduled to be on the ice are
		//	told to get off the bench, players from other lines
		//	may also be used to fill in the available positions.
		//
		catch( task___ice_players *t)
		{
			std::vector <phPerson *> iced;
			us->get_people( iced, "ice");
			
			if( int( iced.size()) != us->max_iced_players()) {
				// First tell everyone to get on the bench.
				message_players( iced,
					new phAI_Message( coa->tcb(),
						"bench"));

				// Now tell the scheduled lines to get on the
				// ice.
				std::vector <phPerson *> sline;
				us->get_people( sline, us->get_sched_lines());
				message_players( sline,
					new phAI_Message( coa->tcb(),
						"ice"));

				{
					std::stringstream tmp;
					us->list_people( tmp, sline);
					AIOUT( "Icing line : " << tmp.str()
						<< "\n");
				}
			}
			continue_previous_task();
			wait_for_next_frame;
		}



		// Description: Watches the game, ocassionally checking the
		//	offensive and defensive state of the team. As well
		//	checking to see if players are following the game
		//	plan. And finally, monitors the other team.
		catch( task___watch_game *)
		{
			wait_for_next_frame;
		}



		catch( phAI_Task *) { wait_for_next_frame;}
	}

}
