/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "ai_forward.h"
#include "engine/engine_game.h"
#include <rg3DRom.h>
#include "arena.h"
#include "team.h"
#include "player.h"
#include "puck.h"
#include "referee.h"
#include "coach.h"

std::map <const std::string, THING_COMMAND_FN> phAI_Forward::m_ccmd;

extern "C" gmThing *spw_phAI_Forward()
{
	return new phAI_Forward();
}

extern "C" int del_phAI_Forward( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phAI_Forward)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phAI_Forward",
		spw_phAI_Forward);

	p_ge->m_thing_manager->add_to_delete_list( "phAI_Forward",
		del_phAI_Forward);

	return 0;
}

phAI_Forward::phAI_Forward()
{

	if( !(m_ccmd.empty())) return;
}

phAI_Forward::~phAI_Forward()
{
}

gmTE phAI_Forward::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phAI_Player::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phAI_Forward::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phAI_Forward::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += phAI_Player::list_ccmd( p_prefix);
	return tmp;
}


// XXX: Tasks that the forward AI can perform.
DECLARE_AI_TASK( task___idle_at_bench);
DECLARE_AI_TASK( task___get_puck);
DECLARE_AI_TASK( task___get_into_scoring_position);
DECLARE_AI_TASK( task___try_to_score);
DECLARE_AI_TASK( task___setup_for_faceoff);

/*
class task___prepare_to_drop_puck : public phAI_Task
{
public:
	bool is_ready;
	int m_stand_frame_delay;

	task___prepare_to_drop_puck() { is_ready = false;}

	void do_task() { throw this;}
};
*/

void phAI_Forward::process_coach_message( phAI_Message *m, phReferee *ref,
	phCoach *coa, phTeam *us, phTeam *them)
{
	//AIOUT( m->message(0) << "\n");
	if( "faceoff" == m->message( 0)) {
		set_single_task( new task___setup_for_faceoff);
	}
}

void phAI_Forward::process( ThingManager *tm)
{
	//process_higher_level_AI( tm);

	// Make sure there is a host player to think for.
	phPlayer *ply;
	{
		ThingControlBlock *h = get_host();
		if( !h) return;
		ply = (phPlayer *)(h->m_thing);
	}

	// Get the arena and accessories
	phArena *t_arena = ply->arena();
	if( !t_arena) return;
	phPuck *t_puck = t_arena->game_puck();
	if( !t_puck) return;

	// Get the teams
	phTeam *us = ply->team();
	phTeam *them = t_arena->home_team();
	if( us == them) them = t_arena->away_team();
	if( (!us) || (!them)) return;

	// Get the coach and referee
	phCoach *coa = us->get_coach();
	phReferee *ref = t_arena->primary_referee();
	if( (!coa) || (!ref)) return;

	check_messages( ref, coa, us, them);

	// XXX:
	//if( m_task.empty()) set_task( new task___idle_at_bench);
	if( m_task.empty()) {
		set_task( new task___idle_at_bench);
		set_sub_task( new task___skate_to(
			t_puck,
			0.5f,
			0.1f,
			250,
			"fast_skate",
			"medium"));
	}

	while( !(m_task.empty())) {
		try { (current_task())->do_task();}


		catch( task___idle_at_bench *t)
		{
			if( !(ply->near_game_puck( 0.75)))
				wait_for_next_frame;
			ply->bind_model_to_box( "inside_blade", t_puck);
			wait_for_next_frame;
		}




		catch( task___skate_to *t)
                {
			if( skate_to( t) != "done")
				wait_for_next_frame;
			continue_previous_task();
		}


		catch( task___skate_heading *t)
		{
			skate_heading( t);
			wait_for_next_frame;
		}

	
		catch( task___get_puck *)
		{
			if( !(ply->near_game_puck( 0.75))) {
				set_sub_task( new task___skate_to(
					t_puck,
					0.5f,
					0.1f,
					250,
					"fast_skate",
					"medium"));
			} else {
				set_sub_task( new task___ply_get_puck_control);
			}
			wait_for_next_frame;
		}


		catch( task___ply_get_puck_control *)
		{
			// FIXME: Player should make an effort to move to the
			//	puck (if the puck is "slightly" out of reach.
			//	Also, player should put up a quick fight if the
			//	puck is controlled by a nearby player).
			if( !(ply->near_game_puck( 1.5))) {
				continue_previous_task();
				wait_for_next_frame;
			}
			ply->bind_model_to_box( "inside_blade", t_puck);
			wait_for_next_frame;
		}


		catch( task___setup_for_faceoff *t)
		{
			if( t->m_first_time) {
				AIOUT( "I am going to setup for the faceoff.\n"
					);
				us->add_person( "ice", ply->tcb()); 
				t->m_first_time = false;
			}
			wait_for_next_frame;
		}


		catch( task___ply_ready_stand *t)
		{
			if( t->m_first_time) {
				ply->animate( "ready_stand", -1, 200, 0);
				t->m_first_time = false;
			}
			wait_for_next_frame;
		}


		catch( task___ply_shoot *t)
		{
			// FIXME: Move puck release into "shoot" member
			//	function.
			if( t->m_first_time && (!t->m_windup)) {
				// XXX: Note that the player releases the puck
				//	they have (which may or may not be the
				//	game puck "t_puck").
				phPuck *p = ply->has_game_puck();
				if( !p) {
					continue_previous_task();
					wait_for_next_frame;
				}
				gmModelBinder *b = ply->get_binding( p->tcb());
				if( b) b->m_bind_type = "yaxis";
				ply->release_binding( p);

				// FIXME: Use players shot power and other
				// attributes to direct the puck. Also, should
				// direct the puck towards the net.
				const float hz =
					float( (ge())->game_clock_hz());
				p->m_velocity = t->m_heading * (22.0f / hz);
				p->m_velocity.y = (5.0f / hz);

				t->m_first_time = false;
			}

			if( "done" == shoot( ply, t))
				continue_previous_task();
			wait_for_next_frame;
		}


		catch( phAI_Task *) { wait_for_next_frame;}
	}
}
