/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "ai_player.h"
#include "engine/engine_game.h"
#include <rg3DRom.h>
#include "arena.h"
#include "team.h"
#include "player.h"
#include "puck.h"
#include "referee.h"
#include "coach.h"
#include "user_control.h"

std::map <const std::string, THING_COMMAND_FN> phAI_Player::m_ccmd;

extern "C" gmThing *spw_phAI_Player()
{
	return new phAI_Player();
}

extern "C" int del_phAI_Player( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phAI_Player)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phAI_Player",
		spw_phAI_Player);

	p_ge->m_thing_manager->add_to_delete_list( "phAI_Player",
		del_phAI_Player);

	return 0;
}

phAI_Player::phAI_Player()
{

	if( !(m_ccmd.empty())) return;
}

phAI_Player::~phAI_Player()
{
}

gmTE phAI_Player::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phAI_Skater::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phAI_Player::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phAI_Player::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += phAI_Skater::list_ccmd( p_prefix);
	return tmp;
}

std::string phAI_Player::shoot( phPlayer *ply, task___ply_shoot *t)
{
	ply->animate( t->m_anim, -1, 25, 0);
	if( t->m_windup) return std::string();
	if( ply->seq_stopped()) {
		if( t->m_call_back) t->m_call_back( t);
		return "done";
	}
	return std::string();
}

// XXX: Tasks that the forward AI can perform.
//DECLARE_AI_TASK( task___idle_at_bench);

/*
class task___prepare_to_drop_puck : public phAI_Task
{
public:
	bool is_ready;
	int m_stand_frame_delay;

	task___prepare_to_drop_puck() { is_ready = false;}

	void do_task() { throw this;}
};
*/


void phAI_Player::check_messages( phReferee *ref, phCoach *coa, phTeam *us,
	phTeam *them)
{
	// XXX: Check messages.
	while( message_count() > 0) {
		phAI_Message *m = message( 0);
		if( !(m->m_sender->is_free())) {
			if( m->m_sender == (coa->tcb())) {
				process_coach_message( m, ref, coa, us, them);

			} else if( m->m_sender == (ref->tcb())) {
				process_referee_message( m, ref, coa, us,
					them);

			} else if( m->m_sender->m_thing->is_a( "phPerson")) {
				phPerson *p =
					(phPerson *)(m->m_sender->m_thing);
				if( p->team() == us) {
					process_teammate_message( m, ref, coa,
						us, them);
				} else {
					process_opponent_message( m, ref, coa,
						us, them);
				}
			}
		}
		delete_message( 0);
	}
}

// XXX: "phAI_Player" is just a generic player AI designed to provide
//	a place for AI common to forwards, defense, goalies, etc. Thus
//	nothing is done in this function since it is specific to the
//	player type.
/*
void phAI_Player::process( ThingManager *tm)
{

	// Make sure there is a host player to think for.
	phPlayer *ply;
	{
		ThingControlBlock *h = get_host();
		if( !h) return;
		if( !(h->m_thing->is_a( "phPlayer"))) return;
		ply = (phPlayer *)(h->m_thing);
	}

	// Get the arena and accessories
	phArena *t_arena = ply->arena();
	if( !t_arena) return;
	phPuck *t_puck = t_arena->game_puck();
	if( !t_puck) return;

	// Get the teams
	phTeam *us = ply->team();
	phTeam *them = t_arena->home_team();
	if( us == them) them = t_arena->away_team();
	if( (!us) || (!them)) return;

	// Get the coach and referee
	phCoach *coa = us->get_coach();
	phReferee *ref = t_arena->primary_referee();
	if( (!coa) || (!ref)) return;

	// XXX:
	if( m_task.empty()) set_task( new task___idle_at_bench);

	while( !(m_task.empty())) {
		try { (m_task[ m_task.size() - 1])->do_task();}



		catch( task___idle_at_bench *t)
		{
			wait_for_next_frame;
		}

		catch( phAI_Task *) { wait_for_next_frame;}
	}
}
*/
