/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "ai_referee.h"
#include "ai_skater.h"
#include "engine/engine_game.h"
#include <rg3DRom.h>
#include "arena.h"
#include "player.h"
#include "puck.h"
#include "referee.h"
#include "coach.h"
#include "team.h"

std::map <const std::string, THING_COMMAND_FN> phAI_Referee::m_ccmd;

extern "C" gmThing *spw_phAI_Referee()
{
	return new phAI_Referee();
}

extern "C" int del_phAI_Referee( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phAI_Referee)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phAI_Referee",
		spw_phAI_Referee);

	p_ge->m_thing_manager->add_to_delete_list( "phAI_Referee",
		del_phAI_Referee);

	return 0;
}

phAI_Referee::phAI_Referee()
{
	m_game_started = false;
	m_play_stopped = true;

	if( !(m_ccmd.empty())) return;
}

phAI_Referee::~phAI_Referee()
{
}

gmTE phAI_Referee::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phAI_Skater::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phAI_Referee::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phAI_Referee::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += phAI_Skater::list_ccmd( p_prefix);
	return tmp;
}


// XXX: Tasks that the referee AI can perform.
DECLARE_AI_TASK( task___start_the_game);
DECLARE_AI_TASK( task___skate_to_puck);
DECLARE_AI_TASK( task___pickup_puck);
DECLARE_AI_TASK( task___goto_faceoff_dot);
//DECLARE_AI_TASK( task___prepare_to_drop_puck);
DECLARE_AI_TASK( task___drop_the_puck);
DECLARE_AI_TASK( task___watch_play);

class task___prepare_to_drop_puck : public phAI_Task
{
public:
	int m_faceoff_wait;
	task___prepare_to_drop_puck(){ m_faceoff_wait = 0;}
	void do_task() { throw this;}
};

void phAI_Referee::process( ThingManager *tm)
{
	// Make sure there is a host referee to think for.
	phReferee *ref;
	{
		ThingControlBlock *h = get_host();
		if( !h) return;
		ref = (phReferee *)(h->m_thing);
	}

	phArena *t_arena = ref->arena();
	if( !t_arena) return;
	phPuck *t_puck = t_arena->game_puck();
	if( !t_puck) return;

	// Get the teams
	phTeam *t_home = t_arena->home_team();
	phTeam *t_away = t_arena->away_team();
	if( !t_home) return;
	if( !t_away) return;

	// Get the coachs
	phCoach *t_home_coach = t_home->get_coach();
	phCoach *t_away_coach = t_away->get_coach();
	if( !t_home_coach) return;
	if( !t_away_coach) return;
	

	// FIXME: Testing, make puck stick to player "moja".
/*	ThingControlBlock *t = (ge())->m_thing_manager->get("moja");
	if( t && (!t->is_free())) {
		t_puck->m_use_gravity = false;
		(t_puck->draw_context())->translate(
			(t->m_thing->cs()).p - (t_puck->cs()).p);
	}
*/

	// FIXME: Testing, trying to make draw context matrix no longer be
	//	orthogonal.
	//(ref->draw_context())->rotate( RG_DEGTORAD( 1), rgVector( 0, 1, 0));

	// XXX: Check messages.
	while( message_count() > 0) {
		delete_message( 0);
	}

	// XXX:
	if( m_task.empty()) {
		// FIXME: Temporary. Wait 2 seconds before doing anything
		//	so I can see the referee skate to the face of dot.
		phAI_Task *t = new task___start_the_game;
		t->m_frame_delay = 2 * (ge())->game_clock_hz();
		set_task( t);
	}

	while( !(m_task.empty())) {
		try { (current_task())->do_task();}


		catch( task___start_the_game *t)
		{
			if( (t->m_frame_delay)-- > 0)
				wait_for_next_frame;

			if( m_game_started) {
				continue_previous_task();
				wait_for_next_frame;
			}

			// Tell coaches to get ready for the game.
			send_message( t_home_coach, "prepare_for_game");
			send_message( t_away_coach, "prepare_for_game");

			m_play_stopped = true;
			set_task( new task___goto_faceoff_dot);
			set_sub_task( new task___pickup_puck);
		}


		catch( task___skate_to_puck *t)
		{
			if( t->m_first_time) {
				set_sub_task( new task___skate_to(
					t_puck,
					0.0f,
					0.08f,
					350,
					"casual_skate",
					"casual"));
				t->m_first_time = false;
				wait_for_next_frame;
			}
			continue_previous_task();
		}


		catch( task___pickup_puck *)
		{
			if( ref->near_game_puck( 1.0f)) {
				if( ref->cur_seq_is( "lift_puck_off_ice")) {
					if( ref->seq_stopped())
						continue_previous_task();
					wait_for_next_frame;
				}

				ref->animate( "bend_to_get_puck", -1, 1, 0);
				if( ref->seq_stopped()) {
					ref->animate( "lift_puck_off_ice", -1,
						1, 0);
					ref->bind_model_to_box( "right_hand",
						t_puck);
					//t_puck->m_use_gravity = false;
					//t_puck->m_velocity = rgVector();
					//(t_puck->draw_context())->
					//	translate(
					//	rgVector( 0, 1, 0));
				}
				wait_for_next_frame;
			}

			if( 	ref->cur_seq_is( "right_pickup_puck") &&
				!(ref->seq_stopped()))
				wait_for_next_frame;
			set_sub_task( new task___skate_to_puck);
		}


		catch( task___goto_faceoff_dot *t)
		{
			// Go to faceoff point first.
			if( t->m_first_time) {
				set_sub_task( new task___skate_to(
					rgVector( -1.0f, 0, 0), // Center ice
					0.1f,
					0.08f,
					350,
					"casual_skate",
					"casual"));
				t->m_first_time = false;
				wait_for_next_frame;
			}

			ref->animate( "stand", -1, 1, 0);
			if( turn_to_face( rgVector( 100, 0, 0), "casual") !=
				"done")
				wait_for_next_frame;
		
			ref->animate( "raise_right_arm", -1, 500, 0);
			phAI_Task *t = new task___prepare_to_drop_puck;
			t->m_frame_delay = 3 * (ge())->game_clock_hz();
			set_single_task( t);
		}



		catch( task___prepare_to_drop_puck *t)
		{
			// Tell coachs to get their teams ready for the
			// faceoff (only tell them once).
			if( t->m_first_time) {
				send_message( t_home_coach, "faceoff");
				send_message( t_away_coach, "faceoff");

				t->m_faceoff_wait = ref->faceoff_wait();
				t->m_first_time = false;
				AIOUT( "I told the coaches to prepare for "
					"the faceoff\n");
			}

			// TODO: Wait for players to be ready, then drop puck.
			// NOTE: If we are waiting forever, then both coaches
			//	must signal the referee when to drop the puck.
			if( t->m_faceoff_wait < 0) wait_for_next_frame;

			// Get ready to drop the puck when there is 5 seconds
			// or less remaining.
			if( (t->m_faceoff_wait)-- <= (ge())->sec_to_frames( 3)
				&& (0 == t->m_check_point))
			{
				ref->animate( "faceoff_stance", -1, 400, 0);
				AIOUT( "3 seconds till I drop the puck.\n");
				t->m_check_point++;
			}

			// If I waited long enough, then drop the puck.
			if( (t->m_faceoff_wait) > 0) wait_for_next_frame;
			AIOUT( "I'm dropping the puck, game on!\n");

			ref->animate( "faceoff_drop_ready", -1, 1, 0);
			set_task( new task___drop_the_puck);
			wait_for_next_frame;
		}

		catch( task___drop_the_puck *t)
		{
			if( ref->cur_seq_is( "faceoff_drop_ready")) {
				if( !(ref->seq_stopped())) wait_for_next_frame;
				ref->animate( "faceoff_drop_puck", -1, 1, 0);
				wait_for_next_frame;
			}

			if( !(ref->seq_stopped())) wait_for_next_frame;
			ref->release_binding( t_puck);

			// Start watching play, but first, backup to the
			// penalty box to get out of the players way.
			set_task( new task___watch_play);
			set_sub_task( new task___skate_to(
				rgVector( -15.0f, 0.0f, 0.0f),
				0.5f,
				0.1f,
				250,
				"fast_skate",
				"medium"));
		}

		catch( task___skate_to *t)
		{
			// TODO: Add support for skating to a "thing".
			if( skate_to( t) != "done")
				wait_for_next_frame;
			continue_previous_task();
		}

		catch( task___watch_play *t)
		{
			// TODO: Start skating around inorder to get a good
			// view of the game. Currently, the referee will just
			// stand in one place and face center ice.

			if( t->m_first_time) {
				AIOUT( "I am watching play.\n");
				ref->animate( "active_stand", -1, 1, 0);
				t->m_first_time = false;
			}

			// Make sure to face the players, not the fans.
			if( turn_to_face( rgVector( 100, 0, 0), "casual") !=
				"done")
				wait_for_next_frame;

			wait_for_next_frame;
		}


		catch( phAI_Task *) { wait_for_next_frame;}
	}
}
