/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "ai_skater.h"
#include "player.h"
#include "engine/engine_game.h"
#include <rg3DRom.h>

std::map <const std::string, THING_COMMAND_FN> phAI_Skater::m_ccmd;

extern "C" gmThing *spw_phAI_Skater()
{
	return new phAI_Skater();
}

extern "C" int del_phAI_Skater( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phAI_Skater)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phAI_Skater",
		spw_phAI_Skater);

	p_ge->m_thing_manager->add_to_delete_list( "phAI_Skater",
		del_phAI_Skater);

	return 0;
}

phAI_Skater::phAI_Skater()
{

	if( !(m_ccmd.empty())) return;
}

phAI_Skater::~phAI_Skater()
{
}

gmTE phAI_Skater::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phAI::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phAI_Skater::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phAI_Skater::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += phAI::list_ccmd( p_prefix);
	return tmp;
}

std::string phAI_Skater::skate_to( task___skate_to *p_task)
{
	phPerson *per = (phPerson *)(get_host_thing());
	if( !per) return "done";

	rgVector p( p_task->m_point);
	if( p_task->m_tcb && !(p_task->m_tcb->is_free()))
		p = (p_task->m_tcb->m_thing->cs()).p;
	
	const rgVector up( 0, 1, 0);
	rgVector dir( rgMakePlanar( p, up) - rgMakePlanar( (per->cs()).p, up));
	const float dis_length = dir.length();
	dir.normalize();

	if( dis_length <= p_task->m_radius) {
		per->m_velocity = rgVector();
		return "done";
	}


	turn_to_face( p, p_task->m_intensity);
	{
		const rgVector v( dir * p_task->m_speed);
		per->m_velocity = rgVector( v.x, per->m_velocity.y, v.z);
	}
	per->animate( p_task->m_anim, -1, p_task->m_anim_delay, 0);

	if( dis_length <= p_task->m_speed) {
		per->m_velocity = rgVector();
		return "done";
	}
	return std::string();
}

std::string phAI_Skater::skate_heading( task___skate_heading *p_task)
{
	phPerson *per = (phPerson *)(get_host_thing());
	if( !per) return "no host";

	if( p_task->m_heading.length_squared() < 0.81f) return "bad heading";
	const rgCoordSystem c( per->cs());
	const rgVector &h = c.eye;

	rgVector &v = per->m_velocity;
	const float y = v.y;
	v.y = 0.0f;

	// FIXME: Get speed from player attributes.
	float pace = 1.0f;
	const float hz = float( (ge())->game_clock_hz());
	float max_speed = 7.0f / hz;
//	if( "med" == p_task->m_intensity)
	if( "fast" == per->m_datt_speed) {
		max_speed = 12.0f / hz;
		pace = 2.0f;
	}

	// FIXME: Get accel from player attributes.
	const float accel_speed = 0.25f / hz;
	rgVector accel( h * accel_speed);

	// FIXME: For 3rd parameter of "animate" (transition time), should use
	//	skaters accel speed to determine transition speed.
	//const rgVector u( v * (1.0f / flen));
	const float ilen = v.length();
	{
		const rgVector &u = p_task->m_heading;
		const rgVector w( h.cross( rgVector( 0, 1, 0)));
		float ang = rgDotAngle( h, u);
		const std::string *ani_c_cut;
		const std::string *ani_stop;
		const float t = w.dot( u);
		if( t > 0.0f) {
			ani_c_cut = &(p_task->m_right_cross_cut);
			ani_stop = &(p_task->m_right_stop);

		} else {
			ani_c_cut = &(p_task->m_left_cross_cut);
			ani_stop = &(p_task->m_left_stop);
		}

		bool can_turn = true;
		if( ang > RG_DEGTORAD( 10.0f)) {
			if( ang > RG_DEGTORAD( 80.0f)) {
				if( ilen > (2.0f / hz)) {
					per->animate( *ani_stop, -1, 150, 0);
					accel = v * -0.0625f;
					can_turn = false;
				} else {
				//per->animate( *ani_c_cut, -1, 150, 0);
					per->animate( p_task->m_anim, -1, 150,
						0);
					accel = rgVector();
				}

			} else {
				per->animate( *ani_c_cut, -1, 150, 0);
				if(	(ang > RG_DEGTORAD( 35)) &&
					(ilen < (3.0f / hz)))
					accel = p_task->m_heading * 0.25f *
						accel_speed;
			}
		} else per->animate( p_task->m_anim, -1, 150, 0, pace);
	
		if( can_turn) turn_to_heading(
			p_task->m_heading, p_task->m_intensity);
	}


	// Accelerate in the "heading" direction. Also, prevent the skater
	// from exceeding their maximum speed (by skating). If the skater is
	// already moving beyond their maximum speed (and skating in the
	// "heading" direction does not slow them to below maximum speed) then
	// the skater will continue to move at maximum speed. In short, it
	// means that other things (such as collisions) can move skaters to
	// beyond maximum speed. Skaters simply cannot move beyond their
	// maximum speed by skating of their own accord.
	v += accel;
	float flen = v.length();
	if( (flen > max_speed) && (ilen < flen)) {
		v = v * (ilen / flen);
		flen = ilen;
	}

	v.y = y;

	return std::string();
}

std::string phAI_Skater::turn_to_heading( const rgVector &dir,
	const std::string &p_intensity)
{
	phPerson *per = (phPerson *)(get_host_thing());
	if( !per) return "done";
	const rgVector up( 0, 1, 0);

	rgCoordSystem c( per->cs());
	c.eye = (rgMakePlanar( c.eye, up)).normalize();
	float ang = rgDotAngle( dir, c.eye);
	if( ang < RG_DEGTORAD( 1.0f)) return "done";
	if( (dir.cross( c.eye)).y > 0.0f) ang = -ang;

	if( "casual" == p_intensity) {
		(per->draw_context())->rotate(
			ang / float( (ge())->game_clock_hz()) * 4.0f, up);

	} else {
		(per->draw_context())->rotate(
			ang / float( (ge())->game_clock_hz()) * 6.0f, up);

	}
	return std::string();
}

std::string phAI_Skater::turn_to_face( const rgVector &p,
	const std::string &p_intensity)
{
	phPerson *per = (phPerson *)(get_host_thing());
	if( !per) return "done";
	const rgVector up( 0, 1, 0);

	rgVector dir( per->direction_to( p, up));
	const float len = dir.length_squared();
	if( len < 0.00001f) return "done";
	dir = dir * (1.0f / sqrt( len));
	
	//const rgMatrix m( (per->draw_context())->mat());
	//rgVector eye( m * rgVector( 0, 0, 1));
	rgCoordSystem c( per->cs());
	c.eye = (rgMakePlanar( c.eye, up)).normalize();
	float ang = rgDotAngle( dir, c.eye);
	//eye = (rgMakePlanar( eye, up)).normalize();
	//float ang = rgDotAngle( dir, eye);
	if( ang < RG_DEGTORAD( 1.0f)) return "done";
	if( (dir.cross( c.eye)).y > 0.0f) ang = -ang;

	if( "casual" == p_intensity) {
		(per->draw_context())->rotate(
			ang / float( (ge())->game_clock_hz()) * 4.0f, up);

	} else {
		(per->draw_context())->rotate(
			ang / float( (ge())->game_clock_hz()) * 6.0f, up);

	}
	return std::string();
}

std::string phAI_Skater::turn_to_face( gmThing *t,
	const std::string &p_intensity)
{
	if( !t) return "done";
	return turn_to_face( (t->cs()).p, p_intensity);
}

std::string phAI_Skater::turn_to_face( ThingControlBlock *t,
	const std::string &p_intensity)
{
	if( (!t) || (t->is_free())) return "done";
	return turn_to_face( t->m_thing, p_intensity);
}

void phAI_Skater::process( ThingManager *tm)
{
	// Make sure there is host person to think for.
	phPerson *p;
	{
		ThingControlBlock *h = get_host();
		if( !h) return;
		if( !(h->m_thing->is_a( "phPerson"))) return;
		p = (phPerson *)(h->m_thing);
	}

	//std::cout << "Skater \"" + (tcb())->m_key + "\" is thinking\n";
}
