/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef AI_SKATER_H
#define AI_SKATER_H

#include "ai.h"

class phPuck;

class task___skate_heading : public phAI_Task
{
public:
	std::string m_anim;

	std::string m_right_cross_cut;
	std::string m_left_cross_cut;

	std::string m_right_stop;
	std::string m_left_stop;

	std::string m_intensity;
	rgVector m_heading;

	virtual ~task___skate_heading(){}

	void do_task() { throw this;}
};

class task___skate_to : public phAI_Task
{
public:
	ThingControlBlock *m_tcb;
	rgFloat m_radius;
	rgFloat m_speed;
	int m_anim_delay;
	std::string m_anim;
	std::string m_intensity;
	rgVector m_point;

	task___skate_to() {
		m_tcb = 0;
		m_radius = 1.0f;
		m_speed = 0.1f;
		m_anim_delay = 250;
		m_intensity = "casual";
	}

	task___skate_to(
		const rgVector &p_point,
		const rgFloat p_radius,
		const rgFloat p_speed,
		const int p_anim_delay,
		const std::string &p_anim,
		const std::string &p_intensity)
	{
		m_tcb = 0;
		m_point = p_point;
		m_radius = p_radius;
		m_speed = p_speed;
		m_anim = p_anim;
		m_anim_delay = p_anim_delay;
		m_intensity = p_intensity;
	}

	task___skate_to(
		gmThing *p_thing,
		const rgFloat p_radius,
		const rgFloat p_speed,
		const int p_anim_delay,
		const std::string &p_anim,
		const std::string &p_intensity)
	{
		if( p_thing) m_tcb = p_thing->tcb();
		else m_tcb = 0;
		m_radius = p_radius;
		m_speed = p_speed;
		m_anim = p_anim;
		m_anim_delay = p_anim_delay;
		m_intensity = p_intensity;
	}

	virtual ~task___skate_to(){}

	void do_task() { throw this;}
};

class phAI_Skater : public phAI
{
public:
	RPH_THING_MODULE

	phAI_Skater();
	virtual ~phAI_Skater();

	void process( ThingManager *tm);

	std::string type() const { return std::string( "phAI_Skater");}
	bool is_a( const std::string &s) const {
		if( s != phAI_Skater::type()) return phAI::is_a( s);
		return true;
	}

	gmTE call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	std::string turn_to_heading( const rgVector &dir,
		const std::string &p_intensity);

	std::string turn_to_face( const rgVector &p,
		const std::string &p_intensity);
	std::string turn_to_face( gmThing *t, const std::string &p_intensity);
	std::string turn_to_face( ThingControlBlock *t,
		const std::string &p_intensity);

	std::string skate_to( task___skate_to *p_task);
	std::string skate_heading( task___skate_heading *p_task);

protected:
	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;

	// This AI was designed for skaters.
	bool accept_host( ThingControlBlock *t) {
		if( t->is_free()) return false;
		if( !(t->m_thing->is_a( "phSkater"))) return false;
		return true;
	}

private:

	phAI_Skater( const phAI_Skater &);
	phAI_Skater &operator =( const phAI_Skater &);
};

#endif
