/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef ARENA_H
#define ARENA_H

#include <rgOpenGL.h>
#include "../things/things.h"

class phPuck;
class phTeam;
class phPerson;
class phReferee;
class phCamera;

class phArena : public gmModelThing
{
public:
	RPH_THING_MODULE

	phArena();
	virtual ~phArena();

	void clear();
	int load( const char *fname);
	void draw_ice_surface() const {
		if( dsplst_ice_surface) glCallList( dsplst_ice_surface);}

	void process( ThingManager *tm);
	void draw_boards() const;
	void draw_glass();

	void clear_display_list( GLuint &id);

	std::string type() const { return std::string( "phArena");}
	bool is_a( const std::string &s) const {
		if( s != phArena::type()) return gmModelThing::is_a( s);
		return true;
	}

	gmThingError call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	// Render lists
	bool add_to_render_list( ThingControlBlock *tcb,
		const std::string &des);
	void remove_from_render_list( ThingControlBlock *tcb,
		const std::string &tar);
	void remove_from_render_list( ThingControlBlock *tcb,
		std::map <ThingControlBlock *, void *> *tar);
	std::map <ThingControlBlock *, void *> *get_render_list(
		const std::string &lname);

	phPuck *game_puck() {
		if(	!m_game_puck ||
			(m_game_puck->is_free()) ||
			!(m_game_puck->m_thing->is_a( "phPuck"))) return 0;
		return ((phPuck *)( m_game_puck->m_thing));
	}
	void set_game_puck( ThingControlBlock *p_game_puck) {
		m_game_puck = p_game_puck;}

	phPerson *nearest_puck( const std::vector <phPerson *> &plist);

	phReferee *primary_referee() {
		if(	!m_primary_referee ||
			(m_primary_referee->is_free()) ||
			!(m_primary_referee->m_thing->is_a( "phReferee")))
			return 0;
		return ((phReferee *)(m_primary_referee->m_thing));
	}
	void set_primary_referee( ThingControlBlock *p_primary_referee){
		m_primary_referee = p_primary_referee;}


	phTeam *home_team() {
		if(	!m_home_team ||
			(m_home_team->is_free()) ||
			!(m_home_team->m_thing->is_a( "phTeam"))) return 0;
		return ((phTeam *)(m_home_team->m_thing));
	}
	void set_home_team( ThingControlBlock *p_team) {
		m_home_team = p_team;}

	phTeam *away_team() {
		if(	!m_away_team ||
			(m_away_team->is_free()) ||
			!(m_away_team->m_thing->is_a( "phTeam"))) return 0;
		return ((phTeam *)(m_away_team->m_thing));
	}
	void set_away_team( ThingControlBlock *p_team) {
		m_away_team = p_team;}
	

protected:
	rgVector m_ice_color;
	rgVector m_light_level;
	rgVector m_shadow_color;
	float m_ice_alpha;
	bool m_reflections;
	bool m_stencil_shadows;
	float m_shadow_alpha;


	// The main official that controls the game. There can be additional
	// referees but they would be under the control of the primary
	// referee, and server only to assist the primary referee.
	ThingControlBlock *m_primary_referee;

	// The official game puck used in games. There can be additional pucks,
	// however they won't be official game pucks.
public:
	ThingControlBlock *m_game_puck;


	ThingControlBlock *m_home_team;
	ThingControlBlock *m_away_team;


	// The arena class keeps a number of render lists because certain
	// things require special rendering techniques. The rendering of the
	// arena, players and all other objects is tightly coupled do to the
	// lack of an acceptable generic rendering algorithm.
	std::map <ThingControlBlock *, void *> m_opaque;

	// Players are rendered specially (for optimizations and perhaps some
	// special effects).
	std::map <ThingControlBlock *, void *> m_player;

	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;

	GLuint dsplst_ice_surface;
	GLuint dsplst_boards;

	void compile_face_list( const char *p_name, GLuint &id);
	void compile_face_list( const char *p_name[], const int p_count,
		GLuint &id);

	void render_all_in_list( gmRenderArgs *, const std::string &rlist);


	friend gmTE phArena_set_official_game_thing( const gmCCArg &p);
	friend gmTE phArena_ice_color( const gmCCArg &p);
	friend gmTE phArena_ice_alpha( const gmCCArg &p);
	friend gmTE phArena_light_level( const gmCCArg &p);
	friend gmTE phArena_reflections( const gmCCArg &p);
	friend gmTE phArena_stencil_shadows( const gmCCArg &p);
	friend gmTE phArena_shadow_alpha( const gmCCArg &p);
	friend gmTE phArena_prepare_for_rendering( const gmCCArg &p);
	friend gmTE phArena_add_to_render_list( const gmCCArg &p);
	friend gmTE phArena_remove_from_render_list( const gmCCArg &p);
	friend gmTE phArena_shadow_color( const gmCCArg &p);

private:
	phArena( const phArena &);
	phArena &operator =( const phArena &);
};

phCamera *get_camera_controller( gmCamera *);

#endif
