/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "camera.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
//#include <rg3DRom.h>

std::map <const std::string, THING_COMMAND_FN> phCamera::m_ccmd;
std::map <const std::string, THING_COMMAND_FN> phCameraPitch::m_ccmd;

extern "C" gmThing *spw_phCamera()
{ return new phCamera();}

extern "C" int del_phCamera( gmThing *t)
{ delete t; return 0;}

extern "C" gmThing *spw_phCameraPitch()
{ return new phCameraPitch();}

extern "C" int del_phCameraPitch( gmThing *t)
{ delete t; return 0;}

RPH_THING_MODULE_MAIN_FUNCTION( phCamera)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phCamera", spw_phCamera);
	p_ge->m_thing_manager->add_to_delete_list( "phCamera", del_phCamera);

	p_ge->m_thing_manager->add_to_spawn_list(
		"phCameraPitch", spw_phCameraPitch);
	p_ge->m_thing_manager->add_to_delete_list(
		"phCameraPitch", del_phCameraPitch);

	return 0;
}

phCamera::phCamera()
{
	if( !(m_ccmd.empty())) return;
	add_console_command( "target", phCamera_target, m_ccmd);
}

phCamera::~phCamera()
{
}

gmTE phCamera_target( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: target <thing>\n"
			"NOTE: \"thing\" is the unique name of the thing to "
				"target.\n");
		return GM_THING_ERR_PARAM;
	}
	ThingControlBlock *t = p.ge->m_thing_manager->get( args[0]);
	if( !t) {
		p.ge->rcout( "No such thing \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	((phCamera *)(p.thing))->m_look_at_target = t;
	return 0;
}

gmTE phCamera::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return gmModelThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phCamera::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phCamera::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += gmModelThing::list_ccmd( p_prefix);
	return tmp;
}

void phCamera::process( ThingManager *tm)
{
}




// XXX: Pitch camera
phCameraPitch::phCameraPitch()
{
	m_heading = rgVector( 0, 0, 1);

	if( !(m_ccmd.empty())) return;
	add_console_command( "heading", phCameraPitch_heading, m_ccmd);
}

gmTE phCameraPitch::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phCamera::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phCameraPitch::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phCameraPitch::type() + "::" + (iter->first)
			+ "\n";
	}
	return (tmp + phCamera::list_ccmd( p_prefix));
}

gmTE phCameraPitch_heading( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 3) {
		p.ge->rcout( "usage: heading <x> <y> <z>\n");
		return GM_THING_ERR_PARAM;
	}
	((phCameraPitch *)(p.thing))->m_heading = gm_str_to_vector( args[0],
		args[1], args[2]);
	return 0;
}

rgVector phCameraPitch::viewer_to_world( const rgVector &p_v)
{
	gmCamera *c = controlled_camera();
	if( !c) return p_v;
	//const float ang = rgDotAngle( c->cs.side, rgVector( 1, 0, 0));
	//const rgMatrix m( ang, rgVector( 0, 1, 0));
	rgVector v( p_v);
	rgSwap( v.y, v.z);
	if( (rgVector( 1, 0, 0)).dot( c->cs.side) < 0.0f) {
		v.z = -v.z;
	} else {
		v.x = -v.x;
	}
	return v;
}

void phCameraPitch::process( ThingManager *tm)
{
	// XXX: Camera is disabled when not in use.
	if( (ge())->gfx_engine->current_camera() != controlled_camera())
		return;

	const float m_cam_quickness = 0.0625f;

	gmCamera *c = controlled_camera();
	if( !c) return;

	gmThing *tar = get_look_at_thing();
	if( !tar) return;
	rgVector t( (tar->cs()).p);

	// Position camera to give best view of the zone the "target" is in.
	rgVector p( (cs()).p.x + t.x, (cs()).p.y, t.z);
	float cam_z_min = 10.0f;
	float cam_z_max = 25.0f;
	{
		rgVector h( m_heading);
		p -= h;
		if( p.z < 0.0f) {
			const float w = fabs( (0.0f - p.z) / 30.0f);
			p.y += 2.5f * RG_MIN( 1.0f, w);
			const float pitch_offset = (h.z * RG_MIN( 1.0f, w));
			p.z += pitch_offset;
		}

		if( p.z < -5.0f) {
			const float d = RG_MIN( 5.0f, -(p.z + 5.0f) * 0.5f);
			p.y += d;
			p.z -= d;
			t.z -= d;
		}

		if( p.z > cam_z_min) {
			if( p.z < 0.0f) p.z = -cam_z_min;
			else p.z = cam_z_min;
		}

		if( fabs( p.z) > cam_z_max) {
			if( p.z < 0.0f) p.z = -cam_z_max;
			else p.z = cam_z_max;
		}
	}

	// Pitch camera to give best view of the zone the "target" is in.
	c->move_to( p, m_cam_quickness);
	c->look_at( t, rgVector( 0, 0, 1), m_cam_quickness);
	if( c->cs.eye.z < 0.0f) {
		c->look_at(
			rgVector( c->cs.p.x, t.y, c->cs.p.z),
			(rgVector( m_heading)).normalize());
	}
}

