/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef CAMERA_H
#define CAMERA_H

#include "../things/things.h"

class phCamera : public gmModelThing
{
public:
	RPH_THING_MODULE

	phCamera();
	~phCamera();

	std::string type() const { return std::string( "phCamera");}
	bool is_a( const std::string &s) const {
		if( s != phCamera::type()) return gmModelThing::is_a( s);
		return true;
	}

	gmTE call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	void process( ThingManager *tm);

	void set_look_at( ThingControlBlock *p_tcb) {
		m_look_at_target = p_tcb;}
	ThingControlBlock *get_look_at() { return m_look_at_target;}
	gmThing *get_look_at_thing() {
		ThingControlBlock *t = get_look_at();
		if( !t) return 0;
		return t->m_thing;
	}

	virtual rgVector viewer_to_world( const rgVector &v) { return v;}

protected:

	ThingControlBlock *m_look_at_target;
	gmCamera *m_cam;

	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;

	friend gmTE phCamera_target( const gmCCArg &p);

private:
	phCamera( const phCamera &);
	phCamera &operator =( const phCamera &);
};


class phCameraPitch : public phCamera
{
public:
	phCameraPitch();

	std::string type() const { return std::string( "phCameraPitch");}
	bool is_a( const std::string &s) const {
		if( s != phCameraPitch::type()) return phCamera::is_a( s);
		return true;
	}

	gmTE call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	void process( ThingManager *tm);

	rgVector viewer_to_world( const rgVector &v); 

protected:
	// m_heading is a vector which points either to the "home" side or the
	// "away" side of the arena. This lets the camera know from which side
	// it should view the ice surface from.
	rgVector m_heading;

	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;

	friend gmTE phCameraPitch_heading( const gmCCArg &p);

private:
	phCameraPitch( const phCameraPitch &);
	phCameraPitch &operator =( const phCameraPitch &);
};

#endif
