/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef PLAYER_H
#define PLAYER_H
                                                                                
#include "../things/things.h"

class phTeam;
class phArena;
class phPuck;

class phUniformSubPatch
{
public:
	float m_left;
	float m_bottom;
	float m_width;
	float m_height;
	float m_char_width;
	float m_char_height;
	float m_char_space;

	phUniformSubPatch() {
		m_left = 0.0f;
		m_bottom = 0.0f;
		m_width= 0.0f;
		m_height = 0.0f;
		m_char_width = 0.0f;
		m_char_height = 0.0f;
		m_char_space = 0.0f;
	}
};

class phUniformPatchContext
{
public:
	int m_width;
	int m_height;

	std::string m_base_uniform_page;
	std::string m_patch_page;

	phUniformSubPatch m_big_num_patch;
	phUniformSubPatch m_name_patch;
	phUniformSubPatch m_rnum_patch;
	phUniformSubPatch m_lnum_patch;

	phUniformPatchContext(){}

	phUniformPatchContext( int w, int h) {
		m_width = w;
		m_height = h;
	}
};

class phPersonCursor
{
public:	
	rgMatrix m_mat;
	std::string m_tex_name;
	rgVector m_color;
	float m_alpha;

	phPersonCursor() {
		m_color = rgVector( 1, 1, 1);
		m_alpha = 0.5f;
	}

	void draw( const rgVector p) const;
};

class phPerson : public gmPhyModel
{
public:
	RPH_THING_MODULE

	std::string m_full_name;

	std::string m_datt_speed;

	phPerson();
	virtual ~phPerson();

	std::string type() const { return std::string( "phPerson");}
		bool is_a( const std::string &s) const {
		if( s != phPerson::type()) return gmPhyModel::is_a( s);
		return true;
	}

	gmThingError call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	phTeam *team();
	void set_team( ThingControlBlock *p_team) { m_team = p_team;}

	phArena *arena();
	void set_arena( ThingControlBlock *p_arena) { m_arena = p_arena;}

	phPuck *has_game_puck();


	// XXX: Helper functions
	bool near_game_puck( const float radius);
	bool near_ice_point( rgVector p, const float radius);

	void set_cursor_texture( const std::string &tex) {
		m_cursor.m_tex_name = tex;}

	phPersonCursor *get_cursor() { return &m_cursor;}

protected:


	/////////////////////////////////////////////////////////////////////
	// Some common "things" that "hockey people" would like to know.
	/////////////////////////////////////////////////////////////////////
	ThingControlBlock *m_arena;
	ThingControlBlock *m_team;		// Which team they belong

	phPersonCursor m_cursor;

	std::string m_head_texture;



	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;
	friend gmTE phPerson_set_full_name( const gmCCArg &p);
	friend gmTE phPerson_cursor_color( const gmCCArg &p);
	friend gmTE phPerson_cursor_texture( const gmCCArg &p);
	friend gmTE phPerson_scale_cursor( const gmCCArg &p);
	friend gmTE phPerson_set_head_texture( const gmCCArg &p);

private:

	phPerson( const phPerson &);
	phPerson &operator =( const phPerson &);
};

class phPlayer : public phPerson
{
public:
	int m_rank;
	int m_number;

	phPlayer();
	virtual ~phPlayer();

	std::string type() const { return std::string( "phPlayer");}
		bool is_a( const std::string &s) const {
		if( s != phPlayer::type()) return phPerson::is_a( s);
		return true;
	}

	gmThingError call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	void process( ThingManager *tm);

	void draw( gmRenderArgs *ra);
	void set_position( const std::string &pos) { m_position = pos;}

	std::string create_uniform_patch() {
		return create_uniform_patch( (tcb())->m_key, &m_upc);
	}
	std::string create_uniform_patch(
		const std::string &p_name,
		phUniformPatchContext *upc);

protected:
	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;

	std::string m_position;

	std::string m_uniform_patch;

	phUniformPatchContext m_upc;

	friend gmTE phPlayer_set_attrib( const gmCCArg &p);
	friend gmTE phPlayer_create_uniform_patch( const gmCCArg &p);
	friend gmTE phPlayer_set_sub_patch( const gmCCArg &p);


private:
	void render_uniform_patch(
		rgGLTexture *p_tex,
		phUniformPatchContext *upc);

	void render_base_uniform(
		rgGLTexture *p_tex,
		phUniformPatchContext *upc);

	void render_uniform_string(
		phUniformPatchContext *upc,
		phUniformSubPatch *sp,
		const std::string &str,
		bool use_ascii);

	phPlayer( const phPlayer &);
	phPlayer &operator =( const phPlayer &);
};

#endif
