/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "referee.h"
#include "engine/engine_graphics.h"
#include "engine/engine_game.h"
#include <rg3DRom.h>

std::map <const std::string, THING_COMMAND_FN> phReferee::m_ccmd;

extern "C" gmThing *spw_phReferee()
{
	return new phReferee();
}

extern "C" int del_phReferee( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phReferee)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phReferee", spw_phReferee);

	p_ge->m_thing_manager->add_to_delete_list( "phReferee", del_phReferee);

	return 0;
}

phReferee::phReferee()
{
	m_faceoff_wait = -1;	// Wait forever

	if( !(m_ccmd.empty())) return;
	add_console_command( "set_faceoff_wait", phReferee_faceoff_wait,
		m_ccmd);
}

phReferee::~phReferee()
{
}

gmTE phReferee::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phPerson::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phReferee::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phReferee::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += phPerson::list_ccmd( p_prefix);
	return tmp;
}

gmTE phReferee_faceoff_wait( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: set_faceoff_wait <time in seconds>\n"
			"NOTE: \"time in seconds\" must be an integer.\n"
			"Negative seconds means wait forever.\n");
		return GM_THING_ERR_PARAM;
	}

	phReferee *r = (phReferee *)(p.thing);
	r->m_faceoff_wait = atoi( (args[0]).c_str()) * p.ge->game_clock_hz();
	return 0;
}


void phReferee::process( ThingManager *tm)
{
	gmPhyModel::process( tm);

	// Think...think...think...
	ThingControlBlock *ai = get_ai();
	if( ai && !(ai->is_free())) ai->m_thing->process( tm);

}

void phReferee::draw( gmRenderArgs *p_ra)
{
	if( !p_ra) return;
	gmRenderArgs ra( *p_ra);

	// XXX: The arena already disabled transformation for transparency.
	// if( ra.render_transparent) ra.transform = false;
	if( ra.render_shadow) {

		ModelHandle *mh = ((ge())->gfx_engine->model_man())->get(
			"mdl_referee_shadow");
		if( mh && (mh->model)) {

			(mh->model)->transform_skel_frame( m_draw_context);
			(mh->model)->render_vertex_array( 0);
		}
		return;
	}

	// Render the full opaque part of the model OR render the
	// transparent parts of the model. Transformation for the model
	// is suppressed if transparent parts will be rendered.
	if( ra.render_reflections) ra.lock_vertices = true;
	gmModelThing::draw( &ra);

	if( ra.render_reflections && !(ra.render_transparent)) {

		glPushMatrix();
		glScalef( 1, -1, 1);
		glCullFace( GL_FRONT);

		ra.transform = false;
		gmModelThing::draw( &ra);

		glCullFace( GL_BACK);
		glPopMatrix();
	}
}
