/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "team.h"
#include "arena.h"
#include "player.h"
#include "coach.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
#include <rg3DRom.h>

std::map <const std::string, THING_COMMAND_FN> phTeam::m_ccmd;

extern "C" gmThing *spw_phTeam()
{
	return new phTeam();
}

extern "C" int del_phTeam( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phTeam)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phTeam", spw_phTeam);

	p_ge->m_thing_manager->add_to_delete_list( "phTeam", del_phTeam);

	return 0;
}

phTeam::phTeam()
{
	m_head_coach = 0;

	// 3 forwards, 2 defense and one goalie is the standard combination.
	set_max_iced_players( 6);

	{
		const int total_positions = 22;
		m_official_position =
			new (ThingControlBlock *)[total_positions];
		memset( m_official_position, 0, sizeof( ThingControlBlock *) *
			total_positions);
	}

	{ // Create quick lookup for specific "official" positions.
		typedef std::pair <const std::string, ThingControlBlock **>
			STPair;
		m_position.insert( STPair( "C1", m_official_position));
		m_position.insert( STPair( "C2", m_official_position + 1));
		m_position.insert( STPair( "C3", m_official_position + 2));
		m_position.insert( STPair( "C4", m_official_position + 3));

		m_position.insert( STPair( "LW1", m_official_position + 4));
		m_position.insert( STPair( "LW2", m_official_position + 5));
		m_position.insert( STPair( "LW3", m_official_position + 6));
		m_position.insert( STPair( "LW4", m_official_position + 7));

		m_position.insert( STPair( "RW1", m_official_position + 8));
		m_position.insert( STPair( "RW2", m_official_position + 9));
		m_position.insert( STPair( "RW3", m_official_position + 10));
		m_position.insert( STPair( "RW4", m_official_position + 11));

		m_position.insert( STPair( "LD1", m_official_position + 12));
		m_position.insert( STPair( "LD2", m_official_position + 13));
		m_position.insert( STPair( "LD3", m_official_position + 14));

		m_position.insert( STPair( "RD1", m_official_position + 15));
		m_position.insert( STPair( "RD2", m_official_position + 16));
		m_position.insert( STPair( "RD3", m_official_position + 17));

		m_position.insert( STPair( "E1", m_official_position + 18));
		m_position.insert( STPair( "E2", m_official_position + 19));

		m_position.insert( STPair( "G1", m_official_position + 20));
		m_position.insert( STPair( "G2", m_official_position + 21));
	}


	if( !(m_ccmd.empty())) return;
	add_console_command( "compile_uniform_patches",
		phTeam_compile_uniform_patches, m_ccmd);
	add_console_command( "set_uniform", phTeam_set_uniform, m_ccmd);
	add_console_command( "add_person", phTeam_add_person, m_ccmd);
	add_console_command( "add_person_list", phTeam_add_person_list,
		m_ccmd);
	add_console_command( "dress", phTeam_dress, m_ccmd);
}

phTeam::~phTeam()
{
	if( m_official_position) delete [] m_official_position;
}

gmTE phTeam::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return gmThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phTeam::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter)
		tmp += p_prefix + phTeam::type() + "::" + (iter->first) + "\n";
	tmp += gmThing::list_ccmd( p_prefix);
	return tmp;
}

gmTE phTeam_compile_uniform_patches( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: compile_uniform_patches <patch tex>\n"
			"NOTE: \"patch tex\" must be the name of pre-loaded "
				"textures.\n");
		return GM_THING_ERR_PARAM;
	}
	phTeam *t = (phTeam *)(p.thing);
	const std::string emsg( t->compile_uniform_patches( args[0]));
	if( emsg.length() > 0) return GM_THING_ERR_GENERIC;
	return 0;
}

gmTE phTeam_set_uniform( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 4) {
                p.ge->rcout( "usage: set_uniform <player upper> "
			"<player lower> <goalie upper> <goalie lower>\n");
                return GM_THING_ERR_PARAM;
        }
	phTeam *t = (phTeam *)(p.thing);
	const std::string emsg( t->set_uniform( args[0], args[1], args[2],
		args[3]));
	if( emsg.length() > 0) return GM_THING_ERR_GENERIC;
        return 0;
}

gmTE phTeam_add_person( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 2) {
                p.ge->rcout( "usage: add_person <destination list> <person>\n"
			"NOTE: Person must be of type phCoach or phPlayer.\n"
			"So for example, you cannot add officials ("
				"phReferee).\n");
                return GM_THING_ERR_PARAM;
        }
	phTeam *t = (phTeam *)(p.thing);
	const std::string emsg( t->add_person( args[0], args[1]));
	if( emsg.length() > 0) {
		p.ge->rcout( emsg);
		return GM_THING_ERR_GENERIC;
	}
        return 0;
}

gmTE phTeam_dress( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 2) {
                p.ge->rcout( "usage: dress <player> <position>\n");
                return GM_THING_ERR_PARAM;
        }
	phTeam *t = (phTeam *)(p.thing);
	const std::string emsg( t->dress( args[0], args[1]));
	if( emsg.length() > 0) {
		p.ge->rcout( emsg + "\n");
		return GM_THING_ERR_GENERIC;
	}
        return 0;
}

gmTE phTeam_add_person_list( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 1) {
                p.ge->rcout( "usage: add_person_list <list name>\n"
			"NOTE: The list name \"head_coach\" is reserved and "
				"may not be used.\n");
                return GM_THING_ERR_PARAM;
        }
	phTeam *t = (phTeam *)(p.thing);
	const std::string emsg( t->create_person_list( args[0]));
	if( emsg.length() > 0) {
		p.ge->rcout( emsg);
		return GM_THING_ERR_GENERIC;
	}
        return 0;
}

ThingControlBlock **phTeam::get_position_tcbptr( const std::string &pos)
{
	std::map <const std::string, ThingControlBlock **> :: const_iterator
		iter = m_position.find( pos);
	if( iter == m_position.end()) return 0;
	return (iter->second);
}

phPerson *phTeam::get_position( const std::string &pos)
{
	ThingControlBlock **t = get_position_tcbptr( pos);
	if( !t) return 0;
	if( !(*t)) return 0;
	if( (*t)->is_free()) return 0;
	if( !((*t)->m_thing->is_a( "phPerson"))) return 0;
	return ((phPerson *)((*t)->m_thing));
}

std::string phTeam::get_position_name( const std::string &pos)
{
	phPerson *p = get_position( pos);
	if( !p) return std::string();
	return ((p->tcb())->m_key);
}

std::string phTeam::dress( const std::string &p, const std::string &p_pos)
{
	ThingControlBlock *ply = get_person( "players", p);
	if( !ply || (ply->is_free())) {
		return "player \"" + p + "\" is not a member of team \""
			+ (tcb())->m_key + "\"";
	}
	if( !(ply->m_thing->is_a( "phPlayer")))
		return "Thing \"" + p + "\" is not a player";

	ThingControlBlock **pos = get_position_tcbptr( p_pos);
	if( !pos) return "no such position \"" + p_pos + "\"";

	if( (*pos)) {
		return "player \"" + (*pos)->m_key + "\" is already dressed "
			"for position of \"" + p_pos + "\"";
	}

	(*pos) = ply;
	((phPlayer *)(ply->m_thing))->set_position( p_pos);
	return std::string();
}

ThingControlBlock *phTeam::get_person( const std::string &p_list,
	const std::string &p_person)
{
	Map_String_TCB *tmp = get_person_list( p_list);
	if( !tmp) return 0;

	Map_String_TCB :: const_iterator iter = tmp->find( p_person);
	if( iter == tmp->end()) return 0;
	return (iter->second);
}

Map_String_TCB *phTeam::get_person_list( const std::string &key)
{
	std::map <const std::string, Map_String_TCB *> :: const_iterator iter =
		m_person.find( key);
	if( iter == m_person.end()) return 0;
	return (iter->second);
}

std::string phTeam::add_person( const std::string &p_list,
	const std::string &p_person)
{
	ThingControlBlock *t = (ge())->m_thing_manager->get( p_person);
	if( !t) return ("No such thing \"" + p_person + "\"");
	if( t->is_free())
		return ("Thing \"" + p_person + "\" is free (unusable)");

	if( "head_coach" == p_list) {
		if( m_head_coach)
			return "Must fire current head coach first";
		if( !(t->m_thing->is_a( "phCoach")))
			return "\"" + p_person + "\" is not a coach";
		m_head_coach = t;
		((phCoach *)(t->m_thing))->set_team( tcb());
		((phCoach *)(t->m_thing))->set_arena( m_arena);
		return std::string();
	}

	if( !(t->m_thing->is_a( "phPerson")))
		return "\"" + p_person + "\" is not a person";

	std::map <const std::string, ThingControlBlock *> *tmp =
		get_person_list( p_list);
	if( !tmp) return ("No such person list \"" + p_list + "\"");

	std::string emsg( add_person( tmp, t));
	if( emsg.length() > 0) emsg = "List \"" + p_list + "\": " + emsg;
	else {
		((phPerson *)(t->m_thing))->set_team( tcb());
		((phPerson *)(t->m_thing))->set_arena( m_arena);
	}

	return emsg;
}

std::string phTeam::add_person(
	std::map <const std::string, ThingControlBlock *> *p_list,
	ThingControlBlock *p_tcb)
{
	typedef std::pair < const std::string, ThingControlBlock *> STPair;
	if( false == (p_list->insert( STPair( p_tcb->m_key, p_tcb))).second)
		return "Could not add thing \"" + p_tcb->m_key + "\"";
	return std::string();
}

std::string phTeam::remove_person( const std::string &p_list,
	const std::string &p_person)
{
	// Get the named list.
	Map_String_TCB *tmp = get_person_list( p_list);
	if( !tmp) return ("No such person list \"" + p_list + "\"");

	// Find them in the given list.
	Map_String_TCB :: iterator iter = tmp->find( p_person);
	if( iter == tmp->end()) return ("Person \"" + p_person + "\" not in \""
		+ p_list + "\" list");

	tmp->erase( iter);
	return std::string();
}

std::string phTeam::clear_person_list( const std::string &p_list)
{
	// Get the named list.
	Map_String_TCB *tmp = get_person_list( p_list);
	if( !tmp) return ("No such person list \"" + p_list + "\"");

	Map_String_TCB :: iterator iter = tmp->begin();
	for( ; iter != (tmp->end()); ++iter)
		tmp->erase( iter);
	return std::string();
}

std::string phTeam::create_person_list( const std::string &key)
{
	if( "head_coach" == key) return "list \"" + key + "\" is reserved";
	typedef std::pair <const std::string, Map_String_TCB *> SMPair;
	Map_String_TCB *tmp = new Map_String_TCB;
	if( !tmp) return "Could not create new \"map <string, tcb>\"";
	if( false == (m_person.insert( SMPair( key, tmp))).second) {
		delete tmp;
		return ("Could not add person list \"" + key + "\", "
			"(possibly a duplicate?)");
	}
	return std::string();
}

std::string phTeam::remove_person_list( const std::string &key)
{
	if( "head_coach" == key)
		return "list \"" + key + "\" cannot be removed";
	std::map <const std::string, Map_String_TCB *> :: iterator iter =
		m_person.find( key);
	if( iter == m_person.end()) return "No such list \"" + key + "\"";
	delete (iter->second);
	m_person.erase( iter);
	return std::string();
}

void phTeam::append_people( std::vector <phPerson *> &des,
	const std::string &p_list)
{
	Map_String_TCB *tmp = get_person_list( p_list);
	if( !tmp) return;
	Map_String_TCB :: const_iterator iter = tmp->begin();
	for( ; iter != tmp->end(); ++iter) {
		// TODO: Should remove freeed/deleted people/things from
		//	the list as they are found.
		ThingControlBlock *t = (iter->second);
		if(	t &&
			!(t->is_free()) &&
			(t->m_thing->is_a( "phPerson")))
			des.push_back( (phPerson *)(t->m_thing));
	}
}

void phTeam::list_people( std::ostream &o, const std::vector <phPerson *> &p)
{
	if( p.empty()) return;
	o << "{ ";
	int i = 0;
	for( ; i < int( p.size() - 1); i++) {
		if( p[i]) o << ((p[i])->tcb())->m_key << ", ";
		else o << "NULL, ";
	}
	if( p[i]) o << ((p[i])->tcb())->m_key << "}";
	else o << "NULL}";
}

std::string phTeam::compile_uniform_patches( const std::string &pat)
{
	TextureHandle *t_pat = ((ge())->gfx_engine)->get_texture( pat);
	std::string err;
	if( !t_pat) err += "No such texture \"" + pat + "\"\n";
	if( err.length() > 0) return err;

	if( !t_pat->m_tex)
		err += "Texture \"" + pat + "\" not ready (m_tex == NULL)\n";
	if( err.length() > 0) return err;

	// Prepare uniform patcher to use given uniform and patch.
	m_uniform.m_border_width = 1.0f / float( t_pat->m_tex->getWidth());
	m_uniform.m_patch = pat;
	m_uniform.reset_patch_tables();

	return err;
}

std::string phTeam::set_uniform(
	const std::string &p_up,
	const std::string &p_low,
	const std::string &g_up,
	const std::string &g_low)
{
	TextureHandle *t_up = ((ge())->gfx_engine)->get_texture( p_up);
	TextureHandle *t_low = ((ge())->gfx_engine)->get_texture( p_low);
	TextureHandle *t_gup = ((ge())->gfx_engine)->get_texture( g_up);
	TextureHandle *t_glow = ((ge())->gfx_engine)->get_texture( g_low);
	std::string err;
	if( !t_up) err += "No such texture \"" + p_up + "\"\n";
	if( !t_low) err += "No such texture \"" + p_low + "\"\n";
	if( !t_gup) err += "No such texture \"" + g_up + "\"\n";
	if( !t_glow) err += "No such texture \"" + g_low + "\"\n";
	if( err.length() > 0) return err;

	const char *tmp = "\"not ready (m_tex == NULL)\n";
	if( !t_up->m_tex) err += "Texture \"" + p_up + tmp;
	if( !t_low->m_tex) err += "Texture \"" + p_low + tmp;
	if( !t_gup->m_tex) err += "Texture \"" + g_up + tmp;
	if( !t_glow->m_tex) err += "Texture \"" + g_low + tmp;
	if( err.length() > 0) return err;

	m_ply_upper_uniform = p_up;
	m_ply_lower_uniform = p_low;
	m_goal_lower_uniform = g_up;
	m_goal_lower_uniform = g_low;
	return err;
}

void phTeam::bind_uniform_textures( phPerson *p)
{
	GraphicsEngine *gfx = (ge())->gfx_engine;
	ModelManager *mm = gfx->model_man();
	if( p->is_a( "phGoalie")) {
		ModelHandle *mh = mm->get( "mdl_goalie");
		if( !mh) return;
		mh->set_alternate_texture( "upper", m_goal_upper_uniform, gfx);
		mh->set_alternate_texture( "lower", m_goal_lower_uniform, gfx);

	} else {
		ModelHandle *mh = mm->get( "mdl_player");
		if( !mh) return;
		mh->set_alternate_texture( "upper", m_ply_upper_uniform, gfx);
		mh->set_alternate_texture( "lower", m_ply_lower_uniform, gfx);
	}

}

void phUniformPatcher::reset_big_number_table()
{
	const float w = (1.0f / 8.0f) - (m_border_width * 3.0f);
	const float h = (1.0f / 8.0f) - (m_border_width * 3.0f);
	const float b = m_border_width * 2.0f;
	{ // First row of numbers (0 to 7)
		for( int i = 0; i < 8; i++) {
			m_big_number[i].m_left = i * 0.125f + m_border_width;
			m_big_number[i].m_bottom = 0.875f + b;
			m_big_number[i].m_width = w;
			m_big_number[i].m_height = h;
		}
	}
	{ // Second row of numbers (8 to F)
		for( int i = 8; i < 16; i++) {
			m_big_number[i].m_left = i * 0.125f + m_border_width;
			m_big_number[i].m_bottom = 0.75f + b;
			m_big_number[i].m_width = w;
			m_big_number[i].m_height = h;
		}
	}
}

void phUniformPatcher::reset_pic_table()
{
	{ // First row of pics
		for( int i = 0; i < 8; i++) {
			m_pic[i].m_left = float( i) * 0.125f;
			m_pic[i].m_bottom = 0.625f;
			m_pic[i].m_width = 0.125f;
			m_pic[i].m_height = 0.125f;
		}
	}
	{ // Second row of pics
		for( int i = 8; i < 16; i++) {
			m_pic[i].m_left = float( i) * 0.125f;
			m_pic[i].m_bottom = 0.5f;
			m_pic[i].m_width = 0.125f;
			m_pic[i].m_height = 0.125f;
		}
	}
}

void phUniformPatcher::reset_ascii_table()
{
	const float w = (1.0f / 16.0f) - (m_border_width * 3.0f);
	const float h = (1.0f / 16.0f) - (m_border_width * 3.0f);
	const float b = m_border_width * 2.0f;

	phUniformPatch *p = m_ascii;
	float bottom = 0.5f;
	for( int y = 0; y < 8; y++) {
		bottom -= 0.0625f;
		for( int x = 0; x < 16; x++) {
			p->m_left = float( x) * 0.0625f + m_border_width;
			p->m_bottom = bottom + b;
			p->m_width = w;
			p->m_height = h;
			p++;
		}
	}
}
