/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef TEAM_H
#define TEAM_H

#include "../things/things.h"

class phCoach;
class phArena;
class phPerson;
class phTeam;


typedef std::map <const std::string, ThingControlBlock *> * phTeamPListPtr;

class phUniformPatch
{
public:
	float m_left;
	float m_bottom;	

	float m_width;
	float m_height;

	float m_base;
};

class phUniformContext
{
public:
	int m_big_number_x;
	int m_big_number_y;
	int m_big_number_width;
	int m_big_number_height;

	int m_name_x;
	int m_name_y;
	int m_name_width;
	int m_name_height;
};

class phUniformPatcher
{
public:
	float m_border_width;
	std::string m_patch;
	
	phUniformPatch m_big_number[16];
	phUniformPatch m_pic[16];
	phUniformPatch m_ascii[128];

	phUniformPatcher() { reset_patch_tables();}
	void reset_big_number_table();
	void reset_pic_table();
	void reset_ascii_table();
	void reset_patch_tables() {
		reset_big_number_table();
		reset_pic_table();
		reset_ascii_table();
	}
};

class phTeam : public gmThing
{
public:
	RPH_THING_MODULE

	phUniformPatcher m_uniform;
	phUniformContext m_player;
	phUniformContext m_goalie;

	phTeam();
	virtual ~phTeam();

	std::string type() const { return std::string( "phTeam");}
	bool is_a( const std::string &s) const {
		if( s != phTeam::type()) return gmThing::is_a( s);
		return true;
	}

	gmTE call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	ThingControlBlock **get_position_tcbptr( const std::string &pos);
	phPerson *get_position( const std::string &pos);
	std::string get_position_name( const std::string &pos);
	
	std::string dress( const std::string &p, const std::string &p_pos);

	ThingControlBlock *get_person( const std::string &p_list,
		const std::string &p_person);

	void append_people( std::vector <phPerson *> &des,
		const std::string &p_list);

	void get_people( std::vector <phPerson *> &des,
		const std::vector <std::string> &p_list) {
		des.clear();
		for( int i = 0; i < int( p_list.size()); i++)
			append_people( des, p_list[i]);
	}

	void get_people( std::vector <phPerson *> &des,
		const std::string &p_list) {
		des.clear();
		append_people( des, p_list);
	}


	std::string add_person( const std::string &p_list,
		const std::string &p_person);

	std::string add_person( const std::string &p_list,
		ThingControlBlock *t) {
		if( !t) return "TCB pointer is NULL";
		return add_person( p_list, t->m_key);
	}

	std::string remove_person( const std::string &p_list,
		const std::string &p_person);

	std::string create_person_list( const std::string &key);
	std::string clear_person_list( const std::string &key);

	std::string remove_person_list( const std::string &key);

	Map_String_TCB *get_person_list( const std::string &key);

	phCoach *get_coach() {
		if(	!m_head_coach ||
			(m_head_coach->is_free()) ||
			!(m_head_coach->m_thing->is_a( "phCoach"))) return 0;
		return ((phCoach *)(m_head_coach->m_thing));
	}

	phArena *get_arena() {
		if(	!m_arena ||
			(m_arena->is_free()) ||
			!(m_arena->m_thing->is_a( "phArena"))) return 0;
		return ((phArena *)(m_arena->m_thing));
	}

	void set_arena( ThingControlBlock *p_arena) { m_arena = p_arena;}

	int max_iced_players() const { return m_max_iced_players;}
	void set_max_iced_players( const int n) {
		m_max_iced_players = RG_MAX( 1, n);
	}

	void list_people( std::ostream &o, const std::vector <phPerson *> &p);
	void list_people( std::ostream &o, const std::string &p_list)
	{
		std::vector <phPerson *> tmp;
		get_people( tmp, p_list);
		list_people( o, tmp);
	}

	void schedule_lines( const std::vector <std::string> &p_line) {
		m_sched_lines = p_line;}

	const std::vector <std::string> &get_sched_lines() const {
		return m_sched_lines;}

	std::string compile_uniform_patches( const std::string &pat);

	std::string phTeam::set_uniform(
		const std::string &p_up,
		const std::string &p_low,
		const std::string &g_up,
		const std::string &g_low);

	void bind_uniform_textures( phPerson *p);

	const std::string &player_upper_uniform_texture() const {
		return m_ply_upper_uniform;}

	const std::string &goalie_upper_uniform_texture() const {
		return m_goal_upper_uniform;}

protected:
	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;

	phUniformContext m_player_uc;
	phUniformContext m_goalie_uc;

	std::string m_ply_upper_uniform;
	std::string m_ply_lower_uniform;
	std::string m_goal_upper_uniform;
	std::string m_goal_lower_uniform;

	ThingControlBlock *m_arena;

	// XXX: There is only one head coach per team. Extra coaches (if added)
	//	will be placed in the "persons" list.
	ThingControlBlock *m_head_coach;

	// All players, extra coaches, and other team staff are listed here.
	// That is, everyone except the head coach.
	std::map <const std::string,
		std::map <const std::string, ThingControlBlock *> *> m_person;



	//////////////////////////////////////////////////////////////////////
	// Official team positions (dressed players)
	//////////////////////////////////////////////////////////////////////
	std::map <const std::string, ThingControlBlock **> m_position;
	ThingControlBlock **m_official_position;





	//////////////////////////////////////////////////////////////////////
	// Benched players (currently not on ice)
	//////////////////////////////////////////////////////////////////////
	std::map <const std::string, ThingControlBlock *> m_bench;





	//////////////////////////////////////////////////////////////////////
	// Official team statistics and properties
	//////////////////////////////////////////////////////////////////////

	//    Total number of players allowed on the ice. This "normally" varys
	// between 6 and 4 depending on penalties. The players on the ice can
	// be from any position with the exception that only one goalie is
	// allowed on the ice at any one time. The normal setup is, 3 forwards,
	// 2 defense and one goalie. However combinations such as 6 forwards or
	// 6 defense men are allowed.
	//    In the unlikely event that the official allows more then 6
	// players, then coaches are free to "ice" more players, (though the
	// one goalie limit might still apply).
	int m_max_iced_players;

	// The lines that are currently scheduled to be on the ice. The
	// standard combination is "Fn, Dn, Gn" where "n" is the line number
	// for each line. Fowards are are composed of a line of three players,
	// defense have two players, and goalies are alone on a line. There is
	// also an "E" line for "extra attackers". Some other lines are used
	// for "special teams", such as the "PK" and "PP" lines.
	std::vector <std::string> m_sched_lines;



	std::string add_person(
		std::map <const std::string, ThingControlBlock *> *p_list,
		ThingControlBlock *p_tcb);


	friend gmTE phTeam_compile_uniform_patches( const gmCCArg &p);
	friend gmTE phTeam_set_uniform( const gmCCArg &p);
	friend gmTE phTeam_add_person( const gmCCArg &p);
	friend gmTE phTeam_add_person_list( const gmCCArg &p);
	friend gmTE phTeam_dress( const gmCCArg &p);

private:
	phTeam( const phTeam &);
	phTeam &operator =( const phTeam &);
};

#endif
