/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "user_control.h"
#include "camera.h"
#include "team.h"
#include "player.h"
#include "arena.h"
#include "puck.h"
#include "ai.h"
#include "ai_player.h"
#include "ai_skater.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"

std::map <const std::string, THING_COMMAND_FN> phUserControl::m_ccmd;

extern "C" gmThing *spw_phUserControl()
{ return new phUserControl();}

extern "C" int del_phUserControl( gmThing *t)
{ delete t; return 0;}

RPH_THING_MODULE_MAIN_FUNCTION( phUserControl)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phUserControl",
		spw_phUserControl);
	p_ge->m_thing_manager->add_to_delete_list( "phUserControl",
		del_phUserControl);

	return 0;
}

phUserControl::phUserControl()
{
	m_team = 0;
	m_move_flags = 0;

	m_cursor_alpha = 0.5f;
	m_cursor_scale = rgVector( 1, 1, 1);

	if( !(m_ccmd.empty())) return;
	add_console_command( "select", phUserControl_select_player, m_ccmd);
	add_console_command( "move", phUserControl_move, m_ccmd);
	add_console_command( "speed", phUserControl_speed, m_ccmd);
	add_console_command( "check", phUserControl_check, m_ccmd);
	add_console_command( "pass", phUserControl_pass, m_ccmd);
	add_console_command( "puck", phUserControl_puck, m_ccmd);
	add_console_command( "shoot", phUserControl_shoot, m_ccmd);
	add_console_command( "stick", phUserControl_stick, m_ccmd);
	add_console_command( "control_team", phUserControl_control_team,
		m_ccmd);
	add_console_command( "uc_set", phUserControl_uc_set, m_ccmd);
}

phUserControl::~phUserControl()
{
}

gmThingError phUserControl::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return phAI::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phUserControl::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phUserControl::type() + "::" + (iter->first)
			+ "\n";
	}
	tmp += phAI::list_ccmd( p_prefix);
	return tmp;
}

gmTE phUserControl_select_player( const gmCCArg &p)
{
	//const std::vector <std::string> &args = *(p.args);
	//if( args.size() < 1) return 0;
	phUserControl *uc = (phUserControl *)(p.thing);

	// If controlled player has the puck, then do nothing.
	phPerson *per = uc->get_host_person();
	if( per && (per->has_game_puck())) return 0;
	uc->unselect_player();

	// Choose only from players (from controlled team) that are on the ice.
	// FIXME: Cull incapacitated players from "plist" (don't want to select
	//	players that are knocked out for example).
	phTeam *t = uc->get_team();
	if( !t) return 0;
	std::vector <phPerson *> plist;
	t->get_people( plist, "ice");
	if( plist.empty()) return 0;


	// Choose player that is nearest to the puck.
	phArena *a = t->get_arena();
	if( !a) return 0;
	per = a->nearest_puck( plist);
	if( !per) return 0;

	// Start controlling the selected player.
	if( uc->set_host( per->get_ai())) {
		uc->update_player_cursor( per);
		phPersonCursor *pc = per->get_cursor();
		if( pc) {
			pc->m_color = uc->m_cursor_color;
			pc->m_alpha = uc->m_cursor_alpha;
			pc->m_mat.loadIdentity();
			pc->m_mat.scale( uc->m_cursor_scale);
		}

		std::cout << "selected \"" + (per->get_ai())->m_key + "\"\n";
	}

	return 0;
}

gmTE phUserControl_move( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() < 1) return 0;
	phUserControl *uc = (phUserControl *)(p.thing);

	// Setting a direction of movement.
	const std::string &direction = args[0];
	if( "up" == direction) uc->set_move_up();
	else if( "down" == direction) uc->set_move_down();
	else if( "left" == direction) uc->set_move_left();
	else if( "right" == direction) uc->set_move_right();

	// Unsetting a direction of movement.
	else if( "!up" == direction) uc->unset_move_up();
	else if( "!down" == direction) uc->unset_move_down();
	else if( "!left" == direction) uc->unset_move_left();
	else if( "!right" == direction) uc->unset_move_right();

	phAI *ai = uc->get_host_ai();
	if( !ai) return 0;

	// If not moving, then stand in ready position.
	if( uc->no_movement()) {
		ai->set_single_task( new task___ply_ready_stand);
		return 0;
	}

	// Move in specified heading.
	task___skate_heading *t = new task___skate_heading;
	phCamera *cam =
		get_camera_controller( p.ge->gfx_engine->current_camera());
	if( cam)
		t->m_heading = cam->viewer_to_world( uc->move_heading());

	// Set movement speed and intensity (and respective animations).
	// FIXME: Get skating styles (animations) from skater attributes.
	t->m_intensity = "fast";
	t->m_anim = "fast_skate";

	t->m_left_cross_cut = "cross_cut_left";
	t->m_right_cross_cut = "cross_cut_right";

	t->m_left_stop = "lean_left_deep_stop";
	t->m_right_stop = "lean_right_deep_stop";

	ai->set_single_task( t);
	return 0;
}

gmTE phUserControl_speed( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() < 1) return 0;
	phUserControl *uc = (phUserControl *)(p.thing);

	phPerson *per = uc->get_host_person();
	if( !per) return 0;

	per->m_datt_speed = args[0];

//	phAI *ai = uc->get_host_ai();
//	if( !ai) return 0;

	return 0;
}

gmTE phUserControl_check( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() < 1) return 0;
	phUserControl *uc = (phUserControl *)(p.thing);
	uc = uc;
	return 0;
}

gmTE phUserControl_pass( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() < 1) return 0;
	phUserControl *uc = (phUserControl *)(p.thing);
	uc = uc;
	return 0;
}

gmTE phUserControl_puck( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() < 1) return 0;
	phUserControl *uc = (phUserControl *)(p.thing);

	phPerson *per = uc->get_host_person();
	if( !per) return 0;

	phAI *ai = uc->get_host_ai();
	if( !ai) return 0;


	const std::string &act = args[0];
	if( "get" == act) {
		if( per->has_game_puck()) return 0;
		ai->set_single_task( new task___ply_get_puck_control);
	}
	return 0;
}

void phUserControl_shoot_callback( task___ply_shoot *t)
{
	if( !t->m_uc) return;
	if( !t->m_per) return;
	phPersonCursor *pc = t->m_per->get_cursor();
	if( !pc) return;
	pc->m_tex_name = t->m_uc->m_cursor_tex;
}

gmTE phUserControl_shoot( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() < 1) return 0;
	phUserControl *uc = (phUserControl *)(p.thing);

	// Controlled player must have puck inorder to shoot it.
	phPerson *per = uc->get_host_person();
	if( (!per) || (!per->has_game_puck())) return 0;

	phAI *ai = uc->get_host_ai();
	if( !ai) return 0;

	bool windup = false;
	if( (args.size() >= 2) && ("windup" == args[1])) windup = true;

	phCamera *cam = get_camera_controller(
		p.ge->gfx_engine->current_camera());
	if( !cam) return 0;

	task___ply_shoot *t = new task___ply_shoot;
	t->m_heading = cam->viewer_to_world( uc->move_heading());

	const std::string &shot_type = args[0];
	if( "wrist" == shot_type) {
		if( windup) t->m_anim = per->get_action_anim( "wrist_windup");
		else t->m_anim = per->get_action_anim( "wrist_shot");
	}

	t->m_uc = uc;
	t->m_per = per;
	t->m_call_back = phUserControl_shoot_callback;
	t->m_windup = windup;
	ai->set_single_task( t);

	return 0;
}

gmTE phUserControl_stick( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() < 1) return 0;
	phUserControl *uc = (phUserControl *)(p.thing);
	uc = uc;
	return 0;
}

gmTE phUserControl_control_team( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return "usage: control_team <team>\n";
	ThingControlBlock *t = p.ge->m_thing_manager->get( args[0]);
	if( !t) return "No such thing \"" + args[0] + "\"\n";
	const std::string emsg( ((phUserControl *)(p.thing))->set_team( t));
	if( emsg.length() > 0) return emsg + "\n";
	return 0;
}

gmTE phUserControl_uc_set( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() <= 0)
		return "usage: uc_set <attrib1 = value1> [attrib2 = value2] "
			"... [attrib_n = value_n]\n";

	phUserControl *uc = (phUserControl *)(p.thing);
	int i = 0;
	for( ; (i + 2) < int( args.size()); i += 3) {

		// Check for value.
		if( "=" != args[i + 1]) {
			return "expected \"=\" after \"" + args[i] +
				"\" attribute.";
		}
		const std::string &value = args[ i + 2];

		if( "cursor_tex" == args[i])
			uc->m_cursor_tex = value;

		else if( "cursor_puck_tex" == args[i])
			uc->m_cursor_puck_tex = value;

		else if( "cursor_red" == args[i])
			uc->m_cursor_color.x = strtod( value.c_str(), 0);
		else if( "cursor_green" == args[i])
			uc->m_cursor_color.y = strtod( value.c_str(), 0);
		else if( "cursor_blue" == args[i])
			uc->m_cursor_color.z = strtod( value.c_str(), 0);
		else if( "cursor_alpha" == args[i])
			uc->m_cursor_alpha = strtod( value.c_str(), 0);

		else if( "cursor_scale_x" == args[i])
			uc->m_cursor_scale.x = strtod( value.c_str(), 0);
		else if( "cursor_scale_y" == args[i])
			uc->m_cursor_scale.y = strtod( value.c_str(), 0);
		else if( "cursor_scale_z" == args[i])
			uc->m_cursor_scale.z = strtod( value.c_str(), 0);

		else {
			return ("uc_set: unknown attribute \"" + args[i] +
				"\"\n");
		}
	}
	if( i != int( args.size()))
		return ("expected \"" + args[i] + " = <value>\" for "
			"last attribute");
	return 0;
}

phPerson *phUserControl::get_host_person()
{
	phAI *ai = get_host_ai();
	if( !ai) return 0;
	ThingControlBlock *t = ai->get_host();
	if( !t) return 0;
	if( !(t->m_thing->is_a( "phPerson"))) return 0;
	return (phPerson *)(t->m_thing);
}

void phUserControl::update_player_cursor( phPerson *per)
{
	if( !per) return;
	if( per->has_game_puck()) per->set_cursor_texture( m_cursor_puck_tex);
	else per->set_cursor_texture( m_cursor_tex);
}

void phUserControl::unselect_player()
{
	phPerson *per = get_host_person();
	if( !per) return;
	per->set_cursor_texture( std::string());
	set_host( 0);
}

void phUserControl::process( ThingManager *tm)
{
}

