/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef USER_CONTROL_H
#define USER_CONTROL_H

#include "../things/things.h"
#include "ai.h"

class phTeam;
class phPerson;

class task___ply_shoot;

class phUserControl : public phAI
{
public:
	RPH_THING_MODULE

	phUserControl();
	virtual ~phUserControl();

	std::string type() const { return std::string( "phUserControl");}
	bool is_a( const std::string &s) const {
		if( s != phUserControl::type()) return phAI::is_a( s);
		return true;
	}

	gmTE call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	void process( ThingManager *tm);

	std::string set_team( ThingControlBlock *t) {
		m_team = 0;
		if( !t) return "t == NULL";
		if( t->is_free()) return "Thing is free (not read)";
		if( !(t->m_thing->is_a( "phTeam"))) return
			"Thing \"" + t->m_key + "\" is not a \"phTeam\"";
		m_team = t;
		return std::string();
	}
	phTeam *get_team() {
		if(	!m_team ||
			(m_team->is_free()) ||
			!(m_team->m_thing->is_a( "phTeam"))) return 0;
		return ((phTeam *)(m_team->m_thing));
	}

	phPerson *get_host_person();
	phAI *get_host_ai() {
		ThingControlBlock *ai = get_host();
		if( !ai) return 0;
		return (phAI *)(ai->m_thing);
	}


protected:


	// This AI was designed for controlling other AI's.
	bool accept_host( ThingControlBlock *t) {
		if( t->is_free()) return false;
		if( !(t->m_thing->is_a( "phAI"))) return false;
		return true;
	}

	// Each controller can only select hosts beloning to a specific team.
	ThingControlBlock *m_team;

	std::string m_cursor_tex;
	std::string m_cursor_puck_tex;
	rgVector m_cursor_color;
	rgVector m_cursor_scale;
	rgFloat m_cursor_alpha;

	#define UC_MOVE_UP_BIT		0x1
	#define UC_MOVE_DOWN_BIT	0x2
	#define UC_MOVE_LEFT_BIT	0x4
	#define UC_MOVE_RIGHT_BIT	0x8
	rgVector move_heading() const {
		rgVector v;
		if( m_move_flags & UC_MOVE_UP_BIT) v += rgVector( 0, 1, 0);
		if( m_move_flags & UC_MOVE_DOWN_BIT) v += rgVector( 0, -1, 0);
		if( m_move_flags & UC_MOVE_LEFT_BIT) v += rgVector( -1, 0, 0);
		if( m_move_flags & UC_MOVE_RIGHT_BIT) v += rgVector( 1, 0, 0);
		return (v.normalize());
	}

	bool no_movement() const { return (!m_move_flags);}

	void set_move_up() {
		unset_move_down();
		m_move_flags |= UC_MOVE_UP_BIT;}

	void set_move_down() {
		unset_move_up();
		m_move_flags |= UC_MOVE_DOWN_BIT;}

	void set_move_left() {
		unset_move_right();
		m_move_flags |= UC_MOVE_LEFT_BIT;}

	void set_move_right() {
		unset_move_left();
		m_move_flags |= UC_MOVE_RIGHT_BIT;}

	void unset_move_up() { m_move_flags &= ~UC_MOVE_UP_BIT;}
	void unset_move_down() { m_move_flags &= ~UC_MOVE_DOWN_BIT;}
	void unset_move_left() { m_move_flags &= ~UC_MOVE_LEFT_BIT;}
	void unset_move_right() { m_move_flags &= ~UC_MOVE_RIGHT_BIT;}
	#undef UC_MOVE_UP_BIT
	#undef UC_MOVE_DOWN_BIT
	#undef UC_MOVE_LEFT_BIT
	#undef UC_MOVE_RIGHT_BIT

	void update_player_cursor( phPerson *per);
	void unselect_player();


	friend void phUserControl_shoot_callback( task___ply_shoot *t);

	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;
	friend gmTE phUserControl_select_player( const gmCCArg &p);
	friend gmTE phUserControl_control_team( const gmCCArg &p);
	friend gmTE phUserControl_uc_set( const gmCCArg &p);

	friend gmTE phUserControl_move( const gmCCArg &p);
	friend gmTE phUserControl_speed( const gmCCArg &p);
	friend gmTE phUserControl_check( const gmCCArg &p);
	friend gmTE phUserControl_pass( const gmCCArg &p);
	friend gmTE phUserControl_puck( const gmCCArg &p);
	friend gmTE phUserControl_shoot( const gmCCArg &p);
	friend gmTE phUserControl_stick( const gmCCArg &p);

private:
	int m_move_flags;


	phUserControl( const phUserControl &);
	phUserControl &operator =( const phUserControl &);
};

#endif
