/*
Copyright (C) 2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef ENGINE_FILE_H
#define ENGINE_FILE_H

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <rgPort.h>

class GameEngine;
class FileEngine;

#define MAX_ARCHIVE_FILE_SIZE	67108864	// 64 MB

class gmFileHandle : public rgPortFileHandle
{
public:
	std::string m_emsg;
	int (*m_cfile_close_fn)( FILE *);

	gmFileHandle( FileEngine *p_fe) {
		m_fe = p_fe;
		m_cfile_close_fn = 0;
	}

	bool open_file( const std::string &p_fname);
	FILE *open_c_file( const std::string &p_fname);
	void close_c_file( FILE *f) {
		if( !m_cfile_close_fn) return;
		m_cfile_close_fn( f);
	}

	void set_c_file( FILE *f) { rgPortFileHandle::set_c_file( f);}

protected:
	FileEngine *m_fe;

private:
	gmFileHandle( const gmFileHandle &);
	gmFileHandle &operator =( const gmFileHandle &);
};

class FileEngine
{
public:
	// When loading files, the current working directory (cwd) of the
	// game is checked first. If cwd does not contain the file then
	// compressed media files in "m_media_files" are searched (inorder).
	std::vector <std::string> m_media_files;

	FileEngine( GameEngine *p_ge);

	std::string get_file( const std::string &p_fname, rgPortFileHandle &f);
	std::string get_c_file( const std::string &p_fname, gmFileHandle &f);

	const std::string &temp_directory() const { return m_temp_dir;}

	std::string create_temporary_file( FILE **f, std::string &fname) const;
	std::string create_temporary_file( std::fstream &tmp,
		std::string &fname) const;
	int create_temporary_file( std::string &fname) const;

	void set_temporary_directory( const std::string &dir) {
		m_temp_dir = dir;}

	void set_file_memory_buffer_size( const int n) {
		if( n < 0) return;
		m_max_archive_file_size = n;
	}
	int file_memory_buffer_size() const { return m_max_archive_file_size;}

protected:
	GameEngine *m_ge;

	// The place where temporary files or media can be stored.
	std::string m_temp_dir;

	int m_max_archive_file_size;

private:

	FileEngine( const FileEngine &);
	FileEngine &operator =( const FileEngine &);
};

//int gm_get_from_zip( const std::string &p_zip, const std::string &p_file,
//	std::iostream &s, const int len, std::ostream &p_swap);

#endif
